;; monk-midi.el --- A phony midi track generator -- no shuttle search, jump, loop

;; Sun Dec 31 01:08:43 2000: Started.
;; Tue Feb 26 15:17:45 2002: Fixed, attached version # 0.1.

;;
;; "All are rabish (aren't they)" -- (c) 2000 Shna Ring
;;

(defconst monk-midi-version "0.1")

(require 'monk)
(require 'dired-aux)
(provide 'monk-midi)

(defvar monk-playmidi-command "playmidi" "midi player command.")

(defvar monk-timidity-command "timidity" "midi player command.")

(defvar monk-midi-probe-command nil
  "midi probe command. Set to \"timidity\" if you have it.
Other command perhaps won't work.")

(defvar monk-midi-probe-options '("-idv -Ow -o-")
 "Options for `monk-midi-probe-command'. timidity\(1\) is expected.")

;; Just returns size in a list

(defsubst monk-midi-probe-rabish (file)
  "Probe midi FILE. Rabish."
  (list (cons 'frames (nth 7 (file-attributes file)))))

(defsubst monk-midi-parse (raw)
  "Parse RAW data, build an alist, and return it.
Builds at least \(time . <second>\) data in out.
Rest of records are stored under `spec<N>' key."
  (let ((count 0))
    (mapcar
     (lambda (d)
       (cond
	((string-match "time \\([0-9]+:[0-9]+\\)" d)
	 (cons 'time
	       (monk-cdtime2sec
		(substring d (match-beginning 1) (match-end 1)))))
	(t
	 (cons (make-symbol (format "spec%03d" (setq count (1+ count)))) d))))
     raw)))

(defsubst monk-midi-probe-subr (file)
  "Execute unix command timidity\(1\) to get infos of midi FILE.
Require `monk-midi-probe-command' to be \"timidity\"."
  (cond
   ((file-readable-p file)
    ;; (save-excursion (set-buffer (get-buffer-create monk-info-buffer))
    (monk-with-current-buffer
     (get-buffer-create monk-info-buffer)
     (setq default-directory
	   (monk-put-trailing-/ (file-name-directory file)))
     (erase-buffer)
     (cond
      ((= 0 (call-process
	     ;; I started to think `shell-file-name' might be dengerous (?)
	     ;; shell-file-name
	     "/bin/sh"
	     nil t nil "-c"
	     (concat 
	      monk-midi-probe-command
	      " "
	      (apply 'concat monk-midi-probe-options)
	      (dired-shell-stuff-it "" (list file) nil)
	      " 2>&1 >/dev/null | sed '/^[0-9].*time/q'")))
       (delete
	"" (dired-split
	    "[\n\t]" (buffer-substring (point-min) (point-max))))))))))

(defsubst monk-midi-probe (file)
  "Probe midi FILE with and return the rough-cut result as an alist."
  (cond
   ((null monk-midi-probe-command) (monk-midi-probe-rabish file)) ; Act moronic
   (t (let ((raw (monk-midi-probe-subr file)))
	(cond (raw
	       (append
		(monk-midi-probe-rabish file)
		(monk-midi-parse raw))))))))

(defun monk-midi-build-track-obj (file command)
  "Build MonK `track' object for midi FILE.
COMMAND is used as `command' property."
  (let ( ;; (file (expand-file-name file)) ;; `monk-find-file' does it.
	(fun (lambda (s) (string-match "^.*: \\([\.0-9]+\\)$" s)))
	lis sz tm)
    (cond
     ((setq lis (monk-midi-probe file))
      (setq sz (monk-assq-get-value 'frames lis)
	    tm (monk-assq-get-value 'time lis))
      ;; frames is phony
      (append
       (delq
	nil (list (cons 'file file)
		  (cons 'player command)
		  (cons 'track-opt nil)
		  (cons 'offset-opt nil)
		  (cons 'duration-opt nil)
		  (cons 'post-filter nil)
		  (if tm (cons 'time tm))
		  (if tm (cons 'fps  (round (/ sz tm))) (cons 'fps 1))
		  (cons 'cent  (/ sz 100.0))
		  ;; just prevent error. timer display is really rabish
		  (cons 'type 'mid))) lis)))))

(defsubst monk-midi-insert-track-obj-subr (obj)
  "Insert midi `track' object."
  (let (fname frames time pt)
    (setq fname (monk-assq-get-value 'file obj)
	  frames (monk-assq-get-value 'frames obj)
	  time (monk-assq-get-value 'time obj)
	  pt (point))
    (monk-insert-track obj "mid:"
		       (cond (time
			      (format "%5s.00" (monk-sec2cdtime time)))
			     (t "   NA   "))
		       0 frames
		       (format "\"%s\""
			       (file-name-nondirectory
				fname)) 'newline)
    obj))

(defun monk-playmidi-insert-track-obj (file)
  "Insert track line for playmidi\(1\) at point."
  (interactive "fmidi File:")
  (let ((obj (monk-midi-build-track-obj file monk-playmidi-command)))
    (cond (obj (monk-midi-insert-track-obj-subr obj)))))

(defun monk-timidity-insert-track-obj (file)
  "Insert track line for timidity\(1\) at point."
  (interactive "fmidi File:")
  (let ((obj (monk-midi-build-track-obj file monk-timidity-command)))
    (cond (obj (monk-midi-insert-track-obj-subr obj)))))

;;
;; (monk-timidity-insert-track-obj "~/midi/grief-excerpt.mid") 
;;  mid: [03:17.00]        0     3795 "grief-excerpt.mid"
;; ((file . "~/midi/grief-excerpt.mid")
;;  (player . "timidity")
;;  (track-opt)
;;  (offset-opt)
;;  (duration-opt)
;;  (post-filter)
;;  (time . 197.0)
;;  (fps . 19)
;;  (cent . 37.95)
;;  (type . mid)
;;  (frames . 3795)
;;  (spec001 . "Init soundfonts `8RealGS20.SF2'")
;;  (spec002 . "Playing /home/sn/midi/grief-excerpt.mid")
;;  (spec003 . "MIDI file: /home/sn/midi/grief-excerpt.mid")
;;  (spec004 . "Format: 1  Tracks: 1  Divisions: 192")
;;  (spec005 . "Sequence: str")
;;  (spec006 . "MIDI reset at 0 sec")
;;  (time . 197.0))
;;

;;
;; A function found in my dired-dd-timidity.el.
;;
;; Can be used as
;; (progn (timidity)
;;   (raw-timidity-load-file "~/imm/00midi/prs/triple-1.mid"))
;;

(defun monk-raw-timidity-load-file (fname)
  "Load file with FNAME into `timidiy' buffer and play it."
  ;; (interactive "fMIDI file: ")
  (require 'timidity)			; added for monk
  (monk-stop)				; added for monk
  (if fname
      (progn
	(setq timidity-playing-flag t)
	(timidity-run)
	(setq fname ;; (timidity-expand-file-name fname) ;; too much
	      (expand-file-name fname)) ; good enough
	(garbage-collect)
	(process-send-string timidity-cmd-process
			     (format "L\nPLAY %s\n" fname))
	(message "MIDI File: %s" fname))
    (setq timidity-playing-flag nil)))

;;
;; Run emacs-command `timidy' then play a track with it.  Works.
;;
;;  Should be an official command ?

(and monk-mode-map (keymapp monk-mode-map)
     (define-key monk-mode-map "\e\C-m" 'monk-play-midi-timidity-el))
;;
;; My `timidity-mode-menu.el' and `dired-dd-timidity.el' may be
;; for convenient for "*TiMidity*". They are in my `dired-dd' package.
;;

(defun monk-play-midi-timidity-el ()
  "Play current `midi' track through `timidity' emacs command.
You should have installed timidity.el already."
  (interactive)
  (require 'timidity)
  (let* ((tr (monk-get-track))
	 (file (monk-assq-get-value 'file tr))
	 (type (monk-assq-get-value 'type tr))
	 (buf (get-buffer "*TiMidity*")))
    (cond
     ((eq type 'mid)
      (cond
       ((monk-buffer-live-p buf)
	(switch-to-buffer-other-frame buf)
	(monk-raw-timidity-load-file file))
       (t
	(select-frame (make-frame))
	(timidity)
	(monk-raw-timidity-load-file file))))
     (t (message "The track type is not `mid'")))))
