;; monk-prop-edit.el --- edit `track' property list 

;; This file is still under-construction

;; major mode version
;; Mon Dec 25 19:46:49 2000
;; Wed Dec 27 11:37:32 2000 New track generation cluge.

(require 'pp)
(require 'monk)				; not needed if merged
(provide 'monk-prop-edit)

(autoload 'monk-write-id3-info "monk-mp3" nil t) ;; not yet.

(defconst monk-prop-edit-buffer
  "*monk*prop*edit*" "Buffername for `monk-prop-edit'.")

(defvar monk-prop-edit-mode-map
  (let ((monk-prop-edit-mode-map (copy-keymap emacs-lisp-mode-map)))
    (define-key monk-prop-edit-mode-map "\C-c\C-c" 'monk-prop-edit-finish)
    (define-key monk-prop-edit-mode-map "\C-c\C-g" 'monk-prop-edit-abort)
    (define-key monk-prop-edit-mode-map "\e\C-i" 'monk-complete-prop)
    (define-key monk-prop-edit-mode-map "\C-cI" 'monk-write-id3-info)
    monk-prop-edit-mode-map) "Keymap used in `monk-prop-edit-mode'.")

(defvar monk-prop-edit-mode-hook nil
  "Normal hook run when starting to lkk-j-mode.")

(defun monk-prop-edit-mode ()
  "A mode to edit MonK `track' data, used in `monk-prop-edit-buffer'.

Available key bindings (almost the same as `emacs-lisp-mode' bindings
except \\[monk-prop-edit-finish], and \\[monk-prop-edit-abort]):

\\{monk-prop-edit-mode-map}"
  (interactive)
  ;; (message "This feature is under-construction, so watch out, sister")
  (kill-all-local-variables)
  (setq truncate-lines monk-truncate-lines)
  (use-local-map monk-prop-edit-mode-map)
  (setq indent-line-function 'lisp-indent-line)
  (setq major-mode 'monk-prop-edit-mode)
  (setq mode-name "MonKPropEdit")
  (and window-system (font-lock-mode t)) ; cheap.
  (run-hooks 'monk-prop-edit-mode-hook))

;; To allow even altering *monk* buffer while `monk-prop-edit'.
(defvar monk-prop-edit-marker
  nil "Marker points edit location of `monk-prop-edit'.")

;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc
;; Dummy `track' prop template inserted for a new track
(defvar monk-prop-dummy-template
  '((file . "")
    (player . "")
    (track-opt)
    (offset-opt)
    (duration-opt)
    (options)
    (post-filter)
    (frames)			; required if track time shoud work
    (time)			; required if track time shoud work
    (cent)			; frames / 100
    (fps)			; frames / time
    (type)))
;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc

(fset 'monk-track-edit 'monk-prop-edit)

;;;###autoload
(defun monk-prop-edit (&optional arg readonly)
  "Create a fresh buffer *monk*prop*edit* and insert 'track property.
The track head \(buffer position\) is stored in `monk-prop-edit-marker'.
Optional ARG is currently just ignored, and 2nd optional arg READONLY
causes to open edit buffer with readonly status.
There is no facility to save back the track property to file (ie. id3 info)."
  (interactive "P\nP")
  (let ((m (save-excursion (forward-line 0) (point-marker)))
	(prop (monk-get-track 'raw)) ; No 'file validation check.
	(buf (get-buffer-create monk-prop-edit-buffer))
	newtrack)
    (cond
     ((and prop (listp prop)) (setq prop prop))
     (t (setq newtrack t
	      ;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc
	      ;; Dummy `track' prop template while development of this file
	      prop (copy-list monk-prop-dummy-template)
	      ;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc
	      )))
    (cond
     ((and newtrack readonly)
      (error "You're not on a track, babe"))
     (newtrack
      (message
       "New track created. Hit C-c C-c when edit completed, a#g#f#edc ...")
	       (let (buffer-read-only)	; Still in `monk-buffer'
		 (save-excursion
		   (forward-line 0) (insert ">>NEW TRACK<<\n")))))
    (setq monk-prop-edit-marker m)
    (monk-with-current-buffer
     buf
     (setq buffer-read-only nil)	; the last edit might be in readonly
     (erase-buffer)
     (save-excursion (pp prop (current-buffer)))
     (set-buffer-modified-p nil)
     (cond ((null readonly)
	    (monk-prop-edit-mode)
	    ;;(switch-to-buffer-other-window buf)
	    )))
    ;; Can it make more straight forward ?
    ;; `monk-show-buffer' is used here. *Help* mode is not what we want.
    (cond (readonly
	   (display-buffer buf)
	   (monk-with-current-buffer
	    ;; This `buf' should not be `monk-prop-edit-buffer' reference. Why ?
	    buf				
	    (monk-prop-edit-mode)
	    (setq buffer-read-only t)))
	  (t (switch-to-buffer-other-window buf)))))

;; Merged already:
;; (define-key monk-mode-map "\C-ce" 'monk-edit)
;; (define-key monk-mode-map "\C-c\C-e" 'monk-prop-edit)

(defun monk-prop-edit-finish ()
  "Put prop back to original MonK buffer.
Track line is re-inserted according to the new `track' property."
  (interactive)
  (let ((val (save-excursion (goto-char (point-min))
			     (read (current-buffer))))
	(pt (marker-position monk-prop-edit-marker))
	(buf (marker-buffer monk-prop-edit-marker)))
    (cond
     ((and pt buf)			; allow nil as val. okay ?
      ;;(save-excursion (set-buffer buf)
      (monk-with-current-buffer
       buf
       (let* ((buffer-read-only)
	      (file (monk-assq-get-value 'file val))

	      ;; the track in *monk* value.
	      (prop (monk-get-track 'raw))
	      (on-playing-track
	       ;; (monk-within-overlay-p)
	       (and (monk-proc-running-p)
		    ;; (equal prop monk-current-track)
		    (monk-on-playing-track-p monk-current-track prop))))
	 ;; At least filename should be expanded.  Or `monk-play-start' should
	 ;; be modified to expand it ?  Further, "~/bin/sox" sort of 'player
	 ;; is not supported.  Is that rational ?
	 (cond
	  (file
	   (setq
	    val (monk-prop-set-val val (cons 'file (expand-file-name file))))))

	 (goto-char pt)
	 ;; NOP gracefully if empty line, but should not modify kill-ring.
	 ;; (end-of-line) (kill-line 0)
	 (end-of-line) (delete-region
			(save-excursion (forward-line 0) (point)) (point))
	 (monk-insert-track-versatile val nil)
	 (when on-playing-track
	   (setq monk-current-track val)
	   (when (monk-scratching-p)
	     (setcar monk-loop-data val))
	   ;; RELOCATE `monk-process-marker'.
	   (monk-set-process-marker)
	   (monk-play-take-control-again val)
	   ;; Do not support re-face in emacs-19.28.
	   (or monk-buggy-make-overlay
	       (monk-face-current-track)))
	 ;; This rebuild destroys the edit result: wrong thing to do.
	 ;; Eg. User may want to C-c I (monk-write-id3-info) in *monk*
	 ;; after C-c C-c (this-command). Or, do monk-write-id3-info()
	 ;; here is natural implementation (???)
	 ;; (when monk-rebuild-track-auto (monk-rebuild-track))
	 ))
      (monk-prop-edit-cleanup)
      (pop-to-buffer buf) ))))

(defun monk-prop-edit-cleanup ()
  (move-marker monk-prop-edit-marker nil)
  ;; This should not be done.
  ;; (setq monk-prop-edit-marker nil)
  (cond ((one-window-p) (bury-buffer))
	(t ;;(delete-window (selected-window))
	 (kill-buffer (current-buffer))
	   )))

(defun monk-prop-edit-abort ()
  "Abort prop edit."
  (interactive)
  (let ((pt (marker-position monk-prop-edit-marker))
	(buf (marker-buffer monk-prop-edit-marker)))
    ;; (save-excursion (set-buffer buf)
    (monk-with-current-buffer		; `save-excursion' implicit
     buf
     (goto-char pt)
     (cond ((looking-at ">>NEW TRACK<<\n")
	    (let ((buffer-read-only))
	      (delete-region (match-beginning 0) (match-end 0))))))
    (monk-prop-edit-cleanup)
    (pop-to-buffer buf)			; buf may be nil
    (message "Track edit aborted")
    nil))

;; Fastest alist edit function ever seen. Now in monk.el.
;; (defun monk-prop-set-val (lis newelem) ...)

;;;
;;; Completion System --- M-C-i (monk-complete-prop)
;;;
;;; Started as: monk-id3.el
;;; Merged:  Sun Jan 28 22:23:12 2001
;;; Started: Mon Jan 22 03:51:08 2001, s.n.
;;;

;;
;; `monk-complete-prop' -- command name is right ?
;; for *monk*propedit* buffer.
;;

(defun monk-complete-prop (&optional table)
  "Completes words from `monk-completion-dict', and
optionally `monk-completion-dict' if it is available."
  (interactive)
  ;; (require 'monk-mp3) or (require 'monk-id3) if this is merged to monk.el,
  ;; and others are in monk-mp3 or monk-id3.
  (let* ((b (save-excursion (backward-sexp 1) (point)))
	 (e (point))
	 (seed (buffer-substring b e))
	 (table (or table (append monk-completion-dict
				  (when (boundp 'monk-id3-genre-table)
				    monk-id3-genre-table))))
	 (c (try-completion seed table 'identity)))
    (cond
     ((eq c t))
     ((null c)
      (message "MonK can't complete \"%s\"" seed))
     ((null (string= seed c))
      (delete-region b e)
      (insert c))
     (t (message "MonK making completion list...")
	(let ((lis (sort (all-completions seed table) 'string<)))
	  (with-output-to-temp-buffer "*Completions*"
	    (display-completion-list lis)))
	(message "Making completion list...done")))))

;; Should be in monk.el ?
(defvar monk-completion-dict
  '(("track-opt")
    ("offset-opt")
    ("duration-opt")
    ("track-number")
    ("track-offset")
    ("options")
    ("file")
    ("post-filter")
    ("player")
    ("frames")
    ("time")
    ("cent")
    ("fps")
    ("type")
    ("title")
    ("artist")
    ("album")
    ("year")
    ("comment")
    ("genre")) "MonK's track property keywords to be completed. Rough.")
