;; monk-smpeg.el --- phony mpeg I layer 2 track generation.
;; Wed Dec 27 18:48:23 2000

;;
;; plaympeg(1) from smpeg-0.4.2 works nicely with
;; "plaympeg --novideo --noaudio"
;; No need to redirect stdout/stderr separately, so we can do it
;; without `shell-file-name' and `dired-shell-stuff-it'.
;; 
;; But, another headache: plaympeg(1) always return 0:
;; $ plaympeg --noaudio --novideo /etc/fstab ; echo  $?
;; /etc/fstab: Could not find the beginning of MPEG data
;; 
;; 0
;; $ plaympeg --noaudio --novideo /Gibbs.mp  ; echo $? 
;; /Gibbs.mp: No such file or directory
;; 0
;;

(provide 'monk-smpeg)

(defvar monk-smpeg-command "plaympeg" "mpeg cinema player from smpeg project.")

(defvar monk-smpeg-command-options nil "*Options to `monk-smpeg-command'.")

;; No `monk-smpeg-probe-command' & `monk-smpeg-probe-options' in this file.
;; The probe parameters are hard-coded in this function:

(defun monk-smpeg-probe (file)
  "Probe mpeg cinema FILE with plaympeg\(1\), and return result \
as roughly split list of string."
  (cond
   ((file-readable-p file)
    (monk-with-current-buffer
     (get-buffer-create monk-info-buffer)
     (setq default-directory
	   (monk-put-trailing-/ (file-name-directory file)))
     (erase-buffer)
     (cond
      ((and
	(= 0 (call-process "plaympeg" nil t nil "--novideo" "--noaudio" file))
	;; plaympeg(1) from smpeg-0.4.2 always return 0 even if it failed.
	;; So, here we should check typical error message from plaympeg.
	;; Perhaps this works, since we `expanded' file name.
	(progn (goto-char (point-min)) (null (looking-at "^/"))))
       (delete ""
	       (dired-split "[\n\t]"
			    (buffer-substring (point-min) (point-max))))))))
   (t (message "MonK can't read read file %s" file))))

(defun monk-smpeg-build-track-obj (file)
  (let ((file (expand-file-name file))
	(fun (lambda (s) (string-match "^.*: \\([\.0-9]+\\)$" s)))
	lis spec vd ad sz tm)
    (cond
     ((setq lis (monk-smpeg-probe file))
      (mapcar
       (lambda (e)
	 (cond 
	  ((string-match "^Video" e) (setq vd (cons 'video e)))
	  ((string-match "^Audio" e) (setq ad (cons 'audio e)))
	  ((string-match "^Size" e) (setq sz e))
	  ((string-match "^Total time:" e) (setq tm e))
      	  (t (setq spec (cons 'spec e))))) lis)
      (setq sz (string-to-int
		(cond ((funcall fun sz)
 		       (substring sz (match-beginning 1) (match-end 1)))))
	    tm (string-to-int
		(cond ((funcall fun tm)
		       (substring tm (match-beginning 1) (match-end 1))))))
      ;; Frames is phony -- it's a file size actually.
      (delq nil
	    (list (cons 'file file)
		  (cons 'player monk-smpeg-command)
		  (cons 'track-opt nil)
;;; I know plaympeg(1) has
;;;        --seek N or -S N     Skip N bytes
;;; but it seemed not to work well for big mpg file
;;; Wed Apr 25 14:10:39 2001
;;; Seems like plaympeg(1) (from smpeg-0.4.3) now works right !!!
		  (cons 'offset-opt "-S %d") ; Need a space between opt/arg.
		  (cons 'duration-opt nil) ; but we still want this option !
		  (cons 'post-filter nil)
		  (cons 'options monk-smpeg-command-options)
		  (cons 'frames sz)
		  (cons 'time tm)
		  (cons 'cent  (/ sz 100.0))
		  (cons 'fps  (round (/ sz tm)))
		  (cons 'type 'mpg)
		  spec vd ad))))))

(defun monk-smpeg-insert-track-obj (file)
  "Insert track line at point."
  (interactive "fmpeg File:")
  (let ((obj (condition-case nil     ; sometimes errors (on downloading ?) 
		 (monk-smpeg-build-track-obj file)
	       (error nil)))
	fname frames time pt)
    (cond (obj
	   (setq fname (monk-assq-get-value 'file obj)
		 frames (monk-assq-get-value 'frames obj)
		 time (monk-assq-get-value 'time obj)
		 pt (point))
	   ;; "%-4s [%5s.00]%8d%8d %s" ; old format once was here
	   (monk-insert-track obj "mpg:"
			      (format "%5s.00" (monk-sec2cdtime time))
			      0 frames
			      (format "\"%s\""
				      (file-name-nondirectory
				       fname)) 'newline)
	   obj))))

;;
;; Commands for "-2" option of plaympeg(1)
;; Sun Oct 27 18:24:18 2002
;;

;; Should be merged into monk.el ?
(defsubst monk-replace-in-string (str regex newstr)
  "Bullet-proof version of dired-replace-in-string for command line options."
  (mapconcat
   'identity
   (delete ""
	   (dired-split "[\t ]"
			(dired-replace-in-string regex newstr str))) " "))

(defun monk-smpeg-set-command-options (arg)
  "Operate `monk-smpeg-command-options' with ARG.
Non-nil ARG forces to set `monk-smpeg-command-options' to \"-2s\",
otherwise nil.  Current value is ignored.  May not be useful."
  (interactive "P")
  (when monk-smpeg-command-options
    (setq monk-smpeg-command-options	; firstly cleanup
	  (monk-replace-in-string monk-smpeg-command-options "-2" "")))
  (when arg
    (setq monk-smpeg-command-options
	  (concat "-2"
		  (when (< 0 (length monk-smpeg-command-options)) " ")
		  monk-smpeg-command-options)))
  ;; "" arg to plaympeg(1) causes malfunction, so we should set it to nil.
  (when (string= monk-smpeg-command-options "")
    (setq monk-smpeg-command-options nil))
  (when (interactive-p)
    (message "monk-smpeg-command-options: %S" monk-smpeg-command-options)))

(defun monk-smpeg-resize (arg)
  "Select double or single size for 'mpg track according to ARG.
Restart playing if the relevant track is playing."
  (interactive "P")
  (let* ((prop (monk-get-track))	; with file validity check
	 (playing (and (monk-proc-running-p)
		       (equal monk-current-track prop)))
	 (type (monk-assq-get-value 'type (monk-get-track)))
	 monk-smpeg-command-options)	; Use temp value.
    ;; Fake `monk-smpeg-command-options' within this function only.
    (cond ((eq type 'mpg)
	   (monk-smpeg-set-command-options arg)
	   (when prop
	     (call-interactively 'monk-rebuild-track)
	     (when playing
	       (monk-play-pause 0)	; Restart from current pos.
	       )))
	  (t (error "That track is not 'mpg but '%S" type)))))

;; Should be in `monk-load-hook'
;; (define-key monk-mode-map "\eD" 'monk-smpeg-resize)

;;
;; (monk-smpeg-build-track-obj "~/Ringo/00VIDEO/tmp/Shiina Ringo -  Gibbs.mpg")
;; ((file . "/home/sn/Ringo/00VIDEO/tmp/Shiina Ringo -  Gibbs.mpg")
;;  (player . "plaympeg")
;;  (track-opt)
;;  (offset-opt)
;;  (duration-opt)
;;  (post-filter)
;;  (frames . 59619180)
;;  (time . 337.964144)
;;  (cent . 596191.8)
;;  (fps . 176407)
;;  (type . mpg)
;;  (spec . "Shiina Ringo -  Gibbs.mpg: MPEG system stream (audio/video)")
;;  (video . "Video 352x240 resolution")
;;  (audio . "Audio MPEG-1 Layer 2 224kbit/s 44100Hz stereo"))
;; 
;; (monk-smpeg-probe
;;  (expand-file-name "~/Ringo/00VIDEO/tmp/Shiina Ringo -  Gibbs.mpg"))
;; ("Shiina Ringo -  Gibbs.mpg: MPEG system stream (audio/video)"
;;  "Video 352x240 resolution"
;;  "Audio MPEG-1 Layer 2 224kbit/s 44100Hz stereo"
;;  "Size: 59619180"
;;  "Total time: 337.964144")
;;
