;;; monk.el --- For bleeding hearts,

;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc

;; monk.el --- Multimedia Objct Navigator monK (or Kluge, or what ?)
;; in-fact-it's-free software, not a GPL software.
;; Copyright 2000,2001,2002,2003 Seiichi Namba <sn@asahi-net.email.ne.jp>

;; Started: Sat Nov 18 15:39:43 2000 s.n.
;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc

(defconst monk-version "8")
(defconst monk-edition "3")

(or (fboundp 'float) ; (fboundp 'atan)
    (error
     "You should build emacs with floating point support to run MonK."))

(provide 'monk)

(defun monk-version ()
  (interactive)
  (cond ((interactive-p)
	 (message "MonK Version %s Edition %s" monk-version monk-edition))
	(t (concat monk-version "." monk-edition))))

;; For dired-split.  I know something new is in emacs-20, though.

(eval-and-compile (require 'dired)
		  (require 'dired-aux)
		  (require 'timer)	; for emacs-19
		  (require 'cl))	; for emacs-19 (for last, when...)

;;
;; MonK was a cd player at first, and this is still required unconditionally.
;;

(require 'monk-cdaudio)

;;
;; Autoloads, not so many.
;;
;; The file handler is required via `monk-track-generator-list' and
;; `monk-get-track-generator', rather explicitly.
;;

(autoload 'monk-prop-edit "monk-prop-edit" nil t)
(autoload 'monk-write-id3-info "monk-mp3" nil t)
(autoload 'monk-mp3-mouse-toggle-show-title "monk-mp3" nil t)

;; (autoload 'monk-complete-prop "monk-prop-edit" nil t) ;; no use here
;; (autoload 'monk-show-id3-genre-table "monk-mp3" nil t) ;; no use here

(autoload 'monk-watch-directory-mode "monk-watch" nil t)
(autoload 'monk-watch-check-directory "monk-watch" nil t)

;;;
;;; Some macros.  May not always be used globally in this file.
;;; Sometimes they are not useful, or just ineffective.
;;; Mon Nov 27 17:36:25 2000x
;;;

;; 19.28
(or (fboundp 'buffer-substring-no-properties)
    (fset 'buffer-substring-no-properties 'buffer-substring))

(cond ((fboundp 'buffer-live-p)
       ;; `fset' makes E19.34 to barf
       (defsubst monk-buffer-live-p (b)
	 (and b (buffer-live-p (get-buffer b)))))
      ;; E19.28 says 
      ;;  ** the function buffer-live-p is not known to be defined.
      ;; but it's okay
      (t (defsubst monk-buffer-live-p (b)
	   (condition-case nil
	       (and b (save-excursion (set-buffer (get-buffer b))))
	     (error nil)))))

;; emacs-20 sort of `with-current-buffer' in emacs-19 compatible form.
;; May not be so useful

(defmacro monk-with-current-buffer (buffer &rest body)
  "Eval the forms in BODY with BUFFER as the current buffer.
The value returned is the value of the last form in BODY."
  (` (and (monk-buffer-live-p (, buffer))
					;(get-buffer (, buffer))
					;(save-window-excursion ; required ?
	    (save-excursion
	      (set-buffer (, buffer))
	      ;; (eval (cons (quote progn) (quote (, body))))
	      (,@ body)
	      )
					;)
	    )))

;; `monk-with-monk-window-and-buffer' -- unused currently

;; (defmacro monk-with-monk-window-and-buffer (&rest body)
;;   "Eval the forms in BODY with `monk-buffer' as the current buffer.
;; See also `monk-wcb'."
;;   (` (and (monk-buffer-live-p monk-buffer)
;;           ;; (get-buffer monk-buffer)
;;           (save-window-excursion
;;             (save-excursion
;;               (select-window (get-buffer-window monk-buffer t))
;;               ;; (set-buffer monk-buffer)
;;               ;; (eval (cons (quote progn) (quote (, body))))
;;               (,@ body) )))))

;; 
;; `monk-with-monk-window-or-buffer'
;; The idea is `use-window-if-showing-or-use-buffer'.
;; Conceived to control curso from `monk-proc-sentinel'.
;;

(defmacro monk-with-monk-window-or-buffer (&rest body)
  "Eval BODY with selecting `monk-buffer''s window, \
or selecting `monk-buffer'."
  (` (let ((w (selected-window))
	   (bw))
       (cond
;;; Tried to act differently depending upon frame config, but was no use.
;;;      ((setq bw get-buffer-window monk-buffer nil)
;;;       ;; `save-window-excursion' can't move cursor if in the same frame.
;;;       ;; Use `save-restore' scheme here.
;;;       ;; (save-window-excursion
;;;	  (select-window bw)
;;;	   (,@ body)
;;;        (select-window w)
;;;       )
	;; This is good enough (current buffer will be forced to *monk*
	;; if it is displayed in another frame, though):
	((setq bw (get-buffer-window monk-buffer t))
	 (select-frame (window-frame bw))
	 (select-window bw)
	 (,@ body)
	 (select-window w)
	 )
	;; window is hidden: `save-excursion' moves cursor nicely.
	(;;(and (monk-buffer-live-p monk-buffer) (get-buffer monk-buffer))
	 (monk-buffer-live-p monk-buffer)
	 (save-excursion
	   (set-buffer monk-buffer)
	   (,@ body) ))))))

(defmacro monk-with-monk-buffer (&rest body)
  "Eval the forms in BODY with `monk-buffer' as the current buffer.
See also `monk-wcb'."
  (` (and (monk-buffer-live-p monk-buffer)
          ;; (get-buffer monk-buffer)
          (save-window-excursion        ; required ?
            (save-excursion
              (set-buffer monk-buffer)
              ;; (eval (cons (quote progn) (quote (, body))))
              (,@ body) )))))

;; Purged 'cause inconsistent, and no use.
;; (defmacro monk-with-current-window (buffer &rest body) ...)

(defmacro monk-with-monk-window (&rest body)
  "Eval forms in BODY with selecting window showing `monk-buffer'."
  (` (and (monk-buffer-live-p monk-buffer) (get-buffer monk-buffer)
	  (get-buffer-window monk-buffer t)
	  (save-window-excursion	; surely required.
	    ;; `get-buffer-window' is primitive.  Do it over again and again.
	    (select-window (get-buffer-window monk-buffer t))
	    ;; (eval (cons (quote progn) (quote (, body))))
	    (,@ body) ))))

;; `monk-only-with-monk-modes' -- The police action.
;; Mainly to inhibit M-x execution of some `direct-edit' commands
;; in wrong buffers.

(defmacro monk-only-with-monk-modes (&rest body)
  "Do BODY when `major-mode' is `monk-mode' or `monk-edit-mode'."
  (` (cond ((or (eq major-mode 'monk-mode)
		(eq major-mode 'monk-edit-mode)
		(eq major-mode 'monk-ripper-mode))
	    ;; (eval (cons (quote progn) (quote (, body))))
	    (,@ body))
	   ((interactive-p) (error "Yo in wrong major-mode, brother"))
	   (t nil))))

;;
;; Coding system macro
;; Mon Dec 25 19:14:18 2000
;;
;; Coding system handling commands
;; Mon Apr  2 14:24:31 2001
;;
;; monk-with-coding-system() macro -- emacs-19.28/19.34/20.x compatible
;;
 
(defvar monk-coding-system nil
 "*The coding-system to be bound to `default-process-coding-system',
`file-name-coding-system' and `pathname-coding-system' in file /
process accessing phase in MonK.  If `monk-file-name-coding-system' is
nil, `monk-coding-system' is bound to `file-name-coding-system' as well.

Leaving this `monk-coding-system' as being nil, set up the standard
coding-systems is one idea.

See also `monk-file-name-coding-system' which has priority to this
variable on `file-name-coding-system' and `pathname-coding-system'.")

(defvar monk-file-name-coding-system nil
  "*The coding-system to bound to `file-name-coding-system' \(makes
sense in emacs-20.x\) `pathname-coding-system' \(makes sense in
mule-2.3\) in file / process accessing phase in MonK.  If non-nil, the
value of this variable is set to `file-name-coding-system' prior to
`monk-coding-system'. See also `monk-coding-system'.")

(defmacro monk-with-coding-system (&rest body)
  "Execute BODY with coding-system fiddle.
See `monk-coding-system' and `monk-file-name-coding-system'."
  (` (let* ((default-process-coding-system
	      (if monk-coding-system 
		  (cons monk-coding-system monk-coding-system)
		default-process-coding-system))
	    (file-name-coding-system
	     (or monk-file-name-coding-system monk-coding-system
		 ;; mule-2.3@19.28 warns on compile. 19.34 does not.
		 ;; Why ?  Byte-compile method difference ?
		 (and (boundp 'file-name-coding-system)
		      file-name-coding-system)
		 ))
	    (pathname-coding-system file-name-coding-system))
       ;; (eval (cons (quote progn) (quote (, body))))
       (,@ body))))

(defun monk-set-file-name-coding-system (cs)
  "Set coding system CS to `monk-file-name-coding-system'."
  (interactive
   (list
    (read-coding-system
     (format
      "monk-file-name-coding-system ? (current:%s, RET sets nil): "
      monk-file-name-coding-system))))
  (check-coding-system cs)
  ;; (when cs (setq monk-file-name-coding-system cs))
  (setq monk-file-name-coding-system cs))

(defun monk-set-coding-system (cs)
  "Set coding system CS to `monk-coding-system'."
  (interactive
   (list
    (read-coding-system
     (format
      "monk-coding-system ? (current:%s, RET sets nil): "
      monk-coding-system))))
  (check-coding-system cs)
  ;; (when cs (setq monk-coding-system cs))
  (setq monk-coding-system cs))

(defvar monk-describe-coding-system-use-buffer nil
  "*Use an independent buffer or not for `monk-describe-coding-system' output.
Make sense only when emacs supports variable `message-truncate-lines'.
Defaults nil.")

(defun monk-describe-coding-system (arg)
  "Show `default-process-coding-system', `file-name-coding-system',
`monk-coding-system', and `monk-file-name-coding-system'.
If `path-name-coding-system' is bound, it is referred instead of
`file-name-coding-system'.

Non-nil ARG toggles the value of `monk-describe-coding-system-use-buffer'
temporally while executing this command."
  (interactive "P")
  (let ((monk-describe-coding-system-use-buffer
	 (cond (arg
		(null monk-describe-coding-system-use-buffer))
	       (t monk-describe-coding-system-use-buffer)))
	fcs fcsn msg buf)
    (cond ((boundp 'file-name-coding-system)
	   (setq fcs file-name-coding-system fcsn "file-name-coding-system"))
	  ((boundp 'path-name-coding-system)
	   (setq fcs path-name-coding-system fcsn "path-name-coding-system"))
	  (t (setq fcsn nil)))
    (setq msg (concat
	       (format "%29s: %S\n" "default-process-coding-system"
		       default-process-coding-system )
	       (when fcsn (format "%29s: %S\n" fcsn fcs))
	       (format "%29s: %S\n" "monk-coding-system" monk-coding-system)
	       (format "%29s: %S" "monk-file-name-coding-system"
		       monk-file-name-coding-system)))
    (cond ((and (boundp 'message-truncate-lines) ; This is emacs-21,
		(null message-truncate-lines) ; and echo area can expand.
		(null monk-describe-coding-system-use-buffer))
	   (message msg))
	  (t (monk-with-current-buffer
	      (setq buf (get-buffer-create "*monk*coding-system*"))
	      (erase-buffer)
	      (insert msg "\n"))
	     (monk-show-buffer buf)))))

;;
;; Slash handling in file name.
;; duplicated // is works as libc (?) functions
;; in emacs-20, but not in emacs-19:
;;

(defsubst monk-hack-trailing-/ (dir)
  "Remove trailing slashes at the end of string DIR."
  (substring dir 0 (string-match "/+$" dir)))

(defsubst monk-put-trailing-/ (dir)
  "Put one trailing slash at the end of string DIR."
  (concat (monk-hack-trailing-/ dir) "/"))

;;
;; Process detection, and jump to top / bottom, map-over-tracks
;;

(defsubst monk-proc-running-p ()
  "Check if `monk' process is running. Checks current buffer only."
  ;; (let ((proc (get-buffer-process (current-buffer)))) ; works, but,
  ;;   (cond (proc (eq (process-status proc) 'run))))
  (eq 'run (condition-case nil		; this is good enough.
	       (process-status nil) (error nil))))

(defsubst monk-jump-to-top ()
  "Goto the track at the top of current buffer."
  (goto-char (point-min)))

(defsubst monk-jump-to-bottom ()
  "Goto the track at the bottom of current buffer."
  (goto-char (point-max))
  (if (bolp) (forward-line -1))		; else last \n is lacked.
  (point))

(defmacro monk-walk-over-tracks (&rest body)
  "Do BODY on each track \(or just a line\) of current buffer."
  (` (save-excursion
       (goto-char (point-min))
       (while (not (eobp))
	 (,@ body)
	 (forward-line 1)))))

(defmacro monk-do-nlines (count &rest body)
  "Do BODY, moving COUNT lines forward.
Object passed as 1st arg COUNT will be modified: It will be 0 if successful,
or number of left move if hit bottom before doing COUNT move."
  ;; Result is the same even if count is not braced as (, count), but
  ;; watch out !, 1st arg is modified (perhaps `arg' in most caller) via
  ;; COUNT, even if the name is different -- this is a macro.
  ;; Try (let ((x 2))
  ;;        (message "x1:%s" x) (monk-do-nlines x (pwd)) (message "x2:%s" x))
  (` (catch 'bottom
       (while (> (, count) 0)
	 (,@ body)
	 (when (or (not (= 0 (forward-line 1)))
		   (eobp)) (throw 'bottom nil))
	 (setq (, count) (1- (, count)))))))

;;
;; Inlines for getting and putting `track' property, inserting track, etc.
;;
;; In MonK, a track property is put as `text-property'.
;; The 'track property location range is (BOL+1) to (BOL+4) on a track line.
;; (BOL+0) is reserved for forthcoming `monk-mark-track' etc.
;;

(defsubst monk-assq-get-value (key proplis)
  "Return value corresponding to KEY in alist PROPLIS."
  (cond ((listp proplis)
	 (cdr (assq key proplis)))))

;; `monk-prop-set-val' -- fastest alist edit function

(defun monk-prop-set-val (alist newelem)
  "Return reconstructed alist out ouf ALIST and NEWELEM.
`delq' is done by `assq'ing the key in NEWELEM in ALIST,
and return `cons' of NEWELEM and the `delq'ed ALIST -- element
order is not maintained, but normally it's okay for alist.
Side effect on ALIST should not be expected."
  (cons newelem (delq (assq (car newelem) alist) alist)))

(defsubst monk-get-track-property ()
  "Get text-property 'track and return it."
  (save-excursion (forward-line 0)
		  (cond ((eobp) nil)	; Can't make (1+ (point))
			((and (bolp) (eolp)) nil) ; Empty line detected
			(t (get-text-property (+ 2 (point)) 'track)))))

(defsubst monk-put-track-property (prop)
  "`put-text-property' PROP \(should be 'track in MonK\) at bol."
  (save-excursion
    (forward-line 0)
    (put-text-property (+ 2 (point)) (+ 5 (point)) 'track prop)))

(defsubst monk-get-track (&optional raw)
  "Get text-property 'track at bol.
If optional arg RAW is supplied returns property as it is without any
validation test of the prop derived from `monk-get-track-property'."
  (let ((prop (monk-get-track-property)))
    (cond ((null prop) nil)
	  (raw prop)			; return raw data without further test
	  ;; ((integerp prop) prop)     ; integer track deprecated.
	  ((listp prop)			; check validity of 'file
	   (monk-with-coding-system
	    (let* ((file (monk-assq-get-value 'file prop))
		   (access (and file (file-readable-p file)))
		   (size (and access (nth 7 (file-attributes file)))))
	      (cond ((null file) prop)	; Allow nil as 'file.
		    ((null access) nil)
		    ;; non-nil size may not be checked: but it's safe.
		    ((and size (<= size 0)) nil)
		    (t prop))))))))

(defsubst monk-insert-track (trk tag time start frames name &optional newline)
  "Generic inline to insert track property TRK.
TAG, TIME, START, FRAMES, and track NAME should be supplied as well, because
sometimes a parameter like TIME must be fiddled by the caller.

If START and FRAMES are number, they are printed as integer.  All
the others should be a string.  A newline is inserted if optional 2nd arg
NEWLINE is non-nil."
  (let (pt)
    (when (numberp start) (setq start (truncate start)))
    (when (numberp frames) (setq frames (truncate frames)))    
    (insert
     ;; (format "%-5s [%8s]%8d%8d " tag time start frames)
     (format "  %-4s [%8s]%9s%9s " tag time start frames))
    (setq pt (point))
    (insert (format "%s" name))
    (put-text-property pt (point) 'mouse-face 'highlight)
    (monk-put-track-property trk)
    (cond (newline (insert "\n")))))

;; More universal version of `monk-insert-track'.
;; You just feed TRACK (and NEWLINE).

(defsubst monk-insert-track-versatile (track &optional newline)
  "Universal version of `monk-insert-track'. Inserts TRACK.
Optional 2nd arg NEWLINE is passed as it is to `monk-insert-track'."
  (let ((type (monk-assq-get-value 'type track))
	(title (monk-assq-get-value 'title track))
	(file (monk-assq-get-value 'file track))
	(artist (monk-assq-get-value 'artist track))
	(brate (monk-assq-get-value 'bitrate track)))
    (monk-insert-track
     track
     (format "%s:" type)
     (format "%5s.00"
	     (monk-sec2cdtime
	      (monk-assq-get-value 'time track)))
     (cond
      ((eq type 'mp3)
       (if brate (format "%db/s" brate) "unknown"))
      ;; `generic' cdaudio track has this.
      (t (or (monk-assq-get-value 'track-offset track) 0)))
     (monk-assq-get-value 'frames track)
     (or
      ;; mp3 `title - artist' track name hanling codes are
      (and (eq type 'mp3)
	   (or (cond
		((and monk-mp3-show-title title
		      (null (string-equal title "")))
		 (format "`%s%s'" title
			 (cond (artist (format " - %s" artist))
			       (t "")))))
	       (format "\"%s\"" (file-name-nondirectory file))))
      (and file (format "\"%s\"" (file-name-nondirectory file)))
      (and (or (eq type 'rip)
	       (eq type 'cda)) (format "\"%s\"" title))
      (format "\"%s\"" (monk-assq-get-value 'player track))
      "Broken Track ?")
     newline)))

(defsubst monk-set-process-marker ()
  "Set `monk-process-marker' to eol of this current track."
  (or monk-process-marker
      (setq monk-process-marker (make-marker)))
  (move-marker monk-process-marker
	       (save-excursion (end-of-line) (point))
	       (current-buffer)))

(defsubst monk-on-process-marker-track-p ()
  "Check if `monk-process-marker' and eol of this current track equals."
  (and (markerp monk-process-marker)
       (eq (current-buffer) (marker-buffer monk-process-marker))
       (= (save-excursion (end-of-line) (point))
	  (marker-position monk-process-marker))))

(defsubst monk-compare-tracks-file (t1 t2)
  "Compare property `file' of tracks T1 and T2."
  (let ((f1 (monk-assq-get-value 'file t1))
	(f2 (monk-assq-get-value 'file t2)))
    (and f1 (string-equal f1 f2))))

(defsubst monk-compare-tracks-track-number (t1 t2)
  "Compare property `track-number' of tracks T1 and T2."
  (let ((f1 (monk-assq-get-value 'track-number t1))
	(f2 (monk-assq-get-value 'track-number t2)))
    (and f1 (= f1 f2))))

(defsubst monk-on-playing-track-p (t1 t2 &optional nomarkerchk)
  "Compare two tracks T1 and T2 loosely. Criterion are:
`monk-process-marker' equals to eol, and, `monk-compare-tracks-file'
or `monk-compare-tracks-track-number' returns non-nil.
If optional 3rd arg NOMARKERCHK is non-nil, the process marker is skipped.
Used to compare current track and `monk-current-track'."
  (and (or nomarkerchk (monk-on-process-marker-track-p))
;;;       (or nomarkerchk
;;;	   (and (markerp monk-process-marker)
;;;		(eq (current-buffer) (marker-buffer monk-process-marker))
;;;		(= (save-excursion (forward-line 0) (point))
;;;		   (marker-position monk-process-marker))))
       (or (monk-compare-tracks-file t1 t2)
	   (monk-compare-tracks-track-number t1 t2))))

;; An APHORISM on buffer and windows:
;; You can get buffer position even if it is not `current',
;; but you can't move cursor unless you select the window.

(defun monk-get-this-track ()
  ;; Now integer `track' does not exist, but keep this as a function.
  "Get track object \(number, or property list\) on which point sits.
Return nil if failed.  See also `monk-insert-track-lines'.

Use current buffer/point context.
If text-property is available for current line, the text-property is used.
If not, searches truck number regex."
  ;; Now first 4 char only hold text-property
  (or (monk-get-track)		; it has also save-ex/forward-line 0
      ;; This failsafe is for really rare case
      ;; in which user destroyed the int/list `track' prop
      (and (save-excursion
	     (forward-line 0)
	     (looking-at "^.T\\([0-9]+\\):"))
	   (string-to-int
	    (buffer-substring
	     (match-beginning 1) (match-end 1))))))

;;;
;;; Player engine
;;;

(defconst monk-info-buffer " *monk*info*" "Internal use only.")

(defconst monk-junk-buffer " *monk*junk*" "Internal use only.")

(defconst monk-buffer-name
  "*monk*" "An `monk' user interface buffer.")

(defconst monk-process-display-buffer-name
  "*monk*display*" "An `monk' user interface buffer.")

;; Going to be deprecated.
(defvar monk-buffer
  nil "Internal use only.  May be deprecated in future versions.
Global.  Points current `monk' buffer at certain timings, but
it may not always point the buffer you are expecting.")

;; `monk-generate-command' -- Command line generator for a MonK track
;; Currently cdaudio track and the other tracks are processed differently.

(defun monk-generate-command (track offset &optional duration)
  "Setup `playing' command line to be passed to `start-process-shell-command'."
  (cond
;;; Not yet. Load monk-ripper.el by hand.
;;;   ((and (listp track)			; ripper prop
;;;	 (eq 'rip (monk-assq-get-value 'type track)))
;;;    (monk-generate-command-ripper track offset))
   ((listp track)			; may return nil 
    (monk-generate-command-generic track offset duration))
   (t (error "BUG:Track data %s broken." track))))

;; Spliting required, to propagate "-S 20" as "-S" "20" as unix ARGVs.

(defsubst monk-command-option-format (format value)
  "Return a list of option string and VALUE.
If FORMAT contains space or tab, the FORMAT string is split there."
  (if (listp format) (setq format (car format)))
  (if format (dired-split "[ \t]" (format format value))))

(defun monk-generate-command-generic (track offset &optional duration)
  "Generates command line as list out of object TRACK.
If a certain property of TRACK was nil, the corresponding command line argument
would not be generated. See `monk-generate-command' for OFFSET and DURATION.

Use `monk-prop-edit' to examine the MonK's `track' format."
  (let ((ofsopt (monk-assq-get-value 'offset-opt track))
	(duropt (monk-assq-get-value 'duration-opt track))

	;; Forthcomming cdaudio generic track support
	(trkopt (monk-assq-get-value 'track-opt track))
	;; cdaudio track must have `track-number' property.
	(track-number (monk-assq-get-value 'track-number track))

	(player (monk-assq-get-value 'player track))
	(options (monk-assq-get-value 'options track))
	(fname (monk-assq-get-value 'file track))
	(postfl (monk-assq-get-value 'post-filter track))
	(size -1) ofs dur command access file)

    (when (and track-number (null trkopt))
      (error "Track must have `track-opt' property  track: %s" track))

    (setq ofs (and ofsopt (monk-command-option-format ofsopt offset))
	  dur (cond ((and duropt duration)
		     (cond ((= duration 0) nil)
			   (t (monk-command-option-format duropt duration)))))
	  command (and player (dired-split "[ \t]" player))
	  trkopt  (and track-number
		       (monk-command-option-format trkopt track-number))

	  options (cond
		   ((and (listp options) ; (options "-2 --noaudio") case
			 (= 1 (length options))
			 (string-match " " (car options)))
		    (dired-split "[ \t]" (car options)))
		   ((stringp options) ; (options . "-2 --noaudio") case
		    (dired-split "[ \t]" options))
		   (t options))	      ; simple list (options "-2" "-options")

	  access (and fname (file-readable-p fname))
	  size (and access (nth 7 (file-attributes fname)))
	  ;; If postfilter is non-nil, `start-process-shell-command' will be
	  ;; used, so no `dired-split' here.
	  postfl (and postfl (list "-" "|" postfl))
	  file (and fname access
		    (cond
		     ;; Must use shell for post filter, so we should quote
		     ;; meta-characters.
		     (postfl
		       (list
			(substring  ; Cheap. Barfs if `dired' changed.
			 (dired-shell-stuff-it
			  "" (list fname) nil) 1)))
		     (t (list fname))))) ; must not split fname.
    (cond (;;(and fname (= size -1)) ;; equiv.
	   (and fname (null access))
	   (message "MonK can't read: %s" track)
	   nil)
	  ((and fname (or (null size) (<= size 0)))
	   (message "MonK can't play 'cause filesize's zero: %s" track)
	   nil)
	  ;; At least just `command' is required
	  ((and command (or file
			    (null fname))) ; allow command only case
	   (append command options trkopt ofs dur file postfl))
	  (t (message
	      "Unknown error, maybe broken track data: %s" track)
	     nil))))

;;
;; monk-play-start() -- the Engine Core, and it's work areas.
;; See also `make-variable-buffer-local' section at bottom.
;;

(defconst monk-cdaudio-fps 75.0 "Frame per second constant for cdaudio.")

(defvar monk-fps 0
  "Frame per second of the track currently played. Internal use only.")

;; Deprecated Wed Jan  3 14:28:34 2001
;; (defvar monk-current-process nil "Internal use only. Player process.")

(defvar monk-process-started-buffers nil
  "Contains process running buffers.
The buffer object will be used, eg., to set current-buffer of
`monk-timer-disp-handler'.  Set by `monk-start-track-timer' and reset
by `monk-cancel-track-timer'.  `monk-timer-disp-handler' should not
use `monk-buffer', because the buffer may be temporally changed eg.,
by `monk-find-file'.")

(defvar monk-process-marker nil
  "Marker indicating currently played track. Internal use only.")
(make-variable-buffer-local 'monk-process-marker) ; New in 5.11
(setq-default monk-process-marker nil)

(defvar monk-current-track nil "Current playing track.
Set by `monk-play-take-control'.")

(defvar monk-current-track-type nil "The type of current playing track.
Set by `monk-play-take-control'.")

(defvar monk-current-track-length nil "Length of current playing track.
Set by `monk-play-take-control'.")
;; Now none refer to this:
;; (defvar monk-current-start nil "Start position in track in sector offset.")

(defvar monk-current-start-sec nil
  "Start position in track in second format as number \(float maybe\).
This variable is referred from `monk-timer-disp-handler' to get playing time.
Set by `monk-play-take-control'.")

(defvar monk-current-offset nil
  "Absolute playing position in sector offset.
Updated by `monk-proc-filter', or other timer handler.")

(defvar monk-current-cent nil
  "Number of sector corresponding the 1% of current track
Set by `monk-play-take-control'.")

;; Track/loop name, to reduce computing in `monk-mode-line-update'

(defvar monk-current-track-name ""
  "Track name string used in `monk-mode-line-update'.")

(defvar monk-current-loop-name ""
  "Loop-data name string used in `monk-mode-line-update'.")

(defvar monk-wait-proc-to-die .1
  "Delay to wait proc to die. Unit: second. Defaults .1 second.
Try increment if your player commands fail to restart. See `monk-stop'.")

(defsubst monk-play-generate-loop-name (track)
  (cond
   (monk-loop-data
    (let ((data (cond (;; (monk-one-track-loop-p)
		       (and (nth 1 monk-loop-data)
			    (atom (nth 1 monk-loop-data)))
		       monk-loop-data)
		      (t (car monk-loop-queue)))))
      (format "%s " (cons (monk-assq-get-value 'type track)
			  (cdr data)))))
   (t "")))

(defsubst monk-play-generate-track-name (track)
  ;; This doesn't support id3 title yet.
  (or 
   (and (monk-assq-get-value 'file track)
	(file-name-nondirectory (monk-assq-get-value 'file track)))
   (monk-assq-get-value 'track-number track)
   (monk-assq-get-value 'player track)
   "?unknown?"))			; will happen ?

(defsubst monk-play-take-control (track offset process disp)
  "Setup track control parameters.  Most of them are buffer local."
  (set-process-filter process (cond (disp (function monk-proc-filter2))
				    (t (function monk-proc-filter))))
;;;  (or monk-process-marker
;;;      (setq monk-process-marker (make-marker)))
;;;   (move-marker monk-process-marker (point) (current-buffer))
  (monk-set-process-marker)

  (set-process-sentinel process (function monk-proc-sentinel))
  ;; Can't start track-timer till these parameters are setup.
  (setq
   monk-fps (monk-assq-get-value 'fps track)
   monk-current-track track		; WRONG ?
   monk-current-track-type (monk-assq-get-value 'type track)

   monk-current-start-sec (/ offset monk-fps)
   monk-current-track-length
   ;;(or (monk-assq-get-value 'frames track) (monk-assq-get-value
   ;;                                          'length track)
   (monk-assq-get-value 'frames track)
   ;; cent is provided in the data
   ;; monk-current-cent (/ monk-current-track-length 100.0)
   monk-current-cent (monk-assq-get-value 'cent track)
   monk-current-loop-name (monk-play-generate-loop-name track)
   monk-current-track-name (monk-play-generate-track-name track)
  ;; The track-timer needs the parameters above to be setup.
  ;; Now track-timer needs `monk-process-started-buffers' as well.
   monk-process-started-buffers (cons (current-buffer)
				      monk-process-started-buffers))

  (monk-start-track-timer)

  ;; enFacing the playing track
  (and monk-use-face (monk-face-current-track))

  ;; For track(s ?) having 'display property.
  (when disp (monk-display-buffer monk-process-display-buffer-name)))

(defun monk-proc-filter2 (p m)
  "Output process P's  message M to `monk-process-display-buffer-name'"
  ;; Discrete version of `monk-with-monk-window-or-buffer'
  (let* ((b (get-buffer monk-process-display-buffer-name))
	 (w (and b (get-buffer-window b t)))
	 (fun (lambda ()
		(save-excursion
		  ;; `process-mark' is not what we use here.
		  (goto-char (point-max))
		  (insert m))
		  (goto-char (point-max)) )))
    (cond (w (save-window-excursion (select-window w)
				    (funcall fun)))
	  (b (save-excursion (set-buffer b)
			     (funcall fun))))))

(defsubst monk-display-buffer (buf)
  "Simplified version of `display-buffer."
  (let ((x (selected-window)))
    (pop-to-buffer buf) (select-window x)))

(defsubst monk-play-get-default-directory (track)
  "Inline for `monk-play-start'.
Return `default-directory', or directory specified by TRACK by `cd' property.
If `cd' is a directory name string and the directory is accessible,
the `cd' value is retruned.  Else if `cd' is non-nil \(like t\),
the `file-name-directory' part is returned if the directory is accessible.
Else current `default-directory' value is returned."
  (let* ((cd (monk-assq-get-value 'cd track))
	 (file (monk-assq-get-value 'file track))
	 (dir (and file (file-name-directory file))))
    (cond ((and cd (stringp cd) (file-accessible-directory-p cd))
	   (monk-put-trailing-/ cd))
	  ;; non-string 'cd case, `file-accessible-directory-p' may be overkill
	  ((and cd file dir (file-accessible-directory-p dir))
	   (monk-put-trailing-/ dir))
	  (t default-directory))))

(defun monk-play-start (track offset &optional duration)
  "Start player process passing options: TRACK OFFSET and DURATION.
All argument must be integer, and
OFFSET and DURATION must be in `sectors'\(aka. frame\) here."
  ;;(progn
  (monk-with-coding-system
   (let* ((disp (monk-assq-get-value 'display track))
	  (buf (current-buffer))
	  (default-directory (monk-play-get-default-directory track))
	  (postfl (monk-assq-get-value 'post-filter track))
	  (command (monk-generate-command track offset duration))
	  proc)
     (cond
      (command
       (cond
	(disp
	 (save-excursion
	   (set-buffer (get-buffer-create monk-process-display-buffer-name))
	   (erase-buffer))))

       (setq monk-buffer buf)

       (setq proc (apply
		   (cond (postfl 'start-process-shell-command)
			 (t 'start-process))
		   "monk-process" buf command))
       (monk-play-take-control track offset proc disp)
       proc)))))

;;
;; `monk-stop' --- Used interactively and non-interactively
;;
;; Old `monk-stop' tries to kill all monk process unconditionally,
;; even if already running process does not conflict to the forthcoming
;; process. The `kill-all' code calling `monk-kill-all-process' is currently
;; commented out, and handles killing process in the current buffer context
;; only.  See `monk-play-track'.
;;
;; (current-buffer) must be available. Eg. the sentinel must assure.
;;

(defsubst monk-process-started-buffers-cleanup (buf)
  "`delq' killed buffers in `monk-process-started-buffers'.
Executed per `monk-stop' call."
  (cond ((markerp monk-process-marker)	; maybe still nil
	 (move-marker monk-process-marker nil)))
  ;; perhaps `delq' can't delete #<killed buffer>, but it's okay,
  (setq
   monk-process-started-buffers (delq buf monk-process-started-buffers))
  ;; since we do this:
  (setq monk-process-started-buffers
	(delq
	 nil (mapcar (lambda (b) (cond ((monk-buffer-live-p b) b)))
		     monk-process-started-buffers))))

;; Unused maybe.
(defsubst monk-kill-all-process ()
  "Kill all running MonK processes.
Scans buffers in `monk-process-started-buffers'."
  (let (proc)
    (cond
     (monk-process-started-buffers
      ;; Current version stops on finding the first process.
      (mapcar
       (lambda (b)
	 (cond ((monk-buffer-live-p b)
		(save-excursion 
		  (set-buffer b)
		  (cond
		   ((and (setq proc (get-buffer-process (current-buffer)))
			 (eq (process-status proc) 'run))
		    (monk-stop)))))))
       monk-process-started-buffers)))))

(defvar monk-type-group-alist
  '((mp3 . dsp)
    (mid . dsp)
    ;; (mid . sequencer)
    ;; CD audio generic track is not yet implemented, but,
    (cda . dsp)
    (mpg . dsp)
    (wav . dsp)
    (ogg . dsp))
  "*Alist of \(TRACKTYPE . DEVICEGROUP\).
Mutually exclusive TRACKTYPEs should belong to the same DEVICEGROUP.
Eg., `mp3' and `mpg' tracks use /dev/dsp, and these should belong to
group `dsp'.")

(defsubst monk-walk-over-processes (fun)
  "Call FUN each process listed by `process-list'.
FUN must accept one arg PROCESS.
`monk-process-started-buffers' is not used here. Gee, I just can't trust it."
  (delq
   () (mapcar (lambda (p)
		(if (string-match "^monk-process" (process-name p))
		    (funcall fun p))) (process-list))))

(defun monk-stop-process-mutex (track &optional keeploop)
  "Stop processes mutually exclusive to TRACK.
Called before playin the TRACK."
  (when track
    (let* ((type (monk-assq-get-value 'type track))
	   (devgroup (monk-assq-get-value type monk-type-group-alist))
	   ty dg)
      (mapcar
       (lambda (b)
	 (save-excursion
	   (set-buffer b)
	   (setq ty (monk-assq-get-value 'type monk-current-track)
		 dg (monk-assq-get-value ty monk-type-group-alist))
	   ;; In case both-nil, don't have to stop.
	   (cond ((or (and ty (eq ty type))
		      (and dg (eq dg devgroup)))
		  (monk-stop keeploop)))))
       (monk-walk-over-processes (lambda (p)
				   (when (eq 'run (process-status p))
				     (process-buffer p))))))))

;; CUrrently `monk-stop-cleanup' does not call `monk-kill-all-process':
;; See comments on `monk-play-track'.

(defsubst monk-stop-cleanup ()
  "Deface the track, cancel track timer, and \
`monk-process-started-buffers-cleanup'
The jobs are all done in the current-buffer context."
  (let (trkpos)
    (when monk-face-overlay
      (monk-unface nil) (sit-for 0))
    ;; Cleanup for `monk-mode-line-update-mpg123' if it's the timer handler
    (when (and (eq monk-track-timer-handler 'monk-mode-line-update-mpg123)
	       (null (eq this-command 'monk-play-pause))
	       monk-process-marker
	       (setq trkpos (marker-position monk-process-marker)))
      (monk-reset-track-time trkpos))
    (monk-cancel-track-timer (current-buffer))
    ;; Without this last monk-mode-line-update, sometimes cleared (mp3
    ;; 0 0) sort of loop name is left spuriously on hitting `s' (monk-stop).
    (when (and (eq monk-track-timer-handler 'monk-mode-line-update)
	       (numberp monk-current-offset)
	       (numberp monk-fps))
      (monk-mode-line-update (/ monk-current-offset monk-fps)))
    (monk-process-started-buffers-cleanup (current-buffer))))

(defun monk-stop (&optional keeploop)
  "Stop playing CD audio.
This interactive command is only the entry point to stop playing.
Non-nil optional arg KEEPLOOP stops just the player process, and leave
the `monk-loop-data'.
THE PROGRAMMER SHOULD USE ONLY THIS TO STOP PLAYING A TRACK."
  (interactive "P")
  ;; The interactive version `monk-loop-clear' may possibly restarts
  ;; playing, and that is not what we want here. Call the core part only:
  (or keeploop (monk-loop-clear0))
  (cond
   ((monk-proc-running-p)
    (let ((proc (get-buffer-process (current-buffer))))
      (delete-process proc)
      (sleep-for monk-wait-proc-to-die)
      ;; Normally this must not loop ('signal must be the status):
      (while (eq 'run (process-status proc)) ()) ; failsafe
      ;; Should not be reset this.  Used to resume playing.
      ;; (setq monk-current-track nil)
	
      ;; If no process in this buffer, try other buffers.
      ;; (monk-process-started-buffers
      ;;  (monk-kill-all-process))	; recurses to this `monk-stop'
      )))
  (monk-stop-cleanup)
  nil)

;; 
;; The `mode-line' printer
;; Getting `monk-current-offset' out of `second' data from track timer
;; (To get finer resolution)
;;

(defvar monk-emulate-loop-track-type '(mpg mid ogg)
  "*Track types for which loop end should be emulated using the track timer.")

(defsubst monk-emulate-loop ()
  "Emulate loop end for certain track type in `monk-emulate-loop-track-type'.
Called by `monk-mode-line-update'."
  (cond
   ((and
     (get-buffer-process (current-buffer))
     (monk-scratching-p)
     (memq monk-current-track-type monk-emulate-loop-track-type)
     (>= monk-current-offset
	 (+ (nth 1 monk-loop-data) (nth 2 monk-loop-data))))
    (interrupt-process (get-buffer-process (current-buffer))))))

(defsubst monk-mode-line-print (mess)
  "Modify mode-line of current buffer with MESS."
  (setq mode-line-buffer-identification (list mess))
  (force-mode-line-update nil))

(defun monk-mode-line-update (second)
  "Update mode-line of current buffer according to SECOND.
Normally called through `monk-track-timer-handler'"

  ;; Record `monk-current-offset' as float. Optimized:
  ;; (setq monk-current-offset (monk-sec2sector second))
  (setq monk-current-offset (* second monk-fps))

  ;; Emulate loop-end (mainly for plaympg)
  (monk-emulate-loop)

  (monk-mode-line-print
   ;; Hairy ?  But, mode-line handler parses "%" once again.
   ;;(format "%02d:%02d (%03d%%%%:%06d) %s%%b Track:%s"

   ;; (format "%02d:%02d (%02d%%%%:%06d) %s%%b Track:%s" ; this is nice too.
   (format "%%b %02d:%02d (%02d%%%%:%06d) %sTrack:%s"
	   (/ second 60.0) (mod second 60) ; The 00:00 data

	   ;; `%' display.
	   (/ monk-current-offset monk-current-cent)
	   monk-current-offset

	   monk-current-loop-name	; The loop ctrl data name
	   ;; Caveat:
	   ;; "Track:%s" must be after *any* "%" including part, otherwise
	   ;; emacs-20.7's digit calculation fails when "Track:%s" includes
	   ;; Japanese (or any other) multi-byte character.
	   monk-current-track-name	; Track #, or name
	   )))

(defsubst monk-replace-track-time (trkpos second)
  "At buffer position TRKPOS, insert SECOND in MM:SS.ss format."
  (goto-char trkpos)
  (beginning-of-line)
  (forward-char 8)
  (delete-char 8)
  (insert (format "%02d:%02d.00" (/ second 60.0) (mod second 60))))

(defsubst monk-reset-track-time (trkpos)
  "Reset the track timing display at TRKPOS.
Called by `monk-stop-cleanup', and calls `monk-replace-track-time'."
  (if monk-current-track
      (let (buffer-read-only)
	(save-excursion
	  (monk-replace-track-time
	   trkpos (monk-assq-get-value 'time monk-current-track))))))

(defun monk-mode-line-update-mpg123 (second)
  "Alternative version of `monk-mode-line-print', mpg123.el alike.
Actually prints track timing in MonK buffer instead of the mode-line.
The emulation is Not perfect since this prints percentage and
current-frame on the mode-line."
  (setq monk-current-offset (* second monk-fps))
  ;; Emulate loop-end (mainly for plaympg)
  (monk-emulate-loop)
  (let (buffer-read-only trkpos)
    (save-excursion
      (cond ((and monk-process-marker
		  (setq trkpos (marker-position monk-process-marker)))
	     (monk-replace-track-time trkpos second)
	     (monk-mode-line-print
	      (format "%%b (%02d%%%%:%06d) %sTrack:%s"
		      ;; `%' display.
		      (/ monk-current-offset monk-current-cent)
		      monk-current-offset

		      monk-current-loop-name ; The loop ctrl data name
		      ;; Caveat:
		;; "Track:%s" must be after *any* "%" including part !
		      monk-current-track-name ; Track #, or name
		      ))
	     )))))

;;
;; Process Filter and Sentinel --- Control loop, or advance to next track
;;

(defun monk-proc-filter (proc msg)
  "Currently nop, and dump away all process output."
  ;; (setq yyy (current-buffer))
  t)

;; Sun Dec 24 17:53:14 2000
;; Wed Jan  3 20:25:16 2001
;; Cleaned up version of `monk-proc-sentinel'

(defvar monk-proc-continue-on-failure nil
  "*Defaults nil, which makes MonK stop on failure of player command.
Safe to keep this to nil.")

(defun monk-proc-abnormal-exit-handler (proc)
  "Called by `monk-proc-sentinel' when process PROC is \"exited abnormally\"."
  (ding)
  (monk-with-current-buffer
   (process-buffer proc)
   (message
    "Player faild (device may be busy or broken file) for track %s"
    (format "%s"
	    (or
	     (monk-assq-get-value 'file monk-current-track)
	     (monk-assq-get-value 'title monk-current-track)))
    (monk-stop)))			; deface, timer should stop
  monk-proc-continue-on-failure)	; if non-nil, try to advance.

;; Wed Jun 19 14:11:23 2002 New
(defvar monk-proc-track-interval nil
  "Interval in seconds between tracks inserted when advaincing to next track.
No effect on manual operation such as `monk-play-prev', `monk-play-next' etc.
\(normally bound to C-p, C-n etc.\). Should be nil if no interval is required.
Might be useful to set this variable to 2, when you record tracks in the
*monk* buffer to your analog recorder.")

(defun monk-proc-sentinel (proc stat)
  "Sentinel for the monk process accepting PROC and STAT."
  (cond
;;; This part is conceived to handle "kill -9|15" from outer shell,
;;; but causes troubles:
;;;   (1) Hitting <mode-line mouse-2> while one-track-loop clears the loop
;;;       and it's annoying.
;;;   (2) May cause spurious error in advancing to the next track (See
;;;       notes on ChangeLog entry on 2001-06-03).
;;; 
;;;   ;; Maybe killed from the shell (not by emacs):
;;;   ((or (string-match "^killed" stat)
;;;	(string-match "^terminated" stat)
;;;	)
;;;    (let ((monk-buffer (process-buffer proc)))
;;;      (save-excursion
;;;	(and (monk-buffer-live-p monk-buffer)
;;;	     (set-buffer monk-buffer)
;;;	     (monk-stop)))		; deface and cleanup.
;;;      ;; (monk-remove-stale-cdda2wav-file)
;;;      ))

   ;; Otherwise the track ended up normally, or with some error.

   ((or (string-match "^finished" stat)	; normal exit
	;; If `monk-proc-continue-on-failure' is nil,
	;; should not step forward, or loop (just stop)
	(when (string-match "^exited abnormally" stat)
	  (monk-proc-abnormal-exit-handler proc))
	;; `buffer killed' case (19.34). Rather rare, so test this at last.
	(string-match "^hangup" stat))
    ;;
    ;; Use (process-buffer proc) instead of monk-buffer
    ;; `monk-buffer' binding is needed to use the old macros.
    ;;
    (let ((monk-buffer (process-buffer proc)))
      (cond
       ;; Buffer killed (should be eq(monk-buffer)).
       ;; Just (monk-stop), and no further job.
       ;; `monk-stop' will cancel track-timer too.
       ((null (monk-buffer-live-p monk-buffer))
	;; (monk-stop) ; cancel-timer is now done by `kill-buffer-hook'.
	;; `deface' is not required (buffer does not exist).
	(monk-process-started-buffers-cleanup monk-buffer))

       ;; Now `monk-buffer' is alive hereunder.

       ((save-excursion
	  (and ;; (monk-buffer-live-p monk-buffer) ;; 
	       (set-buffer monk-buffer)
	       monk-loop-data))		; looping this track
	(save-excursion
	  (set-buffer monk-buffer)
	  (let (data)
	    (monk-stop-cleanup)
	    (cond
	     ;; one-track-loop
	     ((and (nth 1 monk-loop-data)
		   (atom (nth 1 monk-loop-data))) ; `car' is not suitable.
	      (setq data monk-loop-data))
	     ;; else multi-track-loop
	     (t (setq monk-loop-queue (or (cdr monk-loop-queue)	; Advance queue, 
					  monk-loop-data)) ; or back to top.
		(setq data (car monk-loop-queue)) ))
	    ;; Use current buffer context.
	    ;; Mon Apr 30 17:32:47 2001
	    ;; Track rebuilding may take time, so temporally bind
	    ;; `monk-rebuild-track-auto' to nil if we are scratching.
	    (let ((monk-rebuild-track-auto (not (monk-scratching-p))))
	      (prog1
		  (apply (function monk-play-track) (append data '(keeploop)))
		;; Track may be under downloading/ripping.
		;; `monk-play-track' does this too, but current-cent etc.
		;; are not updated correctly without this: it means
		;; this is not the correct fix but just a workaround.
		(when monk-rebuild-track-auto (monk-rebuild-track nil)))))))
       (t
	;; Must cruise in this `monk-buffer'
	;; Just `save-excursion' is not good enough.

	;; (monk-with-monk-window-and-buffer ; old 
	;; (save-excursion (set-buffer monk-buffer)
	(monk-with-monk-window-or-buffer ; new macro, scrolling around.
	 (let ((pos (and (markerp monk-process-marker)
			 (marker-position monk-process-marker))) ; save
	       (track monk-current-track)) ; save
	   ;; (monk-stop) ; the process is surely stopped, so 
	   (monk-stop-cleanup)
	   (monk-post-sentinel0 track)
	   (cond
	    ((and monk-proc-track-interval
		  (numberp monk-proc-track-interval))
	     (message "MonK pausing for %s seconds ..." monk-proc-track-interval)
	     (sleep-for monk-proc-track-interval)
	     (message "MonK pausing for %s seconds ... Done" monk-proc-track-interval)
	     )
	    (t ;; (sleep-for monk-wait-proc-to-die) ;; NOP testing without this wait.
	       t))
	   (cond ((prog1 (monk-play-next 1 pos)
		    (monk-post-sentinel1 track))) ; just success t returns
		 (monk-repeat
		  ;; Maybe buffer is C-c C-k'ed.
		  (message
		   "MonK couldn't continue to repeat. Maybe buffer cleared")
		  nil)
		 (t (message "MonK played all tracks") nil))))))))))

(defsubst monk-post-sentinel0 (track)
  (cond ((eq major-mode 'monk-ripper-mode)
	 (let ((outfile (monk-assq-get-value 'outfile track)))
	   ;; (monk-remove-stale-cdda2wav-file)
	   (message "MonK ripped: %s" outfile)
	   ))))

(defsubst monk-post-sentinel1 (track) ())

;;
;; Face support
;; Thu Dec 21 16:13:27 2000
;;

(defvar monk-use-face nil
  "*Set to non-nil if face display of current playing track is desired.
You can set favourite `face' symbol to
`monk-current-track-X-face-prop' \(for emacs running as X client\), and
`monk-current-track-T-face-prop' \(non-X environment\).")

(defvar monk-current-track-X-face-prop
  'underline "*Face used in MonK when emacs runs in a window system.")
(defvar monk-current-track-T-face-prop
  'underline "*Face used in MonK when emacs runs on console.")

;; Some examples
;; (setq monk-current-track-X-face-prop 'RoyalBlue)
;; (setq monk-current-track-X-face-prop 'custom-invalid-face)
;; (setq monk-current-track-X-face-prop 'custom-set-face)
;; (setq monk-current-track-X-face-prop 'widget-field-face)

(defun monk-face-setup ()
  "Setup face. Called from `monk-mode'.
Copy face from
`monk-current-track-X-face-prop' \(if emacs running as X client\), or
`monk-current-track-T-face-prop' \(non-X environment\) to
`monk-current-track-face'."
  (copy-face (if window-system
		 monk-current-track-X-face-prop
	       monk-current-track-T-face-prop)
	     'monk-current-track-face))

(defvar monk-face-overlay nil "Internal storage for overlay.")
(defvar monk-buggy-make-overlay nil "Internal use only.  Automatically set.")

;; Old move-overlay support
;; Now we have `monk-with-no-face' and this section may not be
;; required using old-style `make-overlay' call only, but I believe this
;; kluge make `monk-with-no-face' run fast.

(cond ((condition-case nil
	   (progn (make-overlay 0 0 nil t t) t) ; This fails in emacs-19.28
	 (error nil))
       ;; emacs-20.x/19.34
       (setq monk-buggy-make-overlay nil)
       (defsubst monk-make-overlay (rb re)
	 ;; Only available in E19.34/20.x:
	 ;; (make-overlay BEG END &optional BUFFER FRONT-ADVANCE REAR-ADVANCE)
	 (make-overlay rb re nil t nil)))
      ;; No other alternative for emacs-19.28'
      (t (setq monk-buggy-make-overlay t)
	 ;; (fset 'monk-make-overlay 'make-overlay) ; won't compile in 19.28
	 (defsubst monk-make-overlay (rb re)
	   ;; (make-overlay BEG END &optional BUFFER TYPE)
	   (make-overlay rb re))
	 ))

;; Caring E19.28's `make-overlay' --- works fine, but hairy.
;; In emacs-19.28, we can use just (make-overlay rb re), and the
;; overlay grows towards BOB if char is inserted at the top of overlay,
;; so that this trick is required. The idea is shrinking overlay to
;; its end.  No harm even if called while (point) is not on the overlay line.

(defmacro monk-with-no-face (&rest body)
  (` (let* (ol
	    ols ole
	    (doit (and monk-buggy-make-overlay
		       monk-face-overlay
		       (setq ol (car monk-face-overlay))
		       (eq (current-buffer) (overlay-buffer ol)))))
       (cond
	(doit (setq 
	       ;; 1+(): needed to inhibit growth when inserted
	       ;; on point (overlay-start ol): In other words, marker
	       ;; stays on (overlay-start ol) without 1+().
	       ols (set-marker (make-marker) (1+ (overlay-start ol)))
	       ole (set-marker (make-marker) (overlay-end ol)))
	      (move-overlay ol ole ole)))
       ;; (eval (cons (quote progn) (quote (, body)))) ; insert may occur here
       (prog1
	   (,@ body)
	 (cond (doit (move-overlay ol (1- ols) ole)))))))

(defsubst monk-face-region (face rb re)
  "`overlay-put' FACE on region RB RE.
Put overlays are stored `monk-face-overlay' via `cons'."
  (let ((ol (monk-make-overlay rb re)))
    (setq monk-face-overlay (cons ol monk-face-overlay))
    (overlay-put ol 'face face)))

(defsubst monk-face-current-track ()
  "Face 'monk-current-track-face on current line."
  (monk-face-region 'monk-current-track-face
		    (save-excursion (beginning-of-line) (point))
		    (save-excursion (end-of-line) (point))))

(defsubst monk-unface (&optional allbuf)
  "Delete overlay stored in `monk-face-overlay' with `delete-overlay'.
Only the overlay of current buffer is deleted unless the optional arg
ALLBUF is non-nil."
  (when monk-face-overlay
    (let (tem)
      (mapcar 
       (lambda (ol)
	 (cond
	  ((or allbuf
	       (eq (current-buffer) (overlay-buffer ol)))
	   (delete-overlay ol))
	  (t (setq tem (cons ol tem))))) ; else copy to tem
       monk-face-overlay)
      (setq monk-face-overlay (nreverse tem)))))

;;;
;;; Track timer 
;;; Sat Nov 25 16:57:37 2000
;;;
;;; Do track timer display by our timer (not by cdda2wav's) handler,
;;; because disply resolution of cdda2wav is too low.
;;;

(defvar monk-track-timer-handler (function monk-mode-line-update)
  "Disired track timer monitoring function.
The function should accept two argument:
CDTIMESTR, and SECOND \(float maybe\).")

(defvar monk-track-timer-start-sec nil "Internal use only for track timer.")
(defvar monk-track-timer-start-usec nil "Internal use only for track timer.")

(defvar monk-track-timer-obj nil "Internal use only for track timer.")

(defconst monk-track-timer-interval
  ;; (if (string< "20" emacs-version) .5 1) ;; much diviation and noisy.
  ;; Now defaults 1 in all emacses.
  1 "Internal use only for track timer.")

(defun monk-start-track-timer (&optional dispfun)
  "Start track timer display.
Already running timer just stopped."
  ;; (interactive "P")
  (let ((now (current-time)))
    (if monk-track-timer-obj (cancel-timer monk-track-timer-obj))

    (setq monk-track-timer-start-sec (nth 1 now)
	  monk-track-timer-start-usec (nth 2 now))

    (monk-timer-disp-handler)		; 1st immediate update
    (setq monk-track-timer-obj
	  ;; This runs even in 19.28 emacs.
	  (run-at-time (format "%ssec" monk-track-timer-interval)
		       monk-track-timer-interval
		       'monk-timer-disp-handler))))

(defun monk-timer-disp-handler ()
  "Calls elapsed time in \"xx:yy\" format.
Relays the time data `monk-track-timer-handler' if it is non-nil."
;;;  (mapcar (lambda (b) (monk-timer-disp-handler-subr b))
;;;	  monk-process-started-buffers)
  (mapcar 'monk-timer-disp-handler-subr monk-process-started-buffers))

(defsubst monk-timer-disp-handler-subr (buf)
  (or (and
       buf (monk-buffer-live-p buf)
       (save-excursion
	 (set-buffer buf)
	 (let* (tem
		(now (current-time))
		(secdelta (- (nth 1 now) monk-track-timer-start-sec))
		(usecdelta
		 (cond
		  ((= secdelta 0)
		   (- (nth 2 now) monk-track-timer-start-usec))
		  (t
		   (setq tem
			 (+ (nth 2 now) (- 1000000 monk-track-timer-start-usec)))
		   (if (< tem 1000000)
		       tem
		     ;; otherwise, usec part overflowed.
		     (setq secdelta (1+ secdelta))
		     (- tem 1000000)))))
		(second
		 (+ monk-current-start-sec
		    (string-to-number (format "%d.%d" secdelta usecdelta)))))
	   ;; (message "%s" second)
	   (if monk-track-timer-handler
	       (funcall monk-track-timer-handler second) ; Returns non-nil
	     second) ;; Seems returning string is okay for the system (?)
	   )))
;;; The Buffer Is Surely Killed (by C-x k), but,
;;; timer can't commit suicide in E19.xx.  So this ghost timer should
;;; be live with NOP, until killed by somebody (sentinel ?).
;;;   (progn
;;;     ;; (monk-cancel-track-timer buf) ; so this can't be used.
;;;     ;; (cancel-timer monk-track-timer-obj) ; direct call won't work either.
;;;     ;; (setq buf nil monk-track-timer-obj nil)
;;;     )
      ))

(defun monk-cancel-track-timer (buf)
  "Stop track timer display.
Do the relevant job only when `monk-track-timer-obj' is non-nil."
  (save-excursion (set-buffer buf)
		  (cond (monk-track-timer-obj
			 (cancel-timer monk-track-timer-obj)
			 (setq monk-track-timer-obj nil)))))

(add-hook 'kill-buffer-hook (lambda ()
			      (monk-cancel-track-timer (current-buffer))))
			      
;; (monk-current-track-time) => 89.397582
;; (monk-current-track-time 'cdtime) => "01:41"
(defun monk-current-track-time (&optional cdtime)
  "Bind `monk-track-timer-handler' to nil, and returns \
`monk-timer-disp-handler' output as second \(float maybe\).
\"xx:yy\" formated string is returned  if optional arg CDTIME is non-nil."
  (let* ((monk-track-timer-handler)
	 (tem (monk-timer-disp-handler)))
    (if cdtime (monk-sec2cdtime tem) tem)))

;;;
;;; Routines for Time parameter operations (some are not used)
;;;

;; Tue Nov 21 16:06:57 2000
;; Support the subseconds part returned by cdda2wav
;; cdda2wav returns sort of "2:59.25" timing data.
;; (query-replace "string-to-int" "string-to-number" nil)
;; (query-replace "integerp" "numberp" nil)

(defconst monk-cdtime-regex "\\([0-9\.]+\\):\\([0-9\.]+\\)")

;; "10:32.08" sort of cdtime => return as number (float maybe)
;; (monk-cdtime2sec "1.2:1.01") => 73.01

(defsubst monk-cdtime2sec (tstr)
  "Convert \"xx:yy\" time string TSTR into seconds and return as number \
\(float maybe\)."
  (save-match-data
    (and (stringp tstr)
	 ;; Loose match on the both sides of ":"
	 (string-match monk-cdtime-regex tstr)
	 ;; Do not round() here to get int.
	 (+ (* 60.0 (string-to-number
		     (substring tstr (match-beginning 1) (match-end 1))))
	    (string-to-number
	     (substring tstr (match-beginning 2) (match-end 2)))))))

;; str/number (should be int) => return cdtime string "10:32"
;; Accept `sec' in float, but it's truncated.
;; (monk-sec2cdtime 72.9) => "01:12"
;; round() is used for most of the other arithmetics operations,
;; but this exceptionally uses truncation to step track-timer nicely.

(defsubst monk-sec2cdtime (sec)
  "Convert time SEC \(float accepted\) to time string \"xx:yy\".
Subsecond part is always truncated to get the output format."
  (and
   (setq sec (or (and (numberp sec) ;; (if float sec (round sec))
		      sec)	    ;; round causes divation. Truncate by "%d".
		 (and (stringp sec) (string-to-int sec))))
   (format "%02d:%02d" (/ sec 60.0)
	   ;; (% sec 60) ;; barfs if sec is float
	   ;; (% (truncate sec) 60) ;; NO!
	   (mod sec 60)
	   )))

;; cdtime str ("12:36.92") => return number diff (float maybe)
(defsubst monk-cdtime-diff (a b &optional cdtime)
  "Return difference of cdtime A B \(A - B\).
Optional non-nil arg CDTIME requests output in cdtime \"xx:yy\" format."
  (let ((sa (monk-cdtime2sec a))
	(sb (monk-cdtime2sec b)))
    (cond
     ((null sa) nil)
     ((null sb) nil)
     (t (if cdtime (monk-sec2cdtime (- sa sb)) (- sa sb))) )))

;; str(float,int) => return integer
(defsubst monk-sec2sector (sec &optional altfps)
  "Convert second data SEC \(integer or float, or string\) \
to sector-number integer, which can be handled by cdda2wav\(1\).
Number of sector corresponding to a second stored in `monk-fps'
is used to get sector number.  If optional 2nd arg ALTFPS is non-nil,
the value is use instead of `monk-fps'."
  (and
   (setq sec (or (and (numberp sec) sec)
		 (and (stringp sec) (string-to-number sec))))
   (round (* sec (or altfps monk-fps)))))

;; str(float,int) => return integer
(defun monk-any2sector (str)
  "Convert any time string STR into sector \(aka. frame\)
Supported format is:
\"01:37\", \"78\" \(in second\), or \"1932f\" \(in frames\)."
  (save-match-data
    (cond
     ((string-equal "" str) 0)		; User may hit just RET.
     ((string-match monk-cdtime-regex str)
      (monk-sec2sector (monk-cdtime2sec str)))
     ((string-match "^[0-9.]+$" str) ; number only float is okay here.
      (monk-sec2sector str))
     ((string-match "^\\([0-9]+\\)f$" str)
      (string-to-number
       (substring str (match-beginning 1) (match-end 1))))
     (t (error "Format of `%s' is not correct" str)))))

(defsubst monk-sector2sec (sector &optional altfps)
  "SECTOR to second to conversion.
`monk-fps' value is used as sector \(aka. frame\) per second data.
If optional 2nd arg ALTFPS is non-nil, the value is use instead of
`monk-fps'."
  (round (/ sector (or altfps monk-fps))))

(defsubst monk-sector2cdtime (sector)
  "SECTOR to cdtime conversion."
  (monk-sec2cdtime (monk-sector2sec sector)))

;;;
;;; Loop playing
;;; Advancing in the list is handled by `monk-proc-sentinel'.
;;;

(defvar monk-loop-data nil "Internal use only.")
(defvar monk-loop-queue nil "Internal use only for repeat songs.")
;; No `shuffle' yet.
(defvar monk-repeat t  "*Non-nil forces repeat-playing all tracks.")

(defun monk-get-or-read-track ()
  "Get current line's track property. or ask user track number for cdaudio.
The minibuffer interface is done on failure of `monk-get-this-track'."
  (let ((trk (monk-get-this-track)))
    (cond (trk trk)
	  (t (error "The track broken: %s" trk)))))

(defun monk-play-loop (track beg dur)
  "Play loop in TRACK from BEG for duration DUR.
Refer to `monk-play-track' for the format of TRACK, BEG, and DUR."
  ;; (interactive "nWhat track: \nsFrom where: \nsDuration: ")
  (interactive (list (cond ((monk-proc-running-p) monk-current-track)
			   (t (monk-get-or-read-track)))
		     (read-string "From where: ")
		     (read-string "Duration: ")))
  (let ((beg (if (numberp beg) beg (monk-any2sector beg)))
	(dur (if (numberp dur) dur (monk-any2sector dur))))
    ;; (monk-stop)                        ; current buffer
    (monk-stop-process-mutex track)
    (setq monk-loop-data (list track beg dur))
    (apply (function monk-play-start) monk-loop-data) ))

(defun monk-play-loop-a-track (track)
  "Loop single TRACK."
  (interactive (list (monk-get-or-read-track)))
  (cond ((and (monk-proc-running-p) (equal track monk-current-track))
	 (setq monk-loop-data (list track 0 0)
	       monk-current-loop-name
	       (cond ((listp track)
		      (format "%s " (cons
				     (monk-assq-get-value 'type track)
				     (cdr monk-loop-data))))
		     (t (format "%s " monk-loop-data)))))
	(t (monk-play-loop track 0 0)))
  (message "Looping %s"
   (cond ((eq 'cda (monk-assq-get-value 'type track))
	  (format "Track %s" (monk-assq-get-value 'title track)))
	 ((listp track)
	  (format "a track for file: \"%s\""
		  (monk-assq-get-value 'file track)))
	 (t nil))))

;; This command makes sense for integer (old 'cda) tracks.
;; Un-bound in `monk-mode-map'.
(defun monk-play-loop-tracks (tracks)
  "Loop tracks specified as comma, tab, or space-separated list."
  (interactive "sTrack list space-separated: ")
  (setq monk-loop-queue
	(mapcar (lambda (tr)
		  (cons tr (list 0 0)))
		(mapcar 'string-to-int (dired-split "[\, \t]+" tracks)))
	monk-loop-data monk-loop-queue) ; Sentinel knows to deal with it.
  ;; (apply (function monk-play-start) (car monk-loop-queue))
  (apply (function monk-play-track) (car monk-loop-queue)))

(defsubst monk-get-tracks-region (rb re)
  "Return tracks from region RB to RE."
  (let (trklist)
    (save-excursion
      (goto-char rb)
      (while (and (not (eobp)) (< (point) re))
	(setq trklist (cons (monk-get-track) trklist))
	(forward-line 1)))
    (nreverse trklist)))

(defsubst monk-goto-region-beginning (rb re)
  "Goto region beginning by `exchange-point-and-mark' unless RB equals point.
Returns t unless `exchange-point-and-mark' is not called, else nil."
  (and
   ;; `integer-or-marker-p' isn't be required because `monk-play-loop-region'
   ;; is interactive("r") command.
   ;;(integer-or-marker-p rb) (integer-or-marker-p re)
   (< rb re)
   (or (= (point) rb)
       (exchange-point-and-mark))))

(defun monk-play-loop-region (rb re)
  "Loop tracks belonging to region RB to RE."
  (interactive "r")
  (monk-goto-region-beginning rb re)
  ;; Almost the same construction as `monk-loop-marked-tracks'.
  (let ((trklist (monk-get-tracks-region rb re)))
    (cond
     (trklist
      (setq monk-loop-queue (mapcar (lambda (x) (cons x (list 0 0))) trklist)
	    monk-loop-data monk-loop-queue)
      (apply (function monk-play-track)
	     (append (car monk-loop-queue) '(keeploop))))
     (t (error "Can't get track out of the region")))))

(defsubst monk-loop-clear0 ()
  (setq monk-loop-data nil
	monk-loop-queue nil
	monk-current-loop-name ""))

(defun monk-loop-clear ()
  "Clear the current loop."
  (interactive)
  (cond (monk-loop-data
	 ;; This does `restart-from-current-offset', but I can't remember
	 ;; why it is needed, and with this condtions.
	 ;; There are three type of loop-data:
	 ;; (1) '(<track> 0 0) : looping a whole track.
	 ;; (2) '((<track2> 0 0) (<track3> 0 0) ...) : multiple tracks 
	 ;;      (currently beg/dur values are always assumed to be 0).
	 ;; (3) '(<track> 380 38) 1 second loop (scratch) from 0:10
	 ;; and only (3) requires the `restart-from-current-offset'.

	 ;; (or     ;; Restart except Type (1): wrong.
	 ;;  (and (monk-proc-running-p) (monk-one-track-loop-p))
	 ;;  (monk-play-track monk-current-track monk-current-offset 0))

	 (when (and (interactive-p) ; failsafe (see monk-stop calls).
		    (monk-proc-running-p)
		    (monk-scratching-p))
	   (monk-play-track monk-current-track monk-current-offset 0))

	 (monk-loop-clear0)
	 (when (interactive-p) (message "Loop data cleared")))
	(t (and (interactive-p) (message "Loop is not running")))))

(defsubst monk-one-track-loop-p ()
  "Test `monk-loop-data' and predicate if playing whole one track loop."
  (and monk-loop-data
       (nth 1 monk-loop-data)
       (atom (nth 1 monk-loop-data))
       (= 0 (+ (car (cdr monk-loop-data)) (car (cdr (cdr monk-loop-data)))))))

(defsubst monk-scratching-p ()
  "Test `monk-loop-data' and predicate if DJ is scratching."
  (and monk-loop-data
       (nth 1 monk-loop-data)
       (atom (nth 1 monk-loop-data))
       (not (= 0 (+ (car (cdr monk-loop-data))
		    (car (cdr (cdr monk-loop-data))))))))

;;;
;;; User interface
;;;

(defvar monk-rebuild-track-auto
  nil "*Defaults nil.
Non-nil causes MonK to reconstruct `track' before play it.
Forced to nil automatically while looping certain part of single track.
See `monk-scratching-p'.")

(defvar monk-play-track-verbose
  t "*If non-nil, `monk-play-track' messages on starting play a new track.
Defaults t, and you can see info on the track in the beginning of a song,
even when MonK buffer is running in an invisible frame.")

(defun monk-play-track (track beg dur &optional keeploop)
  "Simply start to play TRACK, specifying BEG, DUR.
An interactive interface to `monk-play-start'.

BEG can be \"01:37\", \"78\" \(in second\), or \"1932f\" \(in frames\).
DUR can be \"3\" \(in second\) or \"152f\" \(in frames\).
In non-interactive use, BEG and DUR can also be integer, and they are
interpreted as a number of sector.
Non-nil optional 4th arg KEEPLOOP stops just the player process, and leave
the `monk-loop-data'.

Unless `monk-rebuild-track-auto' is nil, the target track data is
automatically updated before playing it."
  (interactive (list (or (and (monk-proc-running-p)
			      monk-current-track)
			 (read-string "What track: "))
		     (read-string "From where: ")
		     (read-string "Duration: ") ))
  (let ((beg (if (numberp beg) beg (monk-any2sector beg)))
	(dur (if (numberp dur) dur (monk-any2sector dur))))

    ;; (monk-stop keeploop)		; current buffer
    (monk-stop-process-mutex track keeploop) ; others

    (when monk-rebuild-track-auto (monk-rebuild-track nil))
    (when (or monk-play-track-verbose
	      ;; else report start if the buffer's window is invisible.
	      (not (window-live-p (get-buffer-window (current-buffer)
						     (selected-frame)))))
      (message "MonK started: %s" (monk-play-generate-track-name track)))

    (monk-play-start track beg dur)))

;; monk-stop moved to player engine section. Sun Dec 31 22:23:40 2000

(defun monk-play-this-track (&optional arg)
  "Play a `track' corresponding to the current."
  (interactive "p")
  (let ((track (monk-get-this-track)))
    (cond (track (monk-play-track track 0 0))
	  (t (and (interactive-p)
		  (message "No track there, or the track is broken, sister"))
	     nil))))			; Should return nil here

;;
;; `monk-play-next' and `monk-play-prev'
;; `monk-play-next' is called also from the sentinel.
;;

(defun monk-play-prev (arg)
  "Move ARG tracks backward and play the track.
If process is running, point is moved to the relevant track before the advance."
  (interactive "p")
  (cond
   ((monk-play-next (- arg)))		; test result returned.
   (t (when (interactive-p)		; move-up warning only
	(message "MonK can't move %d %s up sister..."
		 arg (if (= 1 arg) "line" "lines"))))))

;; Advancing to next monk buffer.
;; This replaces `monk-play-next', and should be
;; loaded from `monk-load-hook'.

;; (setq monk-repeat-buffers t)

;; (provide 'monk-repeat-buffers)

;; ~/elisp/buffer-circulate.el has
;; monk-circulate-list2, monk-circulate-list3, and monk-circulate-list
;; but we have `monk-next-buffer' and `monk-previous-buffer'.

(defvar monk-repeat-buffers nil
  "If non-nil, command `monk-play-prev', `monk-play-next-buffer', \
`monk-play-previous-buffer' or automatic track advancing steps into
next MonK buffer if any.")

(defun monk-toggle-repeat-buffers (arg)
  "Toggles `monk-repeat-buffers'. Non-nil ARG forces turning on."
  (interactive "P")
  (prog1
      (setq monk-repeat-buffers
	    (cond
	     (arg t)
	     (t (null monk-repeat-buffers))))
    (message "Repeat Buffer is now %s"
	     (if monk-repeat-buffers "on" "off"))))

;; Called within `monk-play-next', with testing `monk-repeat-buffers'.
(defun monk-play-next-buffer (arg)
  "Advance to next MonK buffer, and play the buffer."
  (interactive "p\nP")
  (cond ((< (length monk-buffer-list) 2) nil) ; one buffer, or none
	(t (if (> arg 0) (monk-next-buffer arg)
	     (monk-previous-buffer (- arg)))
	   (prog1 (current-buffer)
	     (when (eobp) (goto-char (point-min)))
	     (cond ((monk-get-track) (monk-play-this-track))
		   (t (monk-play-next-buffer arg)))))))

(defun monk-play-previous-buffer (arg)
  "Step into previous MonK buffer. Return nil on failure.
Starts playing if optional arg PLAY is non-nil."
  (interactive "p\nP")
  (monk-play-next-buffer (- arg)))

;; But, `monk-play-next' is very difficult to read. Must re-implementation ?

;; This code works fine by n, p (interactive call), but raises
;; "MonK couldn't continue to repeat. Maybe buffer cleared" error in
;; `monk-proc-sentinel' why ?

(defun monk-play-next (arg &optional pos)
  "Move ARG tracks forward, and play the track.
If process is running, point is moved to the relevant track before stepping."
  (interactive "p")
  (let* ((fwd (>= arg 0))		; 0 is impossible for "p"
	 (count (if fwd arg (- arg)))
	 (pt (point)) (jump 0) (leftmove))

    ;; Start position control,
    (cond 
     (pos (goto-char pos))		; supplied by sentinel.
     ((and (markerp monk-process-marker) ; process-status check un-necessary
	   (marker-position monk-process-marker))
      (goto-char monk-process-marker)))

    ;; then try to move to valid track, moving ARG lines
    (cond
     ((catch 'success
	;;(cond ((= 0 (buffer-size)) (throw 'success nil))) ; done by `jump'
	(save-excursion
	  (while (> count 0)
	    (or (progn (setq leftmove (forward-line (if fwd 1 -1)))
		       (cond ((= leftmove 0) t)      ; else move failed:
			     (monk-repeat nil)
			     (t (throw 'success nil)))) ; `monk-repeat' is null
		(and monk-repeat fwd (eobp)
		     (monk-jump-to-top) (setq jump (1+ jump)))
		(and monk-repeat (null fwd) (bobp)
		     (monk-jump-to-bottom) (setq jump (1+ jump))))
	    (cond ((monk-get-track) (setq count (1- count)))
		  ((< 1 jump) (throw 'success nil))))
	  (setq pt (point)))
	t)				; if while() ends up catch() catches t
      (goto-char pt)
      (cond
       ;; If `monk-repeat-buffers', eob/eop hit and we have 2 bufs or more:
       ((and monk-repeat-buffers (> jump 0) (> (length monk-buffer-list) 1))
	(when (monk-proc-running-p)
	  (sleep-for monk-wait-proc-to-die)) ; sentinel fails without this wait.
	(monk-play-next-buffer (if (> arg 0) 1 -1)))
       (t (monk-play-this-track))))
     (t (when (interactive-p)		; move-down warning only
	  (message "MonK can't move %d %s down sister..."
		   arg (if (= 1 arg) "line" "lines")))
	nil))))

(defvar monk-shuttle-search-resolution 1
  "*Shuttle search resolution in second. Defaults 1 sec.")

(defsubst monk-get-shuttle-search-resolution ()
  "Returns shuttle search resolution in `monk-fps' unit."
  (* monk-fps monk-shuttle-search-resolution))
;; (monk-get-shuttle-search-resolution) =>  30.0

(defun monk-play-backward (arg)
  "Move playing position of current track backward by \
`monk-shuttle-search-resolution'."
  (interactive "p")
  (monk-play-forward (- arg)))

(defun monk-play-forward (arg)
  "Advance playing position of current track by \
`monk-shuttle-search-resolution'."
  (interactive "p")
  (let* ((resol (monk-get-shuttle-search-resolution))
	 (beg (+ monk-current-offset (* resol arg)))
	 (onetrackloop (monk-one-track-loop-p))
	 (track monk-current-track))
    (cond
     ((monk-proc-running-p)
      (cond
       ((>= beg monk-current-track-length)
	(message "Out of range (%s sec), advanced to next track"
		 (* arg monk-shuttle-search-resolution)) ; in second.
	(monk-play-next 1))
       ((< beg 0)
	(message "Out of range (%s sec), rewound to previous track"
		 (* arg monk-shuttle-search-resolution)) ; in second.
	(monk-play-next -1))
       (t ;; (message "%s %s" monk-current-offset beg)

	;; like `monk-play-next'
	(when (and (markerp monk-process-marker) ; no process-status check
		   (marker-position monk-process-marker))
	  (goto-char monk-process-marker)
	  (forward-line 0))

	(monk-play-track track beg 0 onetrackloop))))
     (t (message "Sorry, no track is playing, sister") nil))))

(defun monk-menu-play-forward (arg)
  "Call `monk-play-forward' with x4 multiplied arg."
  (interactive "p")
  (cond ((monk-proc-running-p)
	 ;; "Back To Future" because can't tell how long user has
	 ;; been displayed the menu.
	 (monk-timer-disp-handler)	; update time/offset control data
	 (monk-play-forward (* 4 arg)))))

(defun monk-menu-play-backward (arg)
  "Call `monk-play-backward' with x4 multiplied arg."
  (interactive "p")
  (cond ((monk-proc-running-p)
	 (monk-timer-disp-handler)	; Same fix as `forward' fun above.
	 (monk-play-backward (* 4 arg)))))

(defun monk-play-pause (arg)
  "The name tells all about.
With non-nil ARG, `monk-play-forward' is called with that arg."
  (interactive "P")
  (cond
   ((monk-proc-running-p)
    (if (null arg)
	(monk-stop 'keeploop)
      (monk-play-forward (prefix-numeric-value current-prefix-arg))))
   ;; Else we restart playing from current pos.
   ;; Arg
   (t 
    ;; Don't know what I am doing. Is this correct ?
    ;; Surely (not (monk-proc-running-p)) here, so this stops the
    ;; conflicting process in other buffer.
    (when monk-current-track
      (monk-stop-process-mutex monk-current-track 'keeploop))

    ;; Else resume from the paused position.
    (cond
     ;; loop check should be on top of this cond()
     ;; Ordinary `resume' case sometimes fail.
     ;; This may be inconsistent, 'cause always starts from top of loop.
     ((and monk-current-track ;; over-kill ? just loop-data is good enough ?
	   monk-loop-data)
      (let ((ldata (cond
		    ;; This was not good enough.
		    ;; ((atom (nth 1 monk-loop-data)) monk-loop-data)
		    ((monk-one-track-loop-p)
		     (list monk-current-track
			   monk-current-offset 0))
		    ((monk-scratching-p) monk-loop-data)
		    ;; Multiple-tracks loop.
		    ((listp (car monk-loop-data))
		     (car monk-loop-data))
		    (t (error
			"BUG(?):Wrong loop data %s" monk-loop-data)))))
	;; Surely `paused' here, so, use monk-play-start():
	(apply 'monk-play-start ldata)
	;; (apply 'monk-play-track (append ldata (list 'keeploop)))
	))
     ;; *monk* buffer just loaded, and no track played yet
     ((null monk-current-track) (monk-play-this-track))
     ;; User has been moved.
     ((and monk-current-track
	   ;; `equal' is not good enough if the mp3 file is
	   ;; downloading, or the mp3 track builder yields deviation
	   ;; on `frames', `cent', or time.
	   (null
	    (monk-on-playing-track-p
	     monk-current-track (monk-get-this-track) 'nomarkerchk)) )
      (monk-play-this-track))
     ;; Ordinary `resume' case
     ((and monk-current-track monk-current-offset)
      ;; (monk-play-start monk-current-track monk-current-offset 0)
      (monk-play-track monk-current-track monk-current-offset 0)
      )
     ;; or from first. Should resume.
     (t
      (let ((track (monk-get-this-track)))
	(cond
	 (track	;; (monk-play-start track 0 0)
	  (monk-play-track track 0 0))
	 (t (message "No track there, or the track is broken, sister")))))))))

(defun monk-goto-current-track ()
  "Move to the current playing track."
  (interactive)
  (let ((pt (or (and (markerp monk-process-marker)
		     (marker-position monk-process-marker))
		(point))))
    (when pt
      (goto-char pt)
      ;; (forward-line 0)
      (when goal-column
	(move-to-column goal-column)))))

(defun monk-goto-track-integer (number)
  "Move cursor to the track having 'track-number NUMBER \(must be integer\).
This command makes sense only for cdaudio tracks.
Returns point on success, otherwise nil."
  (interactive "p")
  (let (num pt)
    (when (save-excursion
	    (goto-char (point-min))
	    (catch 'found
	      (while (not (eobp))
		(when (and 
		       (setq num (monk-assq-get-value
				  'track-number (monk-get-track 'raw)))
		       (integerp num)
		       (= num number))
		  (throw 'found (setq pt (point))))
		(forward-line 1))
	      nil))
      (goto-char pt))))

(defun monk-cda-revert ()
  "Revert *monk* buffer with cdaudio track.
All the tracks in the *monk* buffer is blown away.
Current buffer contents are all erased and tracks are read in from cd audio.
This commands exists to read newly inserted cd audio media."
  (interactive)
  (monk-quit)
  (monk "/dev/cdrom"))

(defun monk-quit ()
  "Quit `monk' mode, and kill the buffer."
  (interactive)
  (cond
   ((monk-proc-running-p)		; checks only current-buffer
    ;; No need to resume anymore. Not done by monk-stop().
    (setq monk-current-track nil)
    (monk-stop)))
  (kill-buffer (current-buffer)))

(defvar monk-eject-command "eject"
  "*Shell command used to eject cd media.
Defaults \"eject\", and it may be dependent upon linux.
Set desired shell command line as string if the default is not suitable.")

(defun monk-eject (arg)
  "Eject cd media with `monk-eject-command'.
Non-nil ARG forces eject operation without `y-or-no' query."
  (interactive "P")
  (let ((doit (lambda ()
		;; (monk-stop) ;; keep the buffer with this ?
		(monk-quit)
		(monk-with-coding-system
		 (call-process shell-file-name
			       nil nil nil "-c" monk-eject-command)))))
    (cond ((and (monk-proc-running-p)
		(or arg
		    (y-or-n-p "You're playing cd. Force ejecting it ? ")))
	   (funcall doit))
	  (t (funcall doit)))))

(defun monk-shell-command (command)
  "Execute COMMAND on file property of the track on-point."
  (interactive "sshell command: ")
  (let ((file (monk-assq-get-value 'file (monk-get-this-track))))
    (cond
     (file (shell-command (dired-shell-stuff-it command (list file) nil)))
     (t (message "No file contained in this track")))))

;;
;; Info buffer popup 
;; Maybe for nosy user only
;;

(defsubst monk-show-buffer (buffer)
  "Show BUFFER without selecting it. Place the point at bob."
  (monk-with-current-buffer		; this checks buffer existence
   buffer
   (and (eobp) (goto-char (point-min)))) ; cheap. meaning not-shown ?
  (display-buffer buffer))

(defun monk-show-info (arg)
  "Readonly version of `monk-prop-edit'.
With non-nil arg, displays `monk-info-buffer' instead for debugging."
  (interactive "P")
  (cond (arg (monk-show-buffer monk-info-buffer)) ; it checks buffer existence.
	((listp monk-current-track)
	 (monk-prop-edit nil 'readonly))))

(defun monk-show-cddb-info ()
  "Show `monk-cddb-buffer' containing raw cddb info.
The info has come from local cddb database, or remote server."
  (interactive)
  (monk-show-buffer monk-cddb-buffer))

(defvar monk-dired-fun 'dired-other-window
  "*Emacs function called by `monk-dired-on-this'. Should be one of `dired'
family commands such as `dired', `dired-other-frame' etc.")

(defun monk-dired-on-this (arg)
  "Run dired on directory of `file' property of this `track'.
The `dired' command is the one stored in `monk-dired-fun'."
  (interactive "P")
  ;; `monk-get-track' checks if the file could be accessed
  (let ((file (monk-assq-get-value 'file (monk-get-track))))
    (cond
     (file
      (funcall
       ;; `with-monk-coding-system' is not braced here, because `dired'
       ;; operations are out of our control.  Eg., you may see the first
       ;; directory listing if `with-monk-coding-system' is braced here
       ;; but it wouldn't get access files if you did not setup coding
       ;; system for `dired'.
       monk-dired-fun (file-name-directory file)
       (cond
	(arg (read-string
	      "Dired listing switches: " dired-listing-switches))))
      (or (dired-goto-file file)
	  (revert-buffer)
	  (dired-goto-file file)))
     (t
      (message "MonK can't locate file \"%s\""
	       (monk-assq-get-value 'file (monk-get-track 'raw)))))))

;;
;; volume control,
;; (almost) totally depends on `aumix'.
;;
;; `monk-volume-up' and `monk-volume-query' do not support
;; `monk-volume-device-option' values other than "-w" yet.
;; We need a table to interpret options and message catalog, but
;; isn't it rare that people want to use the device other than /dev/dsp (?).
;; How many users 
;;

(defvar monk-volume-command nil
  "*Volume control command, Eg., \"aumix\".
Should not inculde argument part which is generated by
`monk-volume-up' and `monk-volume-down' according to
`monk-volume-device-option' or `monk-volume-query-option'.  These emacs
commands are dependent on aumix\(1\).")

(defvar monk-volume-device-option '("-w")
  "*The default corresponds to aumix\(1\)'s dsp \(pcm\) device.
Any other value, eg. \"-c\" \(cdaudio\) may work, but the result may
not be reported correctly.")

(defvar monk-volume-query-option '("-q") "*aumix\(1\)'s query option.")

(defvar monk-volume-resolution 1
  "*Resolution of volume control by prefix arg given to
`monk-volume-up' and `monk-volume-down' command.")

(defun monk-volume-up (arg)
  "Swell up playing volume by numeric ARG. Relative control only.
If ARG is 0, just the current volume level is reported."
  (interactive "p")
  (cond
   ((and monk-volume-command (stringp monk-volume-command)
	 (= arg 0))
    ;; Currently pcm vol. only.
    (message "Current volume (unchanged): %S" (car (monk-volume-query)))
    (sit-for 3)
    (message ""))
   ((and monk-volume-command (stringp monk-volume-command)
	 (null (= arg 0)))
    (prog1
	;; (monk-with-coding-system ;; unused.
	(apply 'call-process monk-volume-command nil nil nil
	       (append
		monk-volume-device-option
		(list
		 (format "%s%d"
			 (if (> arg 0) "+" "")
			 (* arg monk-volume-resolution)))))
      ;; Currently pcm vol. only.
      (message "Current volume: %S" (car (monk-volume-query)))
      (sit-for 3)
      (message "")))))

(defun monk-volume-down (arg)
  "Lower playing volume by numeric ARG. Relative control only.
If ARG is 0, just the current volume level is reported."
  (interactive "p")
  (monk-volume-up (- arg)))

;; `monk-volume-query' still reports `pcm' and master `vol' only.
;; May have to be modified to report all of them (?), but
;; `monk-volume-up' must find correct value to use for messaging.

(defun monk-volume-query ()
  "Get current volume settings.
Returns a list of cons of pcm, and master volume."
  ;; (interactive)
  (cond
   ((and monk-volume-command (stringp monk-volume-command))
    (let ((buf (get-buffer-create monk-junk-buffer)) v0 v1 m0 m1)
      (monk-with-current-buffer
       buf
       (let ((default-directory (expand-file-name "~/"))) ; safe (?)
	 (erase-buffer)
	 (cond
	  ((and (= 0;; (monk-with-coding-system ;; unused
		   (apply 'call-process monk-volume-command nil t nil
			  monk-volume-query-option))
		(progn (goto-char (point-min))
		       (search-forward-regexp
			"^pcm +\\([0-9]+\\), +\\([0-9]+\\)" nil t)))
	   (setq v0 (buffer-substring (match-beginning 1) (match-end 1))
		 v1 (buffer-substring (match-beginning 2) (match-end 2)))
	   ;; Also get master vol.
	   (progn (goto-char (point-min))
		  (search-forward-regexp
		   "^vol +\\([0-9]+\\), +\\([0-9]+\\)" nil t))
	   (setq m0 (buffer-substring (match-beginning 1) (match-end 1))
		 m1 (buffer-substring (match-beginning 2) (match-end 2)))
	   (list (cons (string-to-number v0) (string-to-number v1))
		 (cons (string-to-number m0) (string-to-number m1)))
	   ))))))))

(defun monk-toggle-repeat (arg)
  "Toggles `monk-repeat'. Non-nil ARG forces turning on."
  (interactive "P")
  (prog1
      (setq monk-repeat
	    (cond
	     (arg t)
	     (t (null monk-repeat))))
    (message "Repeat is now %s"
	     (if monk-repeat "on" "off"))))

;;
;; monk-edit-mode --- almost the same as text-mode except C-c C-c
;; This might have to be minor mode (but later).
;;

(defvar monk-edit-mode-map
  (let ((monk-edit-mode-map (copy-keymap text-mode-map)))
    (define-key monk-edit-mode-map "\C-c\C-c" 'monk-edit-quit)
    monk-edit-mode-map) "Keymap used in `monk-edit-mode'.")
  
(defvar monk-edit-mode-hook nil)

(defun monk-edit-mode ()
  "A mode to edit *monk* buffer.

Available key bindings -- almost the same as `text-mode-map' except
`monk-edit-quit' binding:

\\{monk-edit-mode-map}"
  (interactive)
  ;; This kill disables all local variable, so advancing to next
  ;; track is disabled.  What is the demerit without
  ;; this (kill-all-local-variables) ?  -- perhaps no harm for MonK
  (or (eq major-mode 'monk-mode)
      (kill-all-local-variables))

  (setq truncate-lines monk-truncate-lines)
  (set (make-local-variable 'revert-buffer-function)
       'monk-revert-buffer-function)

  (use-local-map monk-edit-mode-map)
  (setq major-mode 'monk-edit-mode)
  (setq mode-name "MonKEdit")
  ;; (and window-system (font-lock-mode t)) ; Cheap.
  (when window-system (monk-font-lock-setup))
  (run-hooks 'monk-edit-mode-hook))

(defun monk-edit ()
  "Allows to edit buffer, to alter the track line order to change track \
sequence etc.
Each line has track number as `text-property' so you can use
\\[kill-line] \(kill-line\), \\[yank] \(yank\) etc."
  (interactive)
  (setq buffer-read-only nil)
  (when (fboundp 'monk-enable-mode-line-switches-21)
    (monk-enable-mode-line-switches-21))
  (message "Now you can kill and yank lines to change track sequence")
  (monk-edit-mode))

(defun monk-edit-quit ()
  "Quit `monk-edit-mode' and return to `monk-mode'."
  (interactive)
  (message "Returned to `monk-mode'")
  (setq buffer-read-only t)
  (monk-mode))

;; 
;; The key / mouse bindings
;; 

(defvar monk-navigation-menu
  ;; My rough/easy method to define menu. `easy-menu' ? Who cares.
  (cons "MonK"	'(keymap
		  (start-stop ">|" . monk-play-pause)
		  (next ">>|"  . monk-play-next)
		  (prev "|<<"  . monk-play-prev)
		  (fwd ">>" . monk-menu-play-forward)
		  (bwd "<<" . monk-menu-play-backward)
		  (vup "^"  . monk-mouse-volume-up)
		  (vdn "v"  . monk-mouse-volume-down)
		  (tglrepeat "Toggle Repeat" . monk-toggle-repeat)
		  (tglrepeatbuf "Toggle Repeat Buf" . monk-toggle-repeat-buffers)
		  (tglname "Toggle Name" . monk-mp3-mouse-toggle-show-title)
		  (tglreconst "Toggle AutoRebuild"
			   . monk-toggle-rebuild-track-auto)
		  (nextbuf "Next Buffer" . monk-next-buffer)
		  (prevbuf "Previous Buffer" . monk-previous-buffer)
		  (onew "NewBuffer" . monk-open-new)
		  (clear "Clear" . monk-clear-buffer)
		  (undo "Undo" . monk-undo)
		  (quit "Q"  . monk-quit)
		  "MonK"))
  "Menu keymap for `monk' GUI.")

(defvar monk-mode-map
  (let ((monk-mode-map (make-sparse-keymap)))
    (define-key monk-mode-map "\C-m" 'monk-play-this-track)
    (define-key monk-mode-map " " 'monk-play-pause) ; `play/pause' or fwd by arg
    (define-key monk-mode-map "\ep" 'monk-play-track)

    (define-key monk-mode-map "r" 'monk-toggle-repeat)

    (define-key monk-mode-map "n" 'monk-play-next)
    (define-key monk-mode-map "p" 'monk-play-prev)

    (define-key monk-mode-map [tab] 'monk-play-next)
    (define-key monk-mode-map [S-tab] 'monk-play-prev)

    (define-key monk-mode-map "f" 'monk-play-forward)
    (define-key monk-mode-map "b" 'monk-play-backward)
    (define-key monk-mode-map "\177" 'monk-play-backward)

    (define-key monk-mode-map "." 'monk-goto-current-track)
    (define-key monk-mode-map "\eG" 'monk-goto-track-integer)

    (define-key monk-mode-map "L" 'monk-play-loop-a-track)
    (define-key monk-mode-map "\el" 'monk-loop-clear) ; track/beg/dur IFS
    ;; (define-key monk-mode-map "\eL" 'monk-play-loop-tracks)
    (define-key monk-mode-map "l" 'monk-play-loop) ; track/beg/dur IFS
    (define-key monk-mode-map "\e\C-l" 'monk-play-loop-region)
    (define-key monk-mode-map "s" 'monk-stop)

    (define-key monk-mode-map "q" 'monk-quit)
    (define-key monk-mode-map "G" 'monk-cda-revert)
    ;; (define-key monk-mode-map "\C-x\C-r" 'monk-rebuild-track) ; now "g"
    (define-key monk-mode-map "g" 'monk-rebuild-track)
    (define-key monk-mode-map "!" 'monk-shell-command)

    ;; Marking / Unmarking:
    (define-key monk-mode-map "m" 'monk-mark-track)
    (define-key monk-mode-map "u" 'monk-unmark-track)
    ;; (define-key monk-mode-map "A" 'monk-mark-all-tracks)
    (define-key monk-mode-map "M" 'monk-mark-all-tracks)
    (define-key monk-mode-map "A" 'monk-mark-all-tracks)
    (define-key monk-mode-map "U" 'monk-unmark-all-tracks)
    (define-key monk-mode-map "t" 'monk-toggle-mark)
    (define-key monk-mode-map "T" 'monk-toggle-all-marks)
    (define-key monk-mode-map "\em" 'monk-mark-track-regex)
    (define-key monk-mode-map "%m" 'monk-mark-track-regex)
    (define-key monk-mode-map "\eu" 'monk-unmark-track-regex)

    ;; Marked track operaitions:
    (define-key monk-mode-map "K" 'monk-kill-marked-tracks)
    (define-key monk-mode-map "W" 'monk-copy-marked-tracks)
    (define-key monk-mode-map "D" 'monk-remove-marked-tracks)
    (define-key monk-mode-map "\eL" 'monk-loop-marked-tracks)

    ;; Mouse on marking/unmarking
    (define-key monk-mode-map [double-mouse-3] 'monk-kill-marked-tracks)
    (define-key monk-mode-map [down-mouse-1] 'monk-down-mouse)
    (define-key monk-mode-map [drag-mouse-1] 'monk-mouse-sweep-region)

    ;; These require to modify C-down-mouse-1,2, and it's over-kill.
    ;;  (define-key monk-mode-map [C-mouse-1] 'monk-mouse-yank-track)
    ;;  (define-key monk-mode-map [C-mouse-2] 'monk-mouse-yank-track)
    ;; so, let's use just M-mouse-2.
    (define-key monk-mode-map [M-mouse-2] 'monk-mouse-yank-track)

    ;; C-x C-d is normally bound to the moronic command in standard emacs.
    ;; (define-key monk-mode-map "\C-x\C-d" 'monk-remove-broken-track)
    ;; perhaps you don't need M-g, right ?
    ;; (define-key monk-mode-map "\C-c\C-c" 'monk-remove-broken-track)
    (define-key monk-mode-map "\C-c\C-c" 'monk-rebuild-all-tracks)
    (define-key monk-mode-map "\C-c\C-x" 'monk-remove-broken-tracks)

    ;; Swells up/down in reverse direction of mpg123.  Confused ?
    (define-key monk-mode-map "v" 'monk-volume-up)
    ;; Wrong.  Many people use M-v to scroll backward.  Do this in my .emacs.
    ;; (define-key monk-mode-map "\ev" 'monk-volume-down)
    (define-key monk-mode-map "V" 'monk-volume-down)
    (define-key monk-mode-map [S-down] 'monk-volume-down)
    (define-key monk-mode-map [S-up] 'monk-volume-up)

    ;; Some direct edit
    (define-key monk-mode-map [?\C-_] 'monk-undo)
    (define-key monk-mode-map [?\C-/] 'monk-undo)
    (define-key monk-mode-map "\C-xu" 'monk-undo)
    (define-key monk-mode-map "\C-y" 'monk-yank-track)
    (define-key monk-mode-map "\C-k" 'monk-kill-current-track)
    (define-key monk-mode-map "\C-k" 'monk-kill-current-track)
    (define-key monk-mode-map "w" 'monk-copy-current-track)

    (define-key monk-mode-map "\C-d" 'monk-remove-current-track)
    (define-key monk-mode-map "\C-c\C-k" 'monk-clear-buffer)

    (define-key monk-mode-map "\C-c\C-d" 'monk-show-cddb-info)
    (define-key monk-mode-map "\C-c\C-i" 'monk-show-info)

    (define-key monk-mode-map "d" 'monk-dired-on-this)	; bad binding ?

    (define-key monk-mode-map "\C-cE" 'monk-eject)
    ;; Guess all-CTRL-system for `edit' sounds better...
    ;; (define-key monk-mode-map "\ee" 'monk-edit)
    (define-key monk-mode-map "\C-ce" 'monk-edit)

    ;; monk-prop-edit under construction:
    ;; experimental. Gonna be merged maybe
    (define-key monk-mode-map "\C-c\C-e" 'monk-prop-edit)

    ;; This works also for monk-mode.
    (define-key monk-mode-map "\C-cI" 'monk-write-id3-info)

    (define-key monk-mode-map "\C-c\C-s" 'monk-generate-playlist)
    ;; (define-key monk-mode-map "\C-x\C-w" 'monk-generate-playlist)
    ;; (define-key monk-mode-map "\C-x\C-s" 'monk-generate-playlist)

    (define-key monk-mode-map "\C-o" 'monk-find-file)

    (define-key monk-mode-map "\C-c\C-o" 'monk-open-new)
    (define-key monk-mode-map "\C-co" 'monk-open-new-other-window)
    (define-key monk-mode-map "\C-cO" 'monk-open-new-other-frame)

    (define-key monk-mode-map "-" 'negative-argument)
    (mapcar
     (lambda (ks)
       (define-key monk-mode-map ks 'digit-argument))
     (list "0" "1" "2" "3" "4" "5" "6" "7" "8" "9"))

    ;; Some harmless mouse bindings in menu-bar, and in the buffer.
    ;; See `monk-use-extra-mouse' below for more non-standard ones.
    ;; These `define-key' calls do no harm even if it's not `window-system'.
    (define-key monk-mode-map [menu-bar monk] monk-navigation-menu)
    ;; 3rd arg should be value of `monk-navigation-menu' ?

    (define-key monk-mode-map [mouse-2] 'monk-mouse-play-this-track)
    (define-key monk-mode-map [double-mouse-1] 'monk-mouse-play-this-track)

    (define-key monk-mode-map [C-mouse-4] 'monk-volume-up) ; wheel forward
    (define-key monk-mode-map [C-mouse-5] 'monk-volume-down) ; wheel backward

    ;; Like mode-line prefix...
    (define-key monk-mode-map [M-mouse-4] 'monk-mouse-play-prev)
    (define-key monk-mode-map [M-mouse-5] 'monk-mouse-play-next)
    (define-key monk-mode-map [C-M-mouse-4] 'monk-menu-play-backward)
    (define-key monk-mode-map [C-M-mouse-5] 'monk-menu-play-forward)

    (define-key monk-mode-map "\C-c\C-n" 'monk-next-buffer)
    (define-key monk-mode-map "\C-c\C-p" 'monk-previous-buffer)

    (define-key monk-mode-map [M-S-mouse-4] 'monk-previous-buffer)
    (define-key monk-mode-map [M-S-mouse-5] 'monk-next-buffer)

    monk-mode-map)
  "Key map for `monk-mode'.  See also `monk-use-extra-mouse'.")

;;
;; `monk-define-key-after' -- system's `menu-bar' setup
;;

;; You can change `menu-bar' binding using `monk-define-key-after',
;; but you can't call `monk-define-key-after' without loading `monk',
;; if you want use the binding from the start:
;;
;; (progn
;;   (require 'monk)
;;   (monk-define-key-after "MonK-new" 'monk-open-new 'calendar)
;;   ;;(monk-define-key-after "MonK-o-window" 'monk-other-window 'calendar)
;;   ;;(monk-define-key-after "MonK-o-frame" 'monk-other-frame 'calendar)
;;   )
;;
;; If it's okay that you can't use your binding until `monk' is autoloaded,
;; this is an alternative:
;;
;; (autoload 'monk-open-new "monk" nil t)
;; (add-hook
;;  'monk-load-hook
;;   (lambda () (monk-define-key-after "MonK-open" 'monk-open-new 'calendar)))

(defun monk-define-key-after (menulabel monkfun afterkey &optional keep)
  "Define, with MENULABEL, function MONKFUN after AFTERKEY in `menu-bar'."
  (let ((fmenu (or (lookup-key global-map [menu-bar tools]) ; emacs-20
		   (lookup-key global-map [menu-bar file]))))
    (cond ((or (null keep)		; Only with `keep', do check
	       (null ;; (assq 'monk (cdr fmenu))
		(lookup-key fmenu [monk]) ;; perhaps more compatible than assq
		     ))
	   (define-key-after fmenu [monk] (cons menulabel monkfun) afterkey)
	   (x-popup-menu nil fmenu)))))		; Rebuild just it.

;; Do `menu-bar' binding here in loading this file, but,
;; the user may already defined menu item `monk' with his own hairy procedure,
;; or here he is just reloading `monk' now, so, must pass non-nil
;; as the flag `KEEP'.

(monk-define-key-after "MonK" 'monk 'calendar 'keep-existing-binding)

;;
;; `monk-use-extra-mouse':
;; `monk-mode-map' should have been setup surely when this function is read.
;;
;; To use `extra-mouse', add (in ~/.emacs (or ~/.emacs.el (emacs-20 only))):
;; (add-hook 'monk-load-hook 'monk-use-extra-mouse)
;;

(defvar monk-extra-mouse nil "Internal use only. Controlled by `monk-use-extra-mouse'.")

(defun monk-use-extra-mouse ()
  "Enables mouse commands some more, and popup menu within `*monk*' buffer.
Watch out !  Some of standard mouse binding is disabled for `*monk*' buffer
and its mode-line.  When called variable `monk-extra-mouse' is set to t.
If you dare to use the mouse, include in your ~/.emacs:

\(add-hook 'monk-load-hook 'monk-use-extra-mouse\)"
  (interactive)

  (define-key monk-mode-map [mode-line C-mouse-4] 'monk-menu-play-backward)
  (define-key monk-mode-map [mode-line C-mouse-5] 'monk-menu-play-forward)
  (define-key monk-mode-map [mode-line mouse-4] 'monk-mouse-play-prev)
  (define-key monk-mode-map [mode-line mouse-5] 'monk-mouse-play-next)

  ;; For quick access to the menu, without CTRL key, but deprecated
  (cond
   ((fboundp 'mouse-major-mode-menu)
    (or (fboundp 'mode-line-mode-menu-1) ; E21 has this. We use it.
  	(define-key
  	  monk-mode-map [mode-line down-mouse-3] 'mouse-major-mode-menu))
    ;; (define-key monk-mode-map [down-mouse-3] 'mouse-major-mode-menu)
    ))

  ;; Noisy:
  ;; (define-key monk-mode-map [double-down-mouse-3] 'mouse-major-mode-menu)
  ;; (define-key monk-mode-map [mouse-3] 'mouse-major-mode-menu)

  ;; `Stacked' binding won't work with mode-line prefix --- perhaps
  ;; `mode-line' is prefix --- like numeric arg is not passed correctly
  ;; to `stacked' command even if you pushed back the event into queque.
  ;; Let monk-mode-line-down-mouse() handle all the job.
  ;;
  ;; You may want to use
  ;;    [mode-line mouse-2] (mouse-delete-other-windows)
  ;; for `*monk*' buffer, but I am sorry
  ;; down-mouse-2 and C-down-mouse-2 are handled by the same commands:

  ;; Shuttle search by dragging, and jump by click:
  (define-key monk-mode-map
    [mode-line down-mouse-2] 'monk-mode-line-down-mouse)
  ;; Loop by dragging:
  (define-key monk-mode-map
    [mode-line C-down-mouse-2] 'monk-mode-line-down-mouse)

  ;; Backup popular (seems like) commands (works, but ding()s. Why ?):
  (define-key monk-mode-map [mode-line S-mouse-2] 'mouse-delete-other-windows)
  (define-key monk-mode-map [mode-line M-mouse-3] 'mouse-delete-window)
  (setq monk-extra-mouse t)
  )

;;
;; Support keyboard command in `dired'
;;

;; Fri Jan 19 11:26:58 2001 s.n.

;; Howto use monk-dired keyboard commands:
;; 
;; (1) Use `dired-load-hook',
;; (autoload  'monk-dired-bind-extra-keys "monk" nil t)
;; (add-hook 'dired-load-hook 'monk-dired-bind-extra-keys)
;;
;; (2) use `eval-after-load'
;; (autoload  'monk-dired-bind-extra-keys "monk" nil t)
;; (if (string-match "^19.28" emacs-version)
;;    (eval-after-load "dired" (monk-dired-bind-extra-keys))
;;  (eval-after-load "dired" (quote (monk-dired-bind-extra-keys))))
;;
;; (3) or just load them up (if you can't live without).
;; (require 'dired)
;; (require 'monk)
;; (monk-dired-bind-extra-keys)

(defsubst monk-dired-map-over-marked-files (fun arg)
  "Run FUN on marked files in dired buffer.
ARG is passed to `dired-get-marked-files'."
  (or (eq major-mode 'dired-mode) (error
				   "You must be in dired mode to do that"))
  (let ((fn-list (dired-get-marked-files nil arg)))
    (mapcar (lambda (fn) (funcall fun fn)) fn-list)))

(defun monk-dired-do-monk-find-file (arg)
  "Run `monk-find-file' over filenames generated by `dired-get-marked-files'.
Guess ARG works as in `dired-do-shell-command'."
  (interactive "P")
  (monk-dired-map-over-marked-files (function monk-find-file) arg))

(defvar monk-dired-monk-command (function monk)
  "*Command invoked by `monk-dired-do-monk'.")

(defun monk-dired-do-monk (arg)
  "Run MonK command over filenames generated by `dired-get-marked-files'.
The MonK command is the one registered in `monk-dired-monk-command'.
Guess ARG works as in `dired-do-shell-command'."
  (interactive "P")
  (monk-dired-map-over-marked-files monk-dired-monk-command arg))

(defun monk-dired-bind-extra-keys ()
  (define-key dired-mode-map "\e\C-m" 'monk-dired-do-monk-find-file)
  (define-key dired-mode-map "\em" 'monk-dired-do-monk))

;;
;; `monk-disable-mode-line-switches-21'
;; `monk-enable-mode-line-switches-21'
;;
;; emacs-21 has `fancy' (o really ?) bindings on the certain spots
;; of mode-line.  We must disable, or replace them.
;;

(when (fboundp 'mode-line-mode-menu-1)
  (defun monk-mode-line-mode-menu-1 (event)
    "Raise entity of `monk-navigation-menu', and accept input."
    (interactive "e")
    (save-window-excursion		; good enough
      (select-window (posn-window (event-start event)))
      (let* ((men (cdr monk-navigation-menu))
	     (input (x-popup-menu t men))
	     (fun (lookup-key men (apply 'vector input))))
	(when (commandp fun) (command-execute fun)))))

  (defun monk-disable-mode-line-switches-21 ()
    "Turn off the silly buttons and menus on mode-line in emacs-21."
    (setq mode-line-modified (list "%1*" "%1+")) ; goood enough. buffer-local.
    ;;(set (make-local-variable 'mode-line-minor-mode-keymap) nil)
    (set (make-local-variable 'mode-line-minor-mode-keymap)
	  '(keymap
	    ;; What's this `header-line' item. Who use this ?
;;;	    (header-line keymap
;;;			 (down-mouse-2 . monk-mode-line-down-mouse)
;;;			 (C-down-mouse-2 . monk-mode-line-down-mouse)
;;;			 (down-mouse-3 . monk-mode-line-mode-menu-1))
	    (mode-line keymap
		       (down-mouse-2 . monk-mode-line-down-mouse)
		       (C-down-mouse-2 . monk-mode-line-down-mouse)
		       (down-mouse-3 . monk-mode-line-mode-menu-1)
		       ;; Fri Mar 28 15:27:49 2003 added button-4/5
		       (C-mouse-4 . monk-menu-play-backward)
		       (C-mouse-5 . monk-menu-play-forward)
		       (mouse-4 . monk-mouse-play-prev)
		       (mouse-5 . monk-mouse-play-next)
		       )))
    ;; Sat Sep 29 15:54:42 2001
    ;; To make dragable and clickable mode-line, turn off mouse-3 binding
    ;; on coding-system display at the top of mode-line.
    ;; `mode-line-mule-info' is in bindings.el.
    (when (featurep 'mule)
      (set (make-local-variable 'mode-line-mule-info) nil)))

  (defun monk-enable-mode-line-switches-21 ()
    "Turn on the buttons and menu on mode-line in emacs-21."
    (let ((buf (get-buffer-create monk-junk-buffer))
	  m-l-m-system m-l-m-m-k-system mode-l-m-i)
      (save-excursion (set-buffer buf)
		      (setq m-l-m-system mode-line-modified
			    m-l-m-m-k-system mode-line-minor-mode-keymap
			    mode-l-m-i mode-line-mule-info))
      (setq mode-line-modified m-l-m-system
	    mode-line-minor-mode-keymap m-l-m-m-k-system
	    mode-line-mule-info mode-l-m-i))))

;;
;; `monk-mode'
;;

(defvar monk-buffer-list nil		; Must be global.
 "*monk* buffer list \
containing MonK buffers created by `monk-generate-new-monk-buffer'.")

(defvar monk-mode-hook nil "*Hook called at the end of `monk-mode'.")
(defvar monk-cdparanoia-probe-options (list "-sQq") "*CD detect option.")
(defvar monk-dont-care-no-CD t
  "*Non-nil forces to play CD without any media detection.
Defaults t, but you can set to `nil' if you have cdparanoia\(1\)
to check media existense in the drive via cdparanoia.")

(defvar monk-truncate-lines t		; you like it ?
  "*Set to nil if you don't want to truncate lines in *monk* buffer.")

;; `font-lock' fiddlings -- defaults

(defconst monk-font-lock-keywords
  ;; these two exist even in E19.28
  '(("\\(\".+\"\\)$" . font-lock-string-face)
    ;; ("\\(`.+'\\)" . font-lock-type-face) ; in E20 only.
    ("\\(`.+'\\)$" . font-lock-keyword-face))
  "Faces used in MonK buffers")

(defun monk-font-lock-setup ()
  "Turns on `font-lock-mode'.  Face is setup by `monk-font-lock-keywords'."
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(monk-font-lock-keywords t))
  ;; (turn-on-font-lock) ; This does not exist in emacs-19.28.
  (font-lock-mode 1))

(defun monk-generate-new-monk-buffer (bname)
  "`generate-new-buffer' compatible to create *monk* buffer."
  (let ((buf (generate-new-buffer bname)))
    (setq monk-buffer-list (append monk-buffer-list (list buf)))
    buf))

;; kill-buffer-hook

(defun monk-buffer-list-cleanup ()
  "Remove current buffer from monk-buffer-list.
For `kill-buffer-hook'."
  (when (memq (current-buffer) monk-buffer-list)
    (setq monk-buffer-list (delq (current-buffer) monk-buffer-list))))

(add-hook 'kill-buffer-hook 'monk-buffer-list-cleanup)

(defun monk-walk-on-ring (lis arg current &optional reverse)
  "Take a list LIS as ring, and return ARGth element from CURRENT element.
Optional 4th arg REVERSE specifies tracing the ring in reverse direction."
  (let* ((lis (if reverse (reverse lis) lis))
	 (whole lis)			; keep orig.
	 (lis (member current lis)))
    (while (> arg 0)
      (setq lis (cdr lis))
      (when (null lis) (setq lis whole))
      (setq arg (1- arg)))
    (car lis)))

(defun monk-mode ()
  "A mode to navigate in the sea of multimedia obejects.
Fired up in *monk* buffer by command `monk'.

Available key bindings:

\\{monk-mode-map}"
;;; You can freeze emacs on C-h m (describe-mode)
;;; by putting "." after the keymap citation like this:
;;; "\\{monk-mode-map}. " at the end of docstring.
;;; Perhaps help-mode-finish() uses forward-paragraph() ?
  (interactive)
  ;; This kill disables all local variable, so advancing to next
  ;; track is disabled.  What is the demerit without
  ;; this (kill-all-local-variables) ?  -- perhaps no harm for MonK
  (or (eq major-mode 'monk-edit-mode)
      (kill-all-local-variables))

  (setq truncate-lines monk-truncate-lines)
  (set (make-local-variable 'revert-buffer-function)
       'monk-revert-buffer-function)

  (use-local-map monk-mode-map)
  (setq major-mode 'monk-mode)
  (setq mode-name "MonK")
  ;; (and window-system (font-lock-mode t)) ; Cheap.
  (when window-system (monk-font-lock-setup))
  (setq buffer-read-only t)
  ;; Normally `monk-use-extra-mouse' is called from `monk-load-hook',
  ;; and it sets `monk-extra-mouse' to t.
  (when (and monk-extra-mouse
	     (fboundp 'monk-disable-mode-line-switches-21))
    (monk-disable-mode-line-switches-21))
  (setq goal-column 36)			; Cheap.
  (monk-face-setup)			; New. Unconditional.
  (run-hooks 'monk-mode-hook))

;; This requires `cdparanoia'.  How can I probe disc with `cdda2wav' ?
(defsubst monk-disc-inserted-p ()
  "Cheap. May stuck if device is busy."	; FIXIT
  (cond
   ((or monk-dont-care-no-CD
	(= ;; (monk-with-coding-system ;; unused
	 (apply 'call-process "cdparanoia"
		nil nil nil monk-cdparanoia-probe-options) 0)))
   (t (error))))

(defun monk (file &optional arg)
  "Open up MonK buffer.

If just Return key is hit for the minibuffer interface,
tracks are read in from inserted audio CD.

Other pathname input causes MonK to read the file or directory via
`monk-find-file'.  Just add \".\" after suggested pathname to read
current directory.

With numeric ARG 0, an empty, or already existing MonK buffer with `monk-mode'
is read.  See also `monk-mode', `monk-mode-map', and `monk-use-extra-mouse'.
Other non-nil value for optional 2nd ARG reverses the value of
`monk-find-file-recursively' temporally \(see `monk-file-file'\)."
  (interactive
   (list
    (cond
     ((= 0 (prefix-numeric-value
	    current-prefix-arg)) nil) ; IF 0 arg, no minibuf ifs.
     (t (read-file-name
	 "File (just RET opens cd audio): "
	 nil				; DIR `default-directory' is used
	 "/dev/cdrom"			; DEFAULT-INPUT used if just RET is hit
;;;	 "___MonK_JUNK___"		; DEFAULT-INPUT used if just RET is hit
	 t				; MUST-MATCH to existing pathname
	 nil)))				; INITIAL input
    current-prefix-arg))
  (monk-subr file arg 'switch-to-buffer))

(defun monk-other-window (file &optional arg)
  "Open other-window version of `monk'.  See command `monk'."
  (interactive
   (list
    (cond
     ((= 0 (prefix-numeric-value
	    current-prefix-arg)) nil)	; IF 0 arg, no minibuf ifs.
     (t (read-file-name
	 "File (just RET opens cd audio): "
	 nil				; DIR `default-directory' is used
	 "/dev/cdrom"			; DEFAULT-INPUT used if just RET is hit
;;;	 "___MonK_JUNK___"		; DEFAULT-INPUT used if just RET is hit
	 t				; MUST-MATCH to existing pathname
	 nil)))				; INITIAL input
    current-prefix-arg))
  (monk-subr file arg 'switch-to-buffer-other-window))

(defun monk-other-frame (file &optional arg)
  "Open other-window version of `monk'.  See command `monk'."
  (interactive
   (list
    (cond
     ((= 0 (prefix-numeric-value
	    current-prefix-arg)) nil)	; IF 0 arg, no minibuf ifs.
     (t (read-file-name
	 "File (just RET opens cd audio): "
	 nil				; DIR `default-directory' is used
	 "/dev/cdrom"			; DEFAULT-INPUT used if just RET is hit
;;;	 "___MonK_JUNK___"		; DEFAULT-INPUT used if just RET is hit
	 t				; MUST-MATCH to existing pathname
	 nil)))				; INITIAL input
    current-prefix-arg))
  (monk-subr file arg 'switch-to-buffer-other-frame))

(defun monk-subr (file arg fun)
  "Core of `monk', `monk-other-window', and `monk-other-frame'.
Uses FUN to switch to *monk* buffer. Current buffer object is returned."
  (let ((sz)
	(buf ;; (generate-new-buffer monk-buffer-name)
	     (monk-generate-new-monk-buffer monk-buffer-name)
	     ))
    ;; To `monk-find-file' work.  Don't know if this is right thing to do...
    (or (monk-buffer-live-p monk-buffer) (setq monk-buffer buf))
    (funcall fun buf)
    (monk-mode)
    (setq sz (buffer-size))
    (cond ((= 0 (prefix-numeric-value arg)) t) ; just open buffer
	  ((or (null file)		; non-interactive call use this nil
	       ;; (string-equal "___MonK_JUNK___" file)
	       (string-equal "/dev/cdrom" file))
	   (monk-cda-read-tracks))
	  ;; Use `buf' as `monk-buffer' here,
	  ;; otherwise file(s) goes into existing old `monk-buffer'.
	  ;; any other non-nil arg toggles the `recursive' flag
	  ;; 3rd arg t for `monk-find-file' forces to cd if file is directory.
	  (arg (let ((monk-buffer buf)) (monk-find-file file t t)))
	  (t  (let ((monk-buffer buf)) (monk-find-file file nil t))))
    (cond ((= 0 sz) (goto-char (point-min))))
    (current-buffer)))

;;(defun monk-new-buffer ()
(defun monk-open-new ()			; C-c C-o ?
  "Open new emty monk buffer."
  (interactive)
  (monk "" 0))

(defun monk-open-new-other-window ()
  "Open new emty monk buffer in other window." ; C-c o ; frequently used ?
  (interactive)
  (monk-other-window "" 0))

(defun monk-open-new-other-frame ()
  "Open new emty monk buffer in other frame." ; C-c O ?
  (interactive)
  (monk-other-frame "" 0))

(defun monk-switch-to-buffer (b)
  "`switch-to-buffer' to buffer B, but uses B's window is used if available."
  (let ((win (get-buffer-window b)))
    (if win (select-window win) (switch-to-buffer b))))

(defun monk-next-buffer (arg)
  "Switch to ARGth next MonK buffer."
  (interactive "p")
  (cond ((> arg 0)
	 (let ((b (monk-walk-on-ring monk-buffer-list arg (current-buffer))))
	   (cond ((monk-buffer-live-p b) (monk-switch-to-buffer b))
		 (t (error "Can't switch to buffer %S")))))
	(t (monk-previous-buffer (- arg)))))

(defun monk-previous-buffer (arg)
  "Switch to ARGth previous MonK buffer."
  (interactive "p")
  (cond ((> arg 0)
	 (let ((b (monk-walk-on-ring monk-buffer-list arg (current-buffer) 'reverse)))
	   (cond ((monk-buffer-live-p b) (monk-switch-to-buffer b))
		 (t (error "Can't switch to buffer %S")))))
	(t (monk-next-buffer (- arg)))))

;; `monk-no-cda-prcess-p' -- now uses current-buffer context only.

(defsubst monk-no-cda-prcess-p ()
  "Check if cda audio process is *not* running."
  (or (null (monk-proc-running-p))	; cheap check
      ;; This check is done when procrunning.
      ;; `monk-current-track' is current buffers'.
      ;; (monk-with-monk-buffer
      ;; (monk-with-monk-window-and-buffer
      (and monk-current-track
	   ;; (listp monk-current-track)
	   (not (eq 'cda (monk-assq-get-value 'type monk-current-track)))
	   ))) ; else nil

;; The first, and only generic function name for cd-audio:
(defun monk-cda-read-tracks ()
  "Insert track lines for cd audio.
*monk* buffer should have already been created."
  (cond
   (;;(null (monk-proc-running-p))	; cheap
    (monk-no-cda-prcess-p)		; now check this
    (let (buffer-read-only (sz (buffer-size)))
      (goto-char (point-max))
      (condition-case nil
	  (monk-disc-inserted-p)
	(error
	 (error "Seems like CD is not in drive, sister...")))
      ;; Setup `monk-disc-info', `monk-track-list' etc.
      (monk-get-info)
      ;; Get cddb data, if required (or available)
      (monk-fetch-cddb-data (monk-cddb-build-query-param))
      (monk-insert-track-lines)
      (cond ((= 0 sz) (goto-char (point-min))))))
   (t (kill-buffer (current-buffer))
      (error "cd audio is playing. Stop it first"))))

;;
;; Generic track (mp3, cda, mpg, mid, etc.) insert
;;

(defvar monk-track-generator-list
  (list '("\\.mp3$" monk-mp3-insert-track-obj monk-mp3)
	'("\\.m3u$" monk-m3u-find-file monk-mp3)
	'("\\.mpe*g$" monk-smpeg-insert-track-obj monk-smpeg)
	;; '("\\.midi*$" monk-playmidi-insert-track-obj monk-midi)
	'("\\.midi*$" monk-timidity-insert-track-obj monk-midi)
	;; midi.gz, mid.gz
	'("\\.midi*.gz$" monk-timidity-insert-track-obj monk-midi)
	'("\\.wav$" monk-wav-insert-track-obj monk-wav)
	'("\\.ogg$" monk-ogg-insert-track-obj monk-ogg)

	)
  "*A list of \(REGEX HANDLER FEATURE\), wherein

REGEX:   Regex for filename extension such as .mp3 etc.
HANDLER: Symbol as handler function
FEATURE: The feature should be `require'd to use the HANDLER

Normally, HANDLER function is defined in FEATURE.el which
`provide's the FEATURE.  The REGEX match is done in `monk-get-track-generator'
with non-nil `case-fold-search'.")

(defvar monk-find-file-recursively t
  "*Defaults t which causes `monk-find-file' to find files recursively.
This default is the same as that of xmms when it read a directory.
Set to nil if your directories are too deep or contain too many files.")

(defvar monk-readin-counter 0 "Mostly for cosmetics. Internal use only.")

(defvar monk-readin-redisplay-every 5
  "*MonK does redisplay every this number of files have been readin.
Increase, or set to nil if your vehicle is slow, or you always readin
big directories.")

(defsubst monk-readin-redisplay (&optional reset)
  "Take control on track-readin redisplay, or just do it.
If optional arg RESET is non-nil, just initialize `monk-readin-counter' to 0.
Otherwise actual redisplay is done according to `monk-readin-counter' and
`monk-readin-redisplay-every'."
  (cond (reset
	 (if monk-readin-redisplay-every (setq monk-readin-counter 0)))
	(t (cond (monk-readin-redisplay-every
		  (setq monk-readin-counter (1+ monk-readin-counter))
		  (if (= 0 (% monk-readin-counter
			      monk-readin-redisplay-every))
		      (sit-for 0)))))))

(defsubst monk-get-track-generator (file)
  "Return track generator function for FILE by extension matching.
If the file extention is matched, the `require'd feature is loaded.
See also `monk-track-generator-list'."
  (catch 'found
    (mapcar (lambda (data)
	      (let ((rgx (car data)))
		(cond ((let ((case-fold-search t)) (string-match rgx file))
		       (require (nth 2 data))
		       (throw 'found (nth 1 data))))))
	    monk-track-generator-list)
    nil)) ; nil yielded unless 'found

;;
;; C-o (monk-find-file) --- can also be called even outside of *monk*
;;

;; (fset 'monk-open-file 'monk-find-file)

(defun monk-find-file (file &optional arg fakeinteractive)
  "Insert FILE as track in MonK buffer.
The track inserting handler is selected via `monk-get-track-generator'.

If optional 2nd ARG is non-nil, the value of `monk-find-file-recursively'
is temporally toggleded while reading tracks.  If FILE is directory,
and called interactively from the top level, and `cd' to the directory.
Non-nil optional 3rd arg FAKEINTERACTIVE forces the `cd' operation."
  (interactive "fOpen File: \nP")
  (or (eq major-mode 'monk-mode)   ; eg., *scratch* in monk-mode maybe
      (or (monk-buffer-live-p monk-buffer) (monk "" 0))) ; Get *monk*
  ;; Now at least one *monk* buffer is available.
  (let ((monk-buffer
	 ;; If called another *monk**, use current buffer.
	 (cond ((and (null (eq (current-buffer) monk-buffer))
		     (eq major-mode 'monk-mode))
		(current-buffer))
	       (t monk-buffer))))
    ;; (progn
    (monk-with-coding-system
     (let ((file (expand-file-name file)) handler)
       (cond
	((file-directory-p file)
	 (and (or (interactive-p) fakeinteractive) (cd file))
	 (monk-find-directory file arg)) ; Recurrs to this...
	;; Else take just one file.
	(t
	 (cond
	  ;; file extension is tested here
	  ((setq handler (monk-get-track-generator file))
	   ;; Now use current-buffer
	   (cond
	    ((file-readable-p (expand-file-name file))

	     (cond
	      ((= 0 (nth 7 (file-attributes file)))
	       (message "Can't 0-sized file %s" file))
	      ;; Here we should not use `save-excursion' to
	      ;; insert file in `directory-files' order.
	      ;; (monk-with-monk-window-or-buffer  ;; Okay (maybe) 
	      ;; (monk-with-monk-window-and-buffer ;; Wrong
	      ;; So we use monk-with `monk-with-monk-window':
	      (t (monk-with-monk-window
		  (let (buffer-read-only)
		    (monk-with-no-face ; buggy move-overlay() support.
		     (forward-line 0)
		     ;; Call the the found track generation handler
		     (cond
		      ((funcall handler file)
		       (monk-readin-redisplay nil)
		       (set-buffer-modified-p nil)
		       (message "MonK read file \"%s\"" file) t)
		      (t (message "Probe failed for file %s" file) nil))))))))
	    (t (message "Can't read file %s" file) nil)))
	  (t (message "Can't find handler for %s" file)))))))))

(defun monk-find-directory (dir arg)
  "Open a DIR flat, or recursively if `monk-find-file-recursively'.
Call `pushe-mark' before inserting . ARG reverses
`monk-find-file-recursively' to be used."
  (cond ((file-accessible-directory-p dir)
	 ;; Barfs if the tree is so deep when `monk-find-file-recursively'
	 (let* ((default-directory (monk-put-trailing-/ dir))
		(files (delete
			"." (delete
			     ".."
			     (directory-files dir))))
		(monk-find-file-recursively
		 (if arg (not monk-find-file-recursively)
		   monk-find-file-recursively)))
	   (monk-readin-redisplay 'reset)
	   (or monk-find-file-recursively
	       ;; otherwise squeeze it to non-directory-only list
	       (setq files
		     (delete t (mapcar
				(lambda (o)
				  (or (file-directory-p o) o)) files))))
	   (push-mark (point))
	   (mapcar 'monk-find-file files)))
	(t (message "Can't read directory %s" dir))))

;;
;; Direct edit (in monk-mode)
;;

(fset 'monk-clear-buffer 'monk-erase-buffer)
(defun monk-erase-buffer ()
  "Call `erase-buffer' with turning off `buffer-read-only'.
Current playing is not stopped."
  (interactive)				; not "*P"
  (monk-only-with-monk-modes
   (let (buffer-read-only)
     ;; (monk-stop) ;; not stopped.
     (monk-mode-line-print "%b")
     (erase-buffer))))

(defun monk-undo (arg)
  "Call `undo' with turning off `buffer-read-only'."
  (interactive "P")			; not "*P"
  (monk-only-with-monk-modes
   (let (buffer-read-only)
     (undo arg))))

(defun monk-yank-track (arg)
  "Call `yank' with turning off `buffer-read-only'."
  (interactive "p")			; not "*p"
  (monk-only-with-monk-modes
   (let (buffer-read-only)
     (forward-line 0)
     (yank arg))))

(defun monk-kill-current-track (&optional arg append)
  "Kill current track from *monk* buffer.
The killed tracks are pushed into `kill-ring', and can be yanked back later.

Optional 1st ARG \(numeric arg when interactive\) specifies number of
tracks to be killed.  When optional 2nd arg APPEND is non-nil,
the killed tracks are appended on the top of the `kill-ring'.

Killed tracks can be yanked back with `monk-yank-track'.
Current playing is not stopped even if the playing track is killed.

Variable `this-command' is faked as being `kill-line'.

The relevant file won't be deleted in file system.
Use \\[monk-dired-on-this] \(`monk-dired-on-this'\) \
if you really want to delete the file."
  (interactive "p")
  (monk-only-with-monk-modes
   (let (buffer-read-only)
     (forward-line 0)
     (when append (append-next-kill))
     (prog1 (kill-line arg)
       (setq this-command 'kill-line)))))

(defun monk-copy-current-track (&optional arg append)
  "Copy current track to `kill-ring'.
Arguments ARG and APPEND works as in `monk-kill-current-track' \(which see\)."
  (interactive "p")
  (let (2nd)
    (save-excursion
      (monk-only-with-monk-modes
       (monk-do-nlines
	arg
	(when (or 2nd append) (append-next-kill))
	(prog1
	    (copy-region-as-kill (save-excursion (beginning-of-line) (point))
				 (save-excursion (end-of-line) (1+ (point))))
	  (setq 2nd t
		this-command 'copy-region-as-kill)))))))

(defun monk-remove-current-track (&optional arg)
  "Remove current track from *monk* buffer.
The removed track is not `kill'ed, and no way to restore it
except \\[monk-undo] \(`monk-undo'\).
Optional ARG specifies the number of tracks to be removed.

Current playing is not stopped even if the playing track is removed.

The relevant file won't be deleted in file system.
Use \\[monk-dired-on-this] \(`monk-dired-on-this'\) \
if you really want to delete the file."
  (interactive "p")
  (monk-only-with-monk-modes
   (let (buffer-read-only)
     (while (> arg 0)
       (delete-region
	(save-excursion (beginning-of-line) (point))
	(save-excursion (end-of-line) (1+ (point))))
       (setq arg (1- arg))))))

;;
;; Rebuild, check, or remove tracks
;;

(defun monk-remove-broken-tracks (arg)
  "Remove broken \(eg. the relevant file does not exist\) tracks.
If ARG is not nil, all the non-broken tracks are rebuild as well.
See also `monk-rebuild-all-tracks'."
  (interactive "P")
  (monk-rebuild-all-tracks (if arg t 0)))

;; The old version was based on the old `monk-remove-broken-track'.

;; (defun monk-remove-broken-track (arg)
(defun monk-rebuild-all-tracks (arg)
  "Rebuild all tracks associated with a multimedia file.  
Each track is handled by `monk-rebuild-track', and the behaviour can
be controlled by ARG which is passed to `monk-rebuild-track'."
  (interactive "P")
  (let ((cc (current-column)))
    ;; Don't know why, staying on BOL moves cursor to wrong position
    ;; on 2nd C-u C-c C-c hit. So, do move-to-column().
    (when goal-column (move-to-column goal-column))
    (monk-readin-redisplay 'reset)

    ;; This macro is not adequate here, because can't handle on/off control of
    ;; `forward-line' (See the long version above, or in monk-purged.el):
    ;; (monk-walk-over-tracks
    (save-excursion
      (goto-char (point-min))
      (while (not (eobp))
	;; If track is removed (nil returned) must not move.
	(when (monk-rebuild-track arg)
	  (forward-line 1))
	(monk-readin-redisplay nil)))
    ;; `move-to-column' does not raise error, but ought to be
    (condition-case nil
	(move-to-column cc)
      (error ()))
    nil))

;; Thu Dec 28 22:16:10 2000
;; Now `monk-rebuild-track' is experimentally called according to
;; `monk-rebuild-track-auto' settings.
;;
;; There's `monk-revert'for cdaudio, and the rebuild/revert/reconstruction
;; system sounds hairy...

(defsubst monk-play-take-control-again (track)
  "Re-calibrate playing control parameters again using TRACK.
No use if called while the track is not playing.
Called from `monk-rebuild-track'."
  ;; `monk-current-loop-name', `monk-current-track-name' are okay.
  (setq monk-fps (monk-assq-get-value 'fps track)
	monk-current-track track
	monk-current-track-type (monk-assq-get-value 'type track)
	;; monk-current-start-sec (/ offset monk-fps) ;; can't
	monk-current-track-length (monk-assq-get-value 'frames track)
	monk-current-cent (monk-assq-get-value 'cent track)))

(defun monk-rebuild-track (&optional arg)
  "Reconstruct current track associated with a multimedia file.
The behaviour is controlled by ARG:

nil ARG:     Rebuild normal track, or just warn for inaccessible file.
ARG < 1:     Just remove broken track. No rebuild.
non-nil ARG: Rebuild normal track and remove broken track.

Returns t if rebuilding succeeded, else nil.

In emacs-19.28, re-facing playing track is not supported here.
Why not throw away the goddamn old emacs, will ya ?"
  (interactive "P")
  (monk-only-with-monk-modes
   ;; Can't use `save-excursion' here.
   ;; Moving cursor to BOL is the required cost.
					;(save-excursion
   (forward-line 0)
   ;; not `monk-get-this-track', for speeding
   (let* ((prop (monk-get-track 'raw))
	  (buffer-read-only) (file)
	  (on-playing-track
	   ;; (monk-within-overlay-p)
	   (and (monk-proc-running-p)
		;; (equal prop monk-current-track)
		(monk-on-playing-track-p monk-current-track prop)
		)))
     (monk-with-coding-system
      (cond
       ((and prop (listp prop)
	     (setq file (monk-assq-get-value 'file prop)))
	;; Do if it has `file' property -- NOP for 'cda tracks
	(cond
	 ((null (file-readable-p file))
	  (cond
	   (arg (monk-remove-current-track 1)
		(set-buffer-modified-p nil)
		(message
		 "MonK removed track \"%s\" for MonK can't read it" file)
		nil)
	   (t (message "File \"%s\" is not readable" file) t)))
	 ((<= (nth 7 (file-attributes file)) 0)
	  (cond
	   (arg
	    (monk-remove-current-track 1)
	    (set-buffer-modified-p nil)
	    (message "MonK removed track \"%s\" for it has size zero" file)
	    nil)
	   (t (message "File \"%s\" has size zero" file) t)))
	 ;; Else the file is accessible hereunder.
	 ((and arg (< (prefix-numeric-value arg) 1)) t)
	 (t
	  ;; monk-with-no-face: no use, in E20/21.
					;(monk-with-no-face 
	  ;; Removal and a new track-content insert
	  ;; destroys the `monk-process-marker' in any way,
	  ;; so if we inserted a new track-content, relocate
	  ;; `monk-process-marker'.
	  (let ((handler (monk-get-track-generator file))
		(marked (monk-looking-at-mark-p))
		track)
	    (monk-remove-current-track 1)
	    (cond
	     ((and
	       handler
	       (setq track (save-excursion (funcall handler file))))
	      (when on-playing-track
		;; RELOCATE `monk-process-marker'.
		(monk-set-process-marker)
		(monk-play-take-control-again track)
		;; Do not support re-face in emacs-19.28.
		(or monk-buggy-make-overlay
		    (monk-face-current-track))
		)
	      (when marked (save-excursion (monk-mark-track 1)))
	      (set-buffer-modified-p nil)
	      (when (interactive-p)
		(message "MonK read a track from \"%s\"" file))
	      t)
	     (t
	      (message
	       "%s \"%s\"" (cond (handler (format "%S failed for" handler))
				 (t "No handler for")) file)
		nil )))
					;)
	  ))))))
					;)
   ))

(defun monk-toggle-rebuild-track-auto ()
  "Toggle the value of `monk-rebuild-track-auto'."
  (interactive)
  (setq monk-rebuild-track-auto
	(not monk-rebuild-track-auto))
  (message "Auto track reconstruction is now %s"
	   (cond (monk-rebuild-track-auto "*enabled*")
		 (t "*disabled*"))))

;;
;; `monk-revert-buffer-function' --- To be bound to `revert-buffer-function'
;; A replacement of `monk-revert-buffer-function' defined in monk.el
;; If this file is merged into the MonK archive,
;; the old one in monk.el should be deprecated.
;;
;; Still cursor sometimes goes wrong position when called via `auto-revert-mode',
;; but far better than the old version below.
;;

(defun monk-revert-buffer-function (junk1 junk2)
  "A function to be bound to `revert-buffer-function'.
To make `auto-revert-mode' work.
All track is rebuilt via `monk-watch-check-directory', but only the existing
tracks are rebuilt."
  (set-buffer-modified-p nil)
  ;;(monk-rebuild-all-tracks t)
  (monk-watch-check-directory 'justrevert) )	; in monk-watch.el

;; Old version. Cursor pos. ctrl is not quite right when called via
;; `auto-revert-mode'.
;; (defun monk-revert-buffer-function (junk1 junk2)
;;   "A function to be bound to `revert-buffer-function'.
;; To make `auto-revert-mode' work. Very rough.
;; All track is rebuilt unconditionally."
;;   (set-buffer-modified-p nil)
;;   (monk-rebuild-all-tracks t))

;;
;; Sorting track (merged monk-sort.el)
;;

;; monk-sort.el --- sort tracks.

;; Going back, and re-facing current playing track is not supported yet:
;; The `sort-fields', and `sort-numeric-fields' (sort.el)
;; destroy `monk-process-marker' (`save-excursion' tried in vain).
;; Maybe hash or checksum should be generated to go back to on
;; current playing track (?)

;; Maybe `sort-by-mp3-title', or `sort-by-filename' is needed (but not yet).

;; Mon Mar  5 03:39:44 2001

;;(require 'monk)

(defsubst monk-sort-by-field (field &optional reverse)
  "Do lexical sorting tracks by FIELD via `sort-fields'.
Optional arg REVERSE forces `reverse-region' after sorting."
  monk-process-marker
  (monk-only-with-monk-modes
   (let (buffer-read-only)
       (sort-fields field (point-min) (point-max))
       (when reverse (reverse-region (point-min) (point-max))))))

(defsubst monk-sort-by-numeric-field (field &optional reverse)
  "Do numeric sorting tracks by FIELD via `sort-numeric-fields'.
Optional arg REVERSE forces `reverse-region' after sorting."
  (monk-only-with-monk-modes
   (let (buffer-read-only)
     (sort-numeric-fields field (point-min) (point-max))
     (when reverse (reverse-region (point-min) (point-max))))))

(defun monk-sort-tracks-by-name (reverse)
  "Do sorting tracks by name.
Optional non-nil arg REVERSE forces reverse sorting."
  (interactive "P")
  (monk-sort-by-field 5 reverse))

;; `monk-sort-tracks-by-bps' makes sense only for mp3/ogg tracks.

(defun monk-sort-tracks-by-bps (reverse)
  "Do sorting tracks by name.
Optional non-nil arg REVERSE forces reverse sorting."
  (interactive "P")
  (monk-sort-by-field 3 reverse))

;; `monk-sort-tracks-by-frames' perhaps makes no sense for `cda' tracks.

(defun monk-sort-tracks-by-frames (reverse)
  "Do sorting tracks by frames.
Optional non-nil arg REVERSE forces reverse sorting."
  (interactive "P")
  (monk-sort-by-numeric-field 4 reverse))

(defun monk-sort-tracks-by-time (reverse)
  "Do sorting tracks by time.
Optional non-nil arg REVERSE forces reverse sorting."
  (interactive "P")
  (monk-sort-by-field 2 reverse))

(defun monk-sort-tracks-by-type (reverse)
  "Do sorting tracks by type.
Optional non-nil arg REVERSE forces reverse sorting."
  (interactive "P")
  (monk-sort-by-field 1 reverse))

;;
;; Marking / Unmarking tracks
;;

(defvar monk-mark-char "*" "*The simple track mark in MonK.")

(defsubst monk-looking-at-mark-p (&optional mark)
  "Detect MARK \(optional arg\) or `monk-mark-char' on point."
  (save-excursion
    (forward-line 0)
    (looking-at (concat "\\" (or mark monk-mark-char)))))

(defsubst monk-insert-mark (&optional mark)
  "Insert MARK \(optional arg\) or `monk-mark-char' on point."
  ;; Insert at (BOL+1), then backward-delete-char.
  ;; This retains facing even on the playing track.
  (let (buffer-read-only)
    (save-excursion
      (forward-line 0)
      (forward-char 1)
      (insert (or mark monk-mark-char))
      (backward-char 1)
      (backward-delete-char 1)
      t)))

(defmacro monk-walk-over-marked-tracks (&rest body)
  "Do BODY on each marked track in MonK buffer."
  (` (monk-walk-over-tracks
      (when (monk-looking-at-mark-p) (,@ body)))))

(defsubst monk-get-marked-track ()
  "Return a list of marked tracks with `monk-mark-char'."
  (let (tlis)
    (monk-walk-over-marked-tracks
     (when (listp (monk-get-track 'raw))
      (setq tlis (cons (monk-get-track 'raw) tlis))))
    (nreverse tlis)))

(defmacro monk-mark-tracks-when (condition)
  "Mark tracks if CONDITION fulfilled for the relevant track."
  (` (monk-walk-over-tracks
      (when (, condition)
       (save-excursion (monk-mark-track 1))))))

(defsubst monk-goto-first-mark (&optional mark)
  "Move to the first marked track"
  (let ((pt (save-excursion
	      (goto-char (point-min))
	      (if (search-forward-regexp
		   (concat "^\\" (or mark monk-mark-char)) nil t)
		  (point) nil))))
    (when pt (goto-char pt))))

(defmacro monk-unmark-tracks-when (condition)
  "Unmark tracks if CONDITION fulfilled for the relevant track."
  (` (monk-walk-over-tracks
      (when (, condition)
       (save-excursion (monk-unmark-track 1))))))

(defsubst monk-get-current-line-str (&optional propaswell)
  "Get current line as string.
Non-nil optional arg PROPASWELL forces copying text properties as well."
  (funcall (if propaswell
	       'buffer-substring
	     'buffer-substring-no-properties)
	   (save-excursion (beginning-of-line) (point))
	   (save-excursion (end-of-line) (point))))

(defun monk-mark-track (arg)
  "Mark ARG tracks with `monk-mark-char'."
  (interactive "p")
  (monk-only-with-monk-modes
   (let (ret)
     (monk-do-nlines
      arg
      (cond ((monk-looking-at-mark-p) t)
	    (t (setq ret (monk-insert-mark monk-mark-char)))))
     (set-buffer-modified-p nil)
     ret)))

(defun monk-unmark-track (arg)
  "Unmark the current track."
  (interactive "p")
  (monk-only-with-monk-modes
   (monk-do-nlines
    arg
    (when (monk-looking-at-mark-p)
      (monk-insert-mark " ")))))

(defun monk-mark-all-tracks ()
  "Unmark all tracks."
  (interactive)
  (monk-only-with-monk-modes
   (monk-walk-over-tracks (save-excursion (monk-mark-track 1)))))

(defun monk-unmark-all-tracks ()
  "Unmark all tracks."
  (interactive)
  (monk-only-with-monk-modes
   (monk-walk-over-marked-tracks (save-excursion (monk-unmark-track 1)))))

(defsubst monk-toggle-mark-one ()
  "Toggle mark of the current track without moving cursor."
  (when (listp (monk-get-track 'raw))
    (save-excursion
      (if (monk-looking-at-mark-p)
	  (monk-unmark-track 1) (monk-mark-track 1)))))

(defun monk-toggle-mark (arg)
  "Toggle the state of mark of ARG tracks."
  (interactive "p")
  (monk-only-with-monk-modes
   (monk-do-nlines arg (monk-toggle-mark-one))))

(defun monk-toggle-all-marks ()
  "Toggle the state of all marks."
  (interactive)
  (monk-only-with-monk-modes
   (monk-walk-over-tracks (monk-toggle-mark-one))))

(defun monk-mark-track-regex (regex)
  "Mark tracks matching with REGEX."
  (interactive "sMark tracks (regex): ")
  (monk-only-with-monk-modes
   (when regex
     (monk-mark-tracks-when
      (string-match regex (monk-get-current-line-str))))))

(defun monk-unmark-track-regex (regex)
  "Unmark tracks matching with REGEX."
  (interactive "sUnmark tracks (regex): ")
  (monk-only-with-monk-modes
   (when regex
     (monk-unmark-tracks-when
      (string-match regex (monk-get-current-line-str))))))

(defun monk-loop-marked-tracks ()
  "Loop marked tracks."
  (interactive)
  (let ((trklist (monk-get-marked-track)))
    (cond
     (trklist
      (setq monk-loop-queue (mapcar (lambda (tr) (cons tr (list 0 0))) trklist)
	    monk-loop-data monk-loop-queue) ; Sentinel knows to deal with it.
      (monk-goto-first-mark)
      ;;(apply (function monk-play-start) (car monk-loop-queue))
      (apply (function monk-play-track)
	     (append (car monk-loop-queue) '(keeploop))))
     (t (error "There's no marked track")))))

(defun monk-kill-marked-tracks ()
  "Kill marked tracks. The killed tracks are pushed on the top of `kill-ring'."
  (interactive)
  (monk-only-with-monk-modes
   (let ((first t))
     (save-excursion
       (goto-char (point-min))
       (while (not (eobp))
	 (cond ((monk-looking-at-mark-p)
		(monk-kill-current-track 1 (cond (first (setq first nil) nil)
						 (t 'append))))
	       (t (forward-line 1))))))))

(defun monk-copy-marked-tracks ()
  "Copy marked tracks. The tracks are pushed on the top of `kill-ring'."
  (interactive)
  (monk-only-with-monk-modes
   (let ((first t))
     (save-excursion
       (goto-char (point-min))
       (while (not (eobp))
	 (when (monk-looking-at-mark-p)
	   (monk-copy-current-track 1 (cond (first (setq first nil) nil)
					    (t 'append))))
	 (forward-line 1))))))

(defun monk-remove-marked-tracks ()
  "Kill marked tracks. The killed tracks are pushed on the top of `kill-ring'."
  (interactive)
  (monk-only-with-monk-modes
   (save-excursion
     (goto-char (point-min))
     (while (not (eobp))
       (cond ((monk-looking-at-mark-p)
	      (monk-remove-current-track 1))
	     (t (forward-line 1)))))))

;; 
;; Reporting total time/file size, nice when you're recording tracks into MD walkman.
;; Tue May 28 16:47:30 2002
;;
(defsubst monk-message-total-time (tt) ;; `tt' should be float.
  (message "Total: %s \(%f sec.)"
	   (monk-sec2cdtime tt) tt))

(defun monk-calc-total-time-buffer ()
  "Calculate total time of tracks in this monk buffer."
  (interactive)
  (monk-only-with-monk-modes
   (let ((tt 0.0)
	 (tlis (monk-walk-over-monk-tracks
		(lambda (tr) (monk-assq-get-value 'time tr)))))
     (mapc (lambda (tm) (setq tt (+ tt tm))) tlis)
     (monk-message-total-time tt) )))

(defun monk-calc-total-time-marked ()
  "Calculate total time of marked tracks in this monk buffer."
  (interactive)
  (monk-only-with-monk-modes
   (let* ((tracks (monk-get-marked-track))
	  (tlis (mapcar (lambda (tr) (monk-assq-get-value 'time tr)) tracks))
	  (tt 0.0))
     (cond (tracks
	    (mapc (lambda (tm) (setq tt (+ tt tm))) tlis)
	    (monk-message-total-time tt))
	   (t (message "No marked track, sister"))))))

;; This is not globally used.
(defsubst monk-get-file-size (file)
  (monk-with-coding-system
   (and (file-readable-p (expand-file-name file))
	(nth 7 (file-attributes file)))))

(defsubst monk-message-total-size (fsize)
  (message "Total %d bytes" fsize))

(defun monk-calc-total-file-size-buffer ()
  "Calculate total file size of tracks in this monk buffer."
  (interactive)
  (monk-only-with-monk-modes
    (let ((ts 0)
	  (flis (delq
		 () (monk-walk-over-monk-tracks
		     (lambda (tr) (monk-assq-get-value 'file tr))))))
      (mapc (lambda (f) (setq ts (+ ts (or (monk-get-file-size f) 0)))) flis)
      (monk-message-total-size ts))))

(defun monk-calc-total-file-size-marked ()
  "Calculate total file size of marked tracks in this monk buffer."
  (interactive)
  (monk-only-with-monk-modes
   (let* ((tracks (monk-get-marked-track))
	  (flis (mapcar (lambda (tr) (monk-assq-get-value 'file tr)) tracks))
	  (ts 0))
     (cond
      (tracks
       (mapc (lambda (f) (setq ts (+ ts (or (monk-get-file-size f) 0)))) flis)
       (monk-message-total-size ts))
      (t (message "No marked track, sister"))))))

;;
;; .m3u file generation (and map over all tracks)
;;
;; `monk-generate-playlist' seems to be more generic command in futuristic
;; MonK, so keep this section in monk.el.
;; 
;; Thu Dec 21 02:17:32 2000
;;

;; Fine. working. Generic.
;; This function may be called out of monk-buffer context,
;; So (monk-with-monk-window-and-buffer ...) must be braced.

(defun monk-walk-over-monk-tracks (fun)
  "`mapcar' function FUN over all valid MonK tracks in `monk-buffer'."
  (let (tlis prop)
    (monk-walk-over-tracks
     ;; Get only non-cd-audio (generic) `track'.
     ;; The 'file validity is not checked by `monk-get-track'
     ;; here. Okay ?
     (if (listp (setq prop (monk-get-track 'raw)))
	 (setq tlis (cons prop tlis))))
    (mapcar fun (nreverse tlis))))

;; (monk-walk-over-monk-tracks 'identity) ; get tracks as they are as a list
;; (monk-walk-over-monk-tracks		; get duration-opt, nonsense
;;  (lambda (tr) (monk-assq-get-value 'duration-opt tr)))

;; Create a buffer for .m3u for emacs-{19,20}:

(defun monk-generate-playlist ()
  "Generate playlist contents out of the current buffer.
You may see the filenames in buffers, but you ought to select correct
coding-system of the buffer before saving it."
  (interactive)
  (let ((buf (progn (if (get-buffer "*m3u*")
			(kill-buffer "*m3u*")) ; must kill, for `save-buffer'
		    (get-buffer-create "*m3u*")))
	(files (monk-walk-over-monk-tracks	; get file names from c-buffer
		(lambda (tr) (monk-assq-get-value 'file tr)))))
    (monk-with-current-buffer
     buf				; *m3u* buffer (Not monk-buffer !)
     (mapcar (lambda (fname)
	       (if fname (insert (format "%s\n" fname)))) files)
;;; Over-kill: User should do this by hand.
;;;  (cond ((boundp 'buffer-file-coding-system)
;;;         (setq buffer-file-coding-system monk-coding-system))
;;;        ((boundp 'file-coding-system)
;;;         (setq file-coding-system monk-coding-system)))
     (switch-to-buffer-other-window buf)
     (goto-char (point-min))
     ;; User should determine coding system and save by himself
     ;; (save-buffer)
     ;; emacs-20 will ask coding system anyway if necessary.
     (or (coding-system-p 'sjis)	; or maybe this is emacs-19, so,
	 (message "Now select adequate coding-system and save the playlist"))
     )))

;;;
;;; DragDrop handler
;;;

;; (require 'dired-dd) ;; No harm.  The Hook is not called from any other fun.

;; Current arglist passed by dired-dd: (fn-list to modifiers end-point)
;; You can also use variables such as `modifiers', `event', `start-window'
;; etc., which are defined in upper functions.  Refer to the caller functions
;; dired-dd-drag-drop and dired-dd-drop-from-to in dired-dd.el,
;; for the detail of the parameters.
;;
;; No use `monk-with-coding-system' here -- filename generation is
;; already done.

(defsubst monk-guess-default-directory (fn-list)
  "Guess proper current directory from file name list FN-LIST.
FN-LIST should be in full-path description."
  (let ((dirs (sort			; get the shortest (correct ?)
	       (mapcar 'file-name-directory fn-list) 'string<))
	l)
    (mapcar (lambda (e)
	      (or (member e l) (setq l (cons e l)))) dirs)
    (car (nreverse l))))

(defun monk-dired-dd-drop-handler (arglist)
  "`dired-dd' drop handler for `mpg123'.
The dropped first file or directory is loaded to the playlist by command
`mpg123', then the rest of the files are added via `mpg123-add-new'.
The `default-directory' of drop target buffer is determined by
`monk-guess-default-directory'.

Returns non-nil if it succeeded, otherwise nil."
  ;; (setq monk-last-dired-dd-drop arglist) ; debugger
  (let* ((files	;; (nreverse (car arglist))
	  (car arglist))	 ; Now `monk-find-file' has no save-ex
	 (modifiers (car (cdr (reverse arglist)))) ; use it in arglist
	 (firstdrop t)
	 (openfun (lambda (f)
		    (let ((monk-buffer (current-buffer)))
		      (monk-find-file f))))
	 (sz (buffer-size))		; target buffer
	 ;; Wed Feb 14 21:45:08 2001
	 ;; Accessing `from' in `dired-dd-drop-from-to' may be one idea,
	 ;; but I think it's a wrong thing to do these days.
	 ;; (save-window-excursion (select-window from)
	 ;;			   (dired-current-directory))
	 (dir
	  (monk-guess-default-directory files)) ; may be good enough.
	 result)
    (cond
     ((eq major-mode 'monk-mode)
      ;; Wed Feb 14 21:45:08 2001
      (setq default-directory dir)

      (cond
       ;; C-drag-mouse-2, or S-drag-mouse-2 just adds files to the playlist.
       ((or (memq 'control modifiers)
	    (memq 'shift modifiers))
	(setq firstdrop nil)))
      ;; Multiple file handling 
      (cond ((and (null (= 0 sz)) (> (length files) 1)) (push-mark (point))))
      (setq result
	    (mapcar
	     (lambda (f)
	       (cond
		(firstdrop
		 (let (buffer-read-only) (erase-buffer))
		 (setq firstdrop nil)
		 (funcall openfun f))
		(t (funcall openfun f))))
	     files))
      (cond ((= 0 sz) (goto-char (point-min))))
      ;; result ;; No ! This is a list.  Should be nil/non-nil !
      t)
     (t nil) )))

;; Register handler to the handler list.  add-hook() is good enough.
(add-hook 'dired-dd-non-dired-drop-handlers 'monk-dired-dd-drop-handler)

;;;
;;; Mouse commands
;;;
;;; monk-mouse-play-this-track() and most of the other commands are
;;; taken from ~/dired-dd/tools.el.
;;;

(defun monk-mouse-play-this-track (event &optional arg)
  "Mouse version of `monk-play-this-track'."
  (interactive "e")
  (mouse-set-point event)
  (prog1 (monk-play-this-track arg)
    ;; Duck bug of mouse-drag.el Fri Dec 29 02:21:23 2000
    (setq truncate-lines monk-truncate-lines)))

(defun monk-track-mouse-read-event (&optional debug)
  "Read mouse events till mouse button-up event.
Mouse movement, or switching frame event does not stop reading event.
Finally fixed mouse event status is returned."
  (let (event)
    (track-mouse
      (while (progn
	       (setq event (read-event))
	       (when debug
		 (message "%d:%s:%s"
			  (event-click-count event)
			  event
			  (key-binding (vector (car event)))))
	       (or (mouse-movement-p event)
		   (eq (car-safe event) 'switch-frame)))
	nil)
      event)))

(defun monk-mode-line-down-mouse (event &optional arg)
  "Hook mouse-down event, and take final fixed event via \
`monk-track-mouse-read-event'.
finally settled mouse event in `down' buffer."
  (interactive "e\nP")
  (mouse-minibuffer-check event)
  ;; Can't detect the drag/click event with mode-line prefix. Why ?
  ;; So, `fun' is unused.
  (let* ( ;; fun
	 modifiers
	 (click-count (event-click-count event))
;;; O what a idiot.
;;; I tried to set mouse ptr at current playing position on mode-line before
;;; dragging.  This is impossible, in this function, to foresee if it
;;; will be a `click' or `drag'.  Faking start postion inevitably
;;; disables `click' operation via `monk-mode-line-click-mouse'.
;;; 
;;; Further, if this function is all for the dragging preparation,
;;; there's difficulty somemore: faking mouse pixel position
;;; eventually fakes `event-start' position.  There're lot TODO
;;; to get what I wanted. Mon Dec 11 03:26:40 2000 s.n.
	 )
    (setq event (monk-track-mouse-read-event nil)
	  ;; fun (key-binding (vector (car event)))
	  modifiers (event-modifiers event))
    ;;(message"%S:%S" modifiers event) (sit-for 3)

    (cond
     ;; With sort of `control' modifier, these should
     ;; ((and (listp modifiers) (memq 'meta modifiers) (memq 'click modifiers))
     ;;  (monk-mode-line-click-mouse event arg))
     ((and (listp modifiers)
	   (memq 'control modifiers)
	   (memq 'drag modifiers))
      (monk-mode-line-C-drag-mouse event arg))
     ;; Without `control' modifier, these work
     ((equal modifiers '(click))
      (monk-mode-line-click-mouse event arg)
      )
     ((equal modifiers '(drag))
      (monk-mode-line-drag-mouse event arg))
     (t   ;; Should comment out this block later
      (message "mode-line down mouse")))))

;; (or (fboundp 'frame-parameter) (defun frame-parameter...))  is the
;; shitty `apel' way, and totally confuses programmers/users using
;; both of old and new emacs.  So, we do insist on using ours here.

(cond ((fboundp 'frame-parameter)
       (fset 'monk-frame-parameter 'frame-parameter))
      (t (defsubst monk-frame-parameter (frame param)
	   (let ((prop (assq param (frame-parameters frame))))
	     (and prop (cdr prop))))))

(defun monk-mode-line-click-mouse (event &optional arg)
  "Start playing current track as if mode-line is a time scale of the track.
`monk-loop-data' is reset before start playing from the clicked position."
  (interactive "e\nP")
  (mouse-minibuffer-check event)
  (let* ((frwidth
	  (cond
	   ;; E21.x has `full-frame-width' mode-line (?)
	   ((string< "21" emacs-version) (frame-pixel-width))
	   (t (- (frame-pixel-width)
		 (or (monk-frame-parameter (selected-frame) 'scroll-bar-width)
		     (if (string< "20" emacs-version) 10 13))))))
	 (win (select-window (posn-window (event-end event)))) ; ugly
	 (x (car (posn-x-y (event-end event))))
	 (ratio (/ (float x) frwidth))
	 (procrunning (monk-proc-running-p))
	 (onetrackloop (monk-one-track-loop-p))
	 (track (cond (procrunning monk-current-track)
		      ;; User may have moved to desired track line, so:
		      (t (monk-get-this-track))))    
	 length beg)
    (cond
     (track
      (setq length
	    (or monk-current-track-length ; == nil if *monk* buf just created.
		(and (listp track) (monk-assq-get-value 'frames track)))
	    beg (and length (round (* length ratio))))
      (when (and procrunning (markerp monk-process-marker))
	(goto-char monk-process-marker)
	(forward-line 0))
      ;; The `(and procrunning onetrackloop)' portion works `keeploop flag.
      (monk-play-track track beg 0 (and procrunning onetrackloop)))
     (t (message
	 "Should place cursor on a track first 'cause track is not played now"
	 )))))

(defun monk-mode-line-drag-mouse (event &optional arg)
  "Do `shuttle' drag for current tune.
Steps forward or backward by amount of drag."
  (interactive "e\nP")
  (mouse-minibuffer-check event)
  (let* ((click-count (event-click-count event))
	 (frwidth
	  (cond
	   ;; E21.x has `full-frame-width' mode-line (?)
	   ((string< "21" emacs-version) (frame-pixel-width))
	   (t (- (frame-pixel-width)
		 (or (monk-frame-parameter (selected-frame) 'scroll-bar-width)
		     (if (string< "20" emacs-version) 10 13))))))
	 (win (select-window (posn-window (event-end event)))) ; ugly
	 (drag-amt (- (car (posn-x-y (event-end event)))
		      (car (posn-x-y (event-start event)))))
	 (procrunning (monk-proc-running-p))
	 (onetrackloop (monk-one-track-loop-p))
	 (current-ofs (or monk-current-offset 0))
	 (track (cond (procrunning monk-current-track)
		      ;; User may have moved to desired track line, so:
		      (t (monk-get-this-track))))
	 (ratio (/ (float drag-amt) frwidth))
	 length beg)
;;;    (message "DRAG:%s:%s %s:%s %s:%s"
;;;	     frwidth drag-amt   length current-ofs  ratio beg)
    (cond
     (track
      (setq length
	    (or monk-current-track-length ; == nil if *monk* just created.
		(and (listp track) (monk-assq-get-value 'frames track)))
	    beg (+ current-ofs (* length ratio)))
      (cond
       ((and (<= 0 beg) (<= beg length))
	(when (and procrunning (markerp monk-process-marker))
	  (goto-char monk-process-marker)
	  (forward-line 0))
	(monk-play-track track beg 0 (and procrunning onetrackloop)))
       ((null track)
	(message
	 "Should place cursor on a track first \(track is not played now\)"))
       (t (message "Your drag \(%f%%\) is out of range, brother"
		   (* 100 ratio))) )))))

(defun monk-mode-line-C-drag-mouse (event &optional arg)
  "Loop dragged part of the track."
  (interactive "e\nP")
  (mouse-minibuffer-check event)
  (let* ((click-count (event-click-count event))
	 (frwidth
	  (cond
	   ;; E21.x has `full-frame-width' mode-line (?)
	   ((string< "21" emacs-version) (frame-pixel-width))
	   (t (- (frame-pixel-width)
		 (or (monk-frame-parameter (selected-frame) 'scroll-bar-width)
		     (if (string< "20" emacs-version) 10 13))))))
	 (win (select-window (posn-window (event-end event)))) ; ugly
	 (begx (car (posn-x-y (event-start event))))	 
	 (endx (car (posn-x-y (event-end event))))
	 (procrunning (monk-proc-running-p))
	 (current-ofs (or monk-current-offset 0))
	 (track (cond (procrunning monk-current-track)
		      ;; User may have moved to desired track line, so:
		      (t (monk-get-this-track))))
	 length begofs endofs dur tem)
    (cond
     (track
      (setq length
	    (or monk-current-track-length ; == nil if *monk* just created.
		(and (listp track) (monk-assq-get-value 'frames track)))
	    begofs (* length (/ (float begx) frwidth))
	    endofs (* length (/ (float endx) frwidth))
	    dur (- endofs begofs))
      ;; swap start/end (left/right).
      (when (< dur 0)
	(setq tem begofs
	      begofs endofs
	      endofs tem
	      dur (- dur)))
      ;; Zoom X4 off if drag was shorter than 10 second.
      (when (< dur (* monk-fps 10)) (setq dur (/ dur 4)))
      (when (< 0 dur)
	(when (and procrunning (markerp monk-process-marker))
	  (goto-char monk-process-marker)
	  (forward-line 0))
	  (monk-play-loop track (round begofs) (round dur))))
     ((null track)
      (message
       "Should place cursor on a track first 'cause track is not played now"))
     (t (message "You horizontally dragged 0 pixel, brother")))))

;;
;; Toggling mark by sweeping mouse-1
;; Sat Mar 24 12:26:47 2001
;;

(defun monk-track-mouse-read-event-with-mface (&optional debug)
  "`With-mouse-face' version of `monk-track-mouse-read-event' \(which see\)."
  (let (event mface-st win)
    ;; The buffer (window) is assumed to have been selected already here.
    (move-overlay mouse-drag-overlay (point) (point))
    (track-mouse
      (while (progn
	       (setq event (read-event))
	       (or (mouse-movement-p event)
		   (eq (car-safe event) 'switch-frame)))
	(when debug (message "%d:%s:%s"
			   (event-click-count event)
			   (posn-point (event-start event))
			   (key-binding (vector (car event)))))

	(or win (setq win (posn-window (event-start event))))
	(or mface-st
	    (setq mface-st (posn-point (event-start event))))

	;; Mouse face control:
	(when (and mface-st win)
	  (let* ((end (event-end event))
		 (point (posn-point (event-end event)))
		 (square (window-edges win))
		 (top (nth 1 square))
		 (bottom (cond
			  ((window-minibuffer-p win) (nth 3 square))
			  (t (1- (nth 3 square)))))
		 (ewin (posn-window (event-end event))))
	    (cond
	     ((and (eq ewin win) (integer-or-marker-p point))
	      (goto-char mface-st)
	      (goto-char point)
	      (move-overlay mouse-drag-overlay
			    mface-st point (window-buffer win)))
	     (t
	      (let ((mline (cdr (cdr (mouse-position)))))
		(cond
		 ((null mline))
		 ((< mline top)
		  (mouse-scroll-subr win (- mline top)
				     mouse-drag-overlay mface-st)
		  ;; `mouse-drag-region' (mouse.el) in some emacs version
		  ;; (eg. 20.7/21.0.x) does weird stuff here.
		  )
		 ((>= mline bottom)
		  (mouse-scroll-subr win (1+ (- mline bottom))
				     mouse-drag-overlay mface-st)
		  ;; `mouse-drag-region' (mouse.el) in some emacs version
		  ;; (eg. 20.7/21.0.x) does weird stuff here.
		  ))))))) ))
    (delete-overlay mouse-drag-overlay)
    event))

(defun monk-mouse-sweep-region (event)
  "Toggles the mark of tracks in the region swept by mouse."
  (interactive "e\nP")
  (let ((rb (posn-point (event-start event)))
	(re (posn-point (event-end event)))
	(w1 (posn-window (event-start event)))
	(w2 (posn-window (event-start event)))
	tem)
    ;; Non-integer (nil, mode-line etc.) will be yielded
    ;; if mouse is released out of the window boundary.
    ;; For that case, we can use (point), 'cause
    ;; `monk-track-mouse-read-event-with-mface' updates (point) while draggin.
    ;; (or just use (point) and unuse `re' (?) -- ugly though)
    (or (integer-or-marker-p re) (setq re (point)))
    (when (> rb re) (setq tem re re rb rb tem))
    (goto-char rb)
    (while (and (< (point) re) (null (eobp)))
      (save-excursion (monk-toggle-mark 1))
      (forward-line 1))
    (goto-char re)
    (sit-for 0)))

(defun monk-down-mouse (event &optional arg)
  "Hook mouse-down event, and take final fixed event via \
`monk-track-mouse-read-event'.
Finally call a function bound to the settled mouse event by mouse up event."
  (interactive "e\nP")
  (mouse-minibuffer-check event)
  (mouse-set-point event)
  (let* (fun
	 modifiers
	 (click-count (event-click-count event)))
    ;; (setq event (monk-track-mouse-read-event nil)
    (setq event (monk-track-mouse-read-event-with-mface nil)
	  fun (key-binding (vector (car event)))
	  modifiers (event-modifiers event))
    ;; (message"%S:%S:%S" modifiers event fun) (sit-for 3)
	      ;; This can't handle prefix arg.
	      ;;(setq unread-command-events
	      ;;  (cons event unread-command-events))
    ;; The `arg' value should be propagated implicitly
    ;; (eg. to mouse-set-point).
    (when fun (funcall fun event))))

;;
;; Misc mouse commands.
;; Volume control by mouse wheel, yanking ...
;;

(defun monk-mouse-volume-up (arg)
  "Swell up playing volume by numeric ARG by base of 4%. Relative \
control only."
  (interactive "p")
  (monk-volume-up (* 4 arg)))

(defun monk-mouse-volume-down (arg)
  "Lower playing volume by numeric ARG by base of 4%. Relative control only."
  (interactive "p")
  (monk-volume-down (* 4 arg)))

(defun monk-mouse-yank-track (event &optional arg)
  "Yank back track\(s\) by mouse."
  (interactive "e\np")
  (mouse-set-point event)
  (monk-yank-track arg))

(defun monk-mouse-play-prev (event &optional arg)
  "Mouse-driven version of `monk-play-prev'."
  (interactive "e\np")
  (mouse-set-point event)
  (monk-play-prev arg))

(defun monk-mouse-play-next (event &optional arg)
  "Mouse-driven version of `monk-play-next'."
  (interactive "e\np")
  (mouse-set-point event)
  (monk-play-next arg))

;; Just bind `monk-kill-marked-tracks' to double-mouse-3 is good enough, so
;; this is unused.
;; (define-key monk-mode-map [double-mouse-3] 'monk-mouse-kill-marked-tracks)
;; (defun monk-mouse-kill-marked-tracks (event)
;;  (interactive "e\np")
;;  (monk-kill-marked-tracks))

;;;
;;; Undocumented Lazy CD ripping support
;;; Mon Nov 27 21:12:18 2000
;;;
;;; Just generate shell script, and the displays the buffer.
;;; 
;;; M-x monk-generate-ripper-script RET
;;; is the only entry point
;;;

(defvar monk-mp3-extension ".mp3")
(defvar monk-mp3-directory "/tmp")
(defvar monk-ripper-command "monk-ripper")
(defvar monk-ripper-options nil)
(defconst monk-mp3-junk-buffer "*monk*mp3*junk*" "For junk script output.")

(eval-when-compile (require 'dired))	; dired-replace-in-string

(defun monk-cddb-get-title-list ()
  "Get album title list out of `monk-cddb-data'.
Returns nil if failed."
  (let ((lis (nthcdr 2 monk-cddb-data)) num title)
    (mapcar
     (lambda (data)
       (string-match "TTITLE[0-9]+=\\(.*\\)$" data)
       (substring data (match-beginning 1) (match-end 1))) lis)))

(defun monk-generate-album-directory ()
  "Generate a directory name in which mp3 file should be saved.
Basename part is generated by `monk-cddb-get-album-title'.
See `monk-mp3-directory'."
  (let ((ttl (monk-cddb-get-album-title)))
    (cond (ttl
	   (format "%s/%s" monk-mp3-directory 
		   (dired-replace-in-string " \/ " " - "
					    ttl)))
	  ;;(t (error "MonK can't get album title"))
	  (t (format "%s/%s" monk-mp3-directory "Untitled"))
	  )))

;;
;; monk-generate-ripper-script()
;;
;; In future version,`monk-play-start' should be aware of
;; `ripper-mode', and take the titles to pass the ripper command,
;; then `start-process'.

;; Perhaps I don't have to support monk-coding system here.
;; Have you seen cddb record with SJIS encoding ?

(defun monk-generate-ripper-script (&optional extra-opt)
  "Generate ripper shell script."
  (interactive "P")
  (and monk-cddb-data
       (let* ((num) (lis)
	      (cmdstr monk-ripper-command)
	      (tlist (monk-cddb-get-title-list))
	      (header (format "#!%s -x" shell-file-name))
	      ;; `dired-shell-stuff-it' does not take care of / in dname/fname
	      ;; This handles error():
	      (dir (expand-file-name (monk-generate-album-directory)))
	      (mkdir (dired-shell-stuff-it "mkdir" (list "-p" dir) nil))
	      (chdir (dired-shell-stuff-it "cd" (list dir) nil))
	      (buf (get-buffer-create monk-mp3-junk-buffer)))
	 ;; (monk-with-monk-buffer
	 (progn
	  (goto-char (point-min))
	  (while (setq num (monk-get-this-track))
	    (cond ((listp num)		; generic track
		   (setq num (monk-assq-get-value 'track-number num))))
	    (setq lis
		  (cons
		   (dired-shell-stuff-it
		    cmdstr
		    (list
		     ;; cddb track # starts from 0
		     (dired-replace-in-string
		      "\/" "\\\/" 
		      (concat (nth (1- num) tlist) monk-mp3-extension))
		     "-t" (int-to-string num)
		     (if extra-opt extra-opt "")
		     ) nil) lis))
	    (forward-line 1)))
	 (setq lis (append (list header mkdir chdir) (reverse lis)))

	 (monk-with-current-buffer
	  buf;; (widen)
	  (erase-buffer)
	  (mapcar (lambda (l)
		    (insert l "\n")) lis) )
	 ;;(display-buffer buf)
	 (switch-to-buffer-other-window buf)
	 )))
;; TEST
;; (monk-generate-ripper-script "-e")

;;;
;;; Buffer Local definitions
;;;
;;; `monk-deflocalvar' macro will not be used until I have convinced
;;; that this multiple *monk* buffer design is really working.
;;;

(make-variable-buffer-local 'monk-fps)
;; (make-variable-buffer-local 'monk-current-process) ;; Now deprecated.
;; (make-variable-buffer-local 'monk-process-marker) ; New in 5.11, but not here
(make-variable-buffer-local 'monk-current-track)
(make-variable-buffer-local 'monk-current-track-type)
(make-variable-buffer-local 'monk-current-track-length)
(make-variable-buffer-local 'monk-current-start-sec)
(make-variable-buffer-local 'monk-current-offset)
(make-variable-buffer-local 'monk-current-cent)
(make-variable-buffer-local 'monk-current-track-name)
(make-variable-buffer-local 'monk-current-loop-name)
;; (make-variable-buffer-local 'monk-track-timer-handler) ; this is global
;; (make-variable-buffer-local 'monk-process-started-buffers) ; global
(make-variable-buffer-local 'monk-track-timer-start-sec)
(make-variable-buffer-local 'monk-track-timer-start-usec)
(make-variable-buffer-local 'monk-track-timer-obj)
(make-variable-buffer-local 'monk-loop-data)
(make-variable-buffer-local 'monk-loop-queue)
(make-variable-buffer-local 'monk-repeat)
;; (make-variable-buffer-local 'monk-disc-info)
;; (make-variable-buffer-local 'monk-track-list)
;; (make-variable-buffer-local 'monk-cddb-do-save)
;; (make-variable-buffer-local 'monk-shuttle-search-resolution)
;; (make-variable-buffer-local 'monk-readin-counter)

;;
;; Debugger for global and local variables
;;

(defsubst monk-pp-eval-exp (lis &optional buf)
  (require 'pp)
  (with-output-to-temp-buffer (or buf "*monk*dump*") (pp lis))
  (monk-with-current-buffer (get-buffer (or buf "*monk*dump*"))
			    (emacs-lisp-mode)))

(defun monk-dump-local-var ()
  (interactive)
  (monk-pp-eval-exp
   (let (monk-local)
     (mapcar (lambda (c)
	       (if (local-variable-if-set-p c)
		   (setq monk-local (cons (cons c (eval c)) monk-local))))
	     (apropos-internal "monk" 'boundp))
     (nreverse monk-local))))

(defun monk-dump-global-var ()
  (interactive)
  (monk-pp-eval-exp
   (let (monk-global)
     (mapcar (lambda (c)
	       (if (null (local-variable-if-set-p c))
		   (setq monk-global (cons (cons c (eval c)) monk-global))))
	     (apropos-internal "monk" 'boundp))
     (nreverse (delq (assq 'monk-global monk-global) monk-global)))))

;;
;; Load Hook
;;
(defvar monk-load-hook nil)
(run-hooks 'monk-load-hook)

;; BbAbGbFbEbbDbbBbAbGbFbEbbDbbBbAbGbFbEbbDbbBbAbGbFbEbbDbb
(defun thelonious (arg)
  (interactive "*p")
  (let (glide)
    (cond ((< arg 0)
	   (setq arg (- arg)
		 glide "BbAbGbFbEbbDbb"))
	  (t (setq glide "a#g#f#edc")))
    (while (> arg 0) (insert glide) (setq arg (1- arg)))
    (insert "\n")))
;; a#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edca#g#f#edc
