#!/usr/local/bin/perl
#!/usr/bin/perl

use Getopt::Std;

# use strict;			# Find:: barfs with this.

# Local CDDB database directories.
# Must contain cddb files with `0xHEXDISCID' filename.
#
# Specifying non-existing directory does no harm except warning like
#   "Can't stat /home/user/monk-cddb: No such file or directory"
@cddbdirs=(
	   '~/monk-cddb',	# monk-cddb's
	   '~/.cddbslave/',	# Guess this was for a CDplayer in kde/gnome
	   '/opt/kde/share/apps/kscd/cddb',
	   '/var/X11R6/lib/xmcd/cddb' # xmcd's
	   );

# The remote CDDB server host, and protocol
$cddbserver = "cddb.cddb.com";
# $cddbserver = "freedb.freedb.org";
$cddbproto = 888;

# CDROM drive
$cddbcdrom = "/dev/cdrom";

# cdda2wav device definition
# If undefined, cdda2wav will see $CDDA_DEVICE environment variable
# $cdda2wavdev = "-D/dev/sg1";
# $cdda2wavdev = "-D0,1,0";

# Query parameters file extension (without ".") for -P option
$qparamext="qprm";

# Override the defaults above with user config ~/.monk-cddb
$monk_cddb_rc = "$ENV{'HOME'}/.monk-cddb";
require $monk_cddb_rc if (-r $monk_cddb_rc);

# CDDB_get:: related.
$config{CDDB_HOST} = $cddbserver; # Remote cddb server
$config{CDDB_PORT} = $cddbproto;  # `cddb' protocol
$config{CD_DEVICE} = $cddbcdrom;

my %config;

my $total;
my $toc;

# User interaction option for CDDB_get::
# 0: no user interaction
# 1: ask user if more than one possibility
$config{input}=0;		# Default: Non interactive

# Maybe supplied as ARGV
$discidcode = 0;		# or /CDDB discid:\s*0x/ from cdda2wav data
@queryparam;
$usearg=0;

# Get options

unless (&getopts('hvIrlpPLCidL')) { &usage ; die; }
$::opt_h and &usage and exit;
$::opt_v and $verbose = 1;
$::opt_i and $idonly = 1;
$::opt_r and $remotesearch = 1;
$::opt_l and $localsearch = 1;
$::opt_d and $debug = 1;
$::opt_p and $getqparam = 1;
$::opt_P and $param_file = 1 and $getqparam = 1;
$::opt_L and $skiplocalsearch = 1;
$::opt_p and $getqparam = 1;
$::opt_p and $getqparam = 1;
$::opt_C and $useCDDB_get = 1;
$::opt_I and $config{input} = 1;


#
# Well, then reel it now... 
#

# -p: extract query parameters via cdda2wav, and exit
if ($getqparam) {
    &get_queryparam;
    exit;
}

if ($#ARGV >= 0) {
    $discidcode = shift;		# ARGV[0];
    @queryparam = @ARGV;	# the remainder
    $usearg = 1;
    # print "z1: @queryparam\n";
    # print "z2: $discidcode\n";
    # print "z3: $usearg\n";
}

$discidcode = &get_discidcode() unless ($discidcode);

if ($idonly) {
    print "$discidcode\n"; exit;
}

# Search local cddb first, unless ($remotesearch)
unless ($skiplocalsearch) {
    if ($localsearch || (! $remotesearch)) {
	use File::Find;

	while (<@cddbdirs>) {
	    &find_cddb($_);
	}
    }
}

# Quit here if local search only
exit if ( $localsearch && (! $remotesearch));

#
# Two methods of remote search: 
#  &cddb_jam_XXXX:        Use ARGV as query parameters
#  &get_remote_CDDB_get:  No ARGV is required, and (but) probes CDROM drive
#

# If query parameters are supplied as @ARGV, use
# Force remote search with @ARGV, and quit.
if ($usearg) {
    # use IPC::Open2;		# for telnet method. Unused.
    use FileHandle;

    # &cddb_jam_telnet;         # Not reliable. Unused.
    &cddb_jam_cddbcmd;
    exit;
}

# Otherwise drops into remote search with CDDB_get perl module...

#
# &find_cddb() exits if found, or drops into remote search unless found
# The CD drive is probed here again
#

# With CDDB_get(3) module, for example:
# ./monk-cddb -LC
if ($useCDDB_get) {
    use CDDB_get;
    use Data::Dumper;

    print STDERR
	"Making CDROM probe, and querying $cddbserver via CDDB_get\n"
	    if ($verbose);
    &get_remote_CDDB_get;
} else {
    # Occurs when -vL
    print STDERR "Skipped querying $cddbserver via CDDB_get\n" if ($verbose);
}

exit;

#
# My cddb_jam -- use @ARGV as server search param without probebing CD drive
#

# FIXME:
# Error check is not working ???
# Why open(QQ, "cddbcmd ...|") wouldn't die on "Unknown host!" exit 111 of
# cddbcmd  ?

# /tmp/xxxxxxxxxxxxx is of existence:
# $ perl -e 'open("/tmp/xxxxxxxxxxxxx") or die "xxx\n";' ; echo $?
# xxx
# 2
# That worked nice.  But,
# $ perl -e "open(X,'/bin/junk|') or die \"mmm\n\"; print \"nnn\n\";" ; echo $?
# nnn
# 0
# (perl 5.005_03 with libc-2.1.2.so / linux)

sub cddb_jam_cddbcmd {
    my ($cat,$id,$rest);

    print STDERR
    "cddbcmd -l3 -h $cddbserver:$cddbproto cddb query $discidcode @queryparam\n"
	if ($verbose);
    # -l3 is requred for freedb.freedb.org ?
    open(QQ,
    "cddbcmd -l3 -h $cddbserver:$cddbproto cddb query $discidcode @queryparam|")
	or die "Can't connect to $cddbserver with protocol $cddbproto.\n";
    while(<QQ>) {
	($cat,$id,$rest) = ($_ =~ /(\S+)\s+(\S+)\s+(.*)/);
    }
    print STDERR
	"cddbcmd -l3 -h $cddbserver:$cddbproto cddb read $cat $id\n"
	    if ($verbose);
    open(QQ, "cddbcmd -l3 -h $cddbserver:$cddbproto cddb read $cat $id|")
	or die "Can't connect to $cddbserver with protocol $cddbproto again.\n";

    while(<QQ>) {
	s/\r//g;
	print $_;
    }
}

# Not stable, currently unused.

sub cddb_jam_telnet {
    my  $pid = open2( \*Reader, \*Writer, "telnet $cddbserver $cddbproto");
    Writer->autoflush();

    print STDERR "cddb hello root nowhere.com fastrip 0.77\n" if ($verbose);
    print Writer "cddb hello root nowhere.com fastrip 0.77\n";

    while ($got = <Reader>) {
	print STDERR $got if ($verbose);

	# die "cddb server access/query error " unless ($err =~ /^2/);
	die "Error" if ($got =~ /^[45][0-9]/);

	if ($got =~ /^200/) {
	    print STDERR "cddb query $discidcode @queryparam\n" if ($verbose);
	    print Writer "cddb query $discidcode @queryparam\n";
	} elsif ($got =~ /^(2\d\d)\s+Found/) {

	    $got = <Reader>;	# Read 1st entry only.
	    my ($cat,$id,$rest) = 
		$got =~ /(\S+)\s+(\S+)\s+(.*)/;

	    # print "X:$got";
	    CAT: while ($got = <Reader>) {	# Ignore until next "^.$"
	    # print "Y:$got";
		last CAT if ($got =~ /^\.$/);
	    }

	    # Get the relevant data
	    print STDERR "cddb read $cat $id\n" if ($verbose);
	    print Writer "cddb read $cat $id\n";

	    # Sometimes "^.$" is sent (?) here.  What is this ?
	    #print "$got" unless ($got =~ /^\.$/);

	    while ($got = <Reader>) {
		if ($got =~ /^\.$/) {
		    return 0;
		} elsif ($got =~ /^(2\d\d)\s+/) {
		    print STDERR "$got" if ($verbose);
#		} elsif ($got =~ /Connection closed/) {
#		    ;
		} else {
		    print $got;
		}
	    }
	}
    }
}

#
# CDDB_get --- probe CD drive, then search server
#
sub get_remote_CDDB_get {

    my %cd=get_cddb(\%config);

    print STDERR ("Querying remote server $cddbserver\n") if ($verbose);
    unless(defined $cd{title}) {
	die "no cddb entry found";
    }
    print_xmcd(\%cd);
}

sub print_xmcd {
    my $cd=shift;
    for(@{$cd->{raw}}) {
	s/\r//g;
	print $_;
    }
}   

#
# Find routines
# Based on a script generated by "/usr/bin/find2perl"
#

# In these routines, referring to _ like this:
#     print ("N:$_\n");  
# alters the result ...

sub find_cddb {
    # Can't use $_/@_ directly here (See CDDB_get(3))
    my($dir) = @_;
    find(\&wanted, $dir);
}

# Quit on first hit if (!$remotesearch)
sub wanted {
    my($nm) = $File::Find::name;

    print("Scanning: $File::Find::dir;\n") if ($debug);

    # both $_ and _ are used. Is this okay ? 
    if ( # exist-p, and test -f
	(($dev,$ino,$mode,$nlink,$uid,$gid) = lstat($_))
	 && (-f _  || -l _) &&
	/^$discidcode$/ #&&		# and the name is $discidcode$

	) {

	# There is a total confusion on comma syntax after these file handles.
	# Or LDB is not `handle' like STDERR and being `FILE' structure in C ?
	print STDERR "Local cddb file found: $nm\n" if ($verbose);

	open(LDB, "<$nm") or die "... but can't open local cddb file $nm";
	print $_ while(<LDB>);
	close (LDB);

	exit if (!$remotesearch);
    }
}

#
# cdda2wav related routines.
# Get discidcode only (for this script)
#

# Only for "CDDB discid" (mostly for local filename)
sub get_discidcode {
#    open(PROBE, "cdda2wav $cdda2wavdev -NJv32 2>&1|")
    open(PROBE, "cdda2wav $cdda2wavdev -gNJv34 2>&1|")
	or die "$0: Can't execute cdda2wav.\n";

    while (<PROBE>) {
	chomp;
	if (/CDDB discid:\s*0x/) {
	    s/CDDB discid:\s*0x//;
	    return $_;
	}
    }
}

sub get_queryparam {
    my $param_id;
    my (@param_trk_ttl);
    my $param_offsets =0;
    my $output;

# cdda2wav version dependent. fixed. Sun Dec  3 12:48:17 2000 s.n.
#    open(PROBE, "cdda2wav $cdda2wavdev -NJv32 2>&1|")
    open(PROBE, "cdda2wav $cdda2wavdev -gNJv34 2>&1|")
	or die "$0: Can't execute cdda2wav.\n";

    while (<PROBE>) {
	chomp;

	$param_id =
	    &fetch_discid($_) if (/CDDB discid:\s*0x/);
	@param_trk_ttl =
	    &get_tracks_total($_) if ($_ =~ /^Tracks:/);
	if ($_ =~ /^T\d+:\s+/) {
	    if ($param_offsets) {
		$param_offsets = "$param_offsets" . " " . &get_offsets($_);
	    } else {
		$param_offsets = "$param_offsets" . &get_offsets($_);
	    }
	}
    }

    $output = "$param_id $param_trk_ttl[0] $param_offsets $param_trk_ttl[1]\n";
    if ($param_file) {
	open (CDDBF, ">$param_id".".$qparamext")
	    or die "Can't open file $param_id\n";
	print CDDBF $output;
	print STDERR $output if ($verbose);
	print STDERR "CDDB query parameter file \"".
	    "$param_id".".$qparamext"."\" is generated\n" if ($verbose);
    } else {
	print "$param_id $param_trk_ttl[0] $param_offsets $param_trk_ttl[1]\n";
    }
}

sub fetch_discid {
# CDDB discid: 0x2811d113
    my($did) = @_;

    # print "C:$_\n";
    $did =~ s/CDDB discid:\s*0x//;
    return $did;
}

sub get_tracks_total {
# Tracks:19 76:03.72
    my($trk) = @_;
    my($ttl) = @_;
    my($min); my($sec);

    $trk =~ s/Tracks://;
    $trk =~ s/\s.*//;

    $ttl =~ s/^.*\s//;
    ($min, $sec) = split(":", $ttl);

    # print "XX:trk: $trk\n";
    # print "YY:ttl: $ttl\n";

    # Just int(). Some error seems to be allowed by server.
    $sec = int($sec);
    $ttl = $min * 60 + $sec;

    return ($trk,$ttl);
}

sub get_offsets {
# T01:       0  3:49.55 audio linear copydenied stereo title '' from ''
# T02:   17230  7:35.27 audio linear copydenied stereo title '' from ''
    my($line) = @_;
    $line =~ s/^T\d+:\s+//; # cut head
    $line =~ s/\s.*$//;	    # cut tail
    return $line;
}

sub usage {
    print
"Usage: $0 [-hvirlpPLCIdL] [query_params]

    -v: verbose
    -i: output discid only
    -r: remote search
    -l: local  search
    -h: this message
    -p: probe CDROM for query parameters
    -P: probe CDROM and write query parameters to file DISCID.$qparamext
    -L: Skip local search
    -C: use CDDB_get (local search, then probe CDROM and query cddb server)
    -I: interactive (meaningful only when -C)
    -d: debug

<cddb_parameters> are:
    <discid> <tracks> <ofs1> <ofs2> ... <ofsn> <min>
";
}
